

#include "cnfc.h"


/************************************************************************/
/************************************************************************/


float cnFcInitialValue = -999.0F;


/************************************************************************/
/************************************************************************/


void cnFcInit(cnTFC * fc)
{
  int profile;
  int level;
  int var;


assert(fc != NULL);


  for ( profile = 0; profile < CN_NC_MAX_TIME_DIM; profile++ )
  {
    cnTFCProfile * p = &fc->profiles[profile];

    for ( level = 0; level < CN_NC_MAX_LEVEL_DIM; level++ )
    {
      for ( var = 0; var < CN_NC_MAX_LEVEL_VARS; var++ )
      {
        p->levels[level][var] = cnFcInitialValue;
      }

      for ( var = 0; var < CN_NC_MAX_FLUX_VARS; var++ )
      {
        p->fluxes[level][var] = cnFcInitialValue;
      }
    }

    /* One more flux. */
    for ( var = 0; var < CN_NC_MAX_FLUX_VARS; var++ )
    {
      p->fluxes[level][var] = cnFcInitialValue;
    }

    for ( var = 0; var < CN_NC_MAX_SFC_VARS; var++ )
    {
      p->surface[var] = cnFcInitialValue;
    }

    p->forecastTime = (int) cnFcInitialValue;
  }

  fc->numProfiles    = 0;
  fc->numLevels      = 0;
  fc->numLevelVars   = 0;
  fc->numFluxVars    = 0;
  fc->numSurfaceVars = 0;

  for ( var = 0; var < CN_NC_MAX_LEVEL_VARS; var++ )
  {
    fc->levelIds[var] = -1;
  }

  for ( var = 0; var < CN_NC_MAX_FLUX_VARS; var++ )
  {
    fc->fluxIds[var] = -1;
  }

  for ( var = 0; var < CN_NC_MAX_SFC_VARS; var++ )
  {
    fc->surfaceIds[var] = -1;
  }

  fc->firstTime  = 0;
  fc->step       = 0;
  fc->utcOffset  = 0;
  fc->model      = -1;
  fc->site       = -1;
  fc->coords.lon = 0.0;
  fc->coords.lat = 0.0;
  fc->hres       = 0.0;
}


int cnFcGetProfileTime(cnTFC * fc, int profile)
{
assert(fc != NULL);
assert(0 <= profile && profile < fc->numProfiles);

  return fc->firstTime + (profile * fc->step);
}


int cnFcUp(cnTFC * fc, int pVar)
{
  int profile;
  int retValue = -1;

assert(fc != NULL);
assert(fc->numLevels > 0);
assert((pVar < 0) || (0 <= pVar && pVar < fc->numLevelVars));


  if ( pVar < 0 )
  {
    pVar = cnFcSearchLevelIds(fc, CN_VAR_PRESSURE);
    if ( pVar < 0 )
    {
      return -1;
    }
  }

  /* Find the first profile with valid data. */
  for ( profile = 0; profile < fc->numProfiles; profile++ )
  {
    if ( fc->profiles[profile].levels[0][pVar] > cnFcInitialValue
      && fc->profiles[profile].levels[fc->numLevels - 1][pVar]
        > cnFcInitialValue )
    {
      retValue = ( fc->profiles[profile].levels[0][pVar]
        > fc->profiles[profile].levels[fc->numLevels - 1][pVar] )
        ? 1 : 0;
      break;
    }
  }

  return retValue;
}


void cnFcReverseLevels(cnTFC * fc)
{
  int profile;


assert(fc != NULL);


  for ( profile = 0; profile < fc->numProfiles; profile++ )
  {
    cnTFCProfile * p = &fc->profiles[profile];
    int            loLevel;
    int            hiLevel;
    int            var;

    if ( fc->numLevelVars > 0 )
    {
      for ( loLevel = 0, hiLevel = fc->numLevels - 1;
        loLevel < hiLevel; loLevel++, hiLevel-- )
      {
        for ( var = 0; var < fc->numLevelVars; var++ )
        {
          float temp              = p->levels[loLevel][var];
          p->levels[loLevel][var] = p->levels[hiLevel][var];
          p->levels[hiLevel][var] = temp;
        }
      }
    }

    if ( fc->numFluxVars > 0 )
    {
      for ( loLevel = 0, hiLevel = fc->numLevels;
        loLevel < hiLevel; loLevel++, hiLevel-- )
      {
        for ( var = 0; var < fc->numFluxVars; var++ )
        {
          float temp              = p->fluxes[loLevel][var];
          p->fluxes[loLevel][var] = p->fluxes[hiLevel][var];
          p->fluxes[hiLevel][var] = temp;
        }
      }
    }
  }
}


int cnFcSearch(cnTFC * fc, int time)
{
  int retValue = -1;


assert(fc != NULL);


  time -= fc->firstTime;
  if ( time >= 0 )
  {
    retValue = time / fc->step;
    if ( retValue >= fc->numProfiles || (time % fc->step) != 0 )
    {
      retValue = -1;
    }
  }

  return retValue;
}


static int cnFcSearchInts(int * a, int n, int i)
/* Search the array 'a' with 'n' elements for 'i'. */
{
assert(a != NULL);

  for ( n--; n >= 0 && a[n] != i; n-- ) ;
  return n;
}


int cnFcSearchLevelIds(cnTFC * fc, int id)
{
assert(fc != NULL);

  return cnFcSearchInts(fc->levelIds, fc->numLevelVars, id);
}


int cnFcSearchFluxIds(cnTFC * fc, int id)
{
assert(fc != NULL);

  return cnFcSearchInts(fc->fluxIds, fc->numFluxVars, id);
}


int cnFcSearchSurfaceIds(cnTFC * fc, int id)
{
assert(fc != NULL);

  return cnFcSearchInts(fc->surfaceIds, fc->numSurfaceVars, id);
}


void cnFcGetVarProfile(cnTFC * fc, float * data /*OUT*/,
  int profile, int var)
{
  cnTFCProfile * p;
  int            level;


assert(fc != NULL);
assert(data != NULL);
assert(0 <= profile && profile < fc->numProfiles);
assert(0 <= var && var < fc->numLevelVars);


  p = &fc->profiles[profile];
  for ( level = 0; level < fc->numLevels; level++ )
  {
    data[level] = p->levels[level][var];
  }
}


void cnFcGetFluxProfile(cnTFC * fc, float * data /*OUT*/,
  int profile, int var)
{
  cnTFCProfile * p;
  int            level;


assert(fc != NULL);
assert(data != NULL);
assert(0 <= profile && profile < fc->numProfiles);
assert(0 <= var && var < fc->numFluxVars);


  p = &fc->profiles[profile];
  for ( level = 0; level <= fc->numLevels; level++ )
  {
    data[level] = p->fluxes[level][var];
  }
}


void cnFcGetSurfaceVar(cnTFC * fc, float * data /*OUT*/, int var)
{
  int profile;


assert(fc != NULL);
assert(data != NULL);
assert(0 <= var && var < fc->numSurfaceVars);


  for ( profile = 0; profile < fc->numProfiles; profile++ )
  {
    data[profile] = fc->profiles[profile].surface[var];
  }
}


/************************************************************************/
/************************************************************************/
