/*
 * Stuff for the final forecast data set.
 *
 * HISTORY
 * 2004-03-19 ao developed from 'cndataset' module of previous versions.
 *
 * AUTHORS
 * Andrew Openshaw
 */


#if !defined(CNFC_H)
#define CNFC_H


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <math.h>
#include "cnutils.h"
#include "cnmodels.h"
#include "cnsites.h"
#include "cnnetcdf.h"


/************************************************************************/
/************************************************************************/


/* Structure for a profile. */
typedef struct
{
  /* Minutes since initialization of the model run that this profile
   * was taken from.
   */
  int   forecastTime;

  float levels  [CN_NC_MAX_LEVEL_DIM]     [CN_NC_MAX_LEVEL_VARS];
  float fluxes  [CN_NC_MAX_LEVEL_DIM + 1] [CN_NC_MAX_FLUX_VARS];
  float surface [CN_NC_MAX_SFC_VARS];
} cnTFCProfile;


/* Structure for all the profiles for a data set. */
typedef struct
{
  int          firstTime; /* Time of first profile. Minutes since 1960. */
  int          step;      /* Number minutes between each profile. */
  int          utcOffset;
  int          model;     /* one of 'cnEModels' from "cnmodels.h". */
  int          site;      /* one of 'cnESites' from "cnsites.h". */
  cnTSiteCoords coords;   /* coordinates of the grid point. */
  float        hres;      /* Horizontal resolution of the model. */

  cnTFCProfile profiles[CN_NC_MAX_TIME_DIM];

  int          numProfiles;    /* Number of profiles in the dataset. */
  int          numLevels;      /* Number of levels in each profile. */
  int          numLevelVars;   /* Number of variables in each level. */
  int          numFluxVars;    /* Number of flux variables. */
  int          numSurfaceVars; /* Number of variables for each surface. */


  /* The identifiers stored below should be the 'cnEVariableIds' in
   * "cnmodels.h".
   */

  /* This is an array containing the identifiers for each variable
   * in the order they are stored in a level.
   */
  int          levelIds[CN_NC_MAX_LEVEL_VARS];


  /* This is an array containing the identifiers for each flux variable
   * in the order they are stored in a flux array.
   */
  int          fluxIds[CN_NC_MAX_FLUX_VARS];


  /* This is an array containing the identifier for each surface
   * variable in the order they are stored in the surface array
   * for each profile.
   */
  int          surfaceIds[CN_NC_MAX_SFC_VARS];
} cnTFC;


/************************************************************************/
/************************************************************************/


/* This is the value all variables are initialised to.
 * It is initially set to -999.0.
 */
extern float cnFcInitialValue;


/************************************************************************/
/************************************************************************/


/* Initialises the dataset 'fc'.
 * PRE: fc!=NULL;
 * POST: 
 * All variable's data values and frequencies are set to
 * 'cnFcInitialValue'.
 * All 'ids' and 'model' are set to -1.
 * All other fields are set to 0.
 */
void cnFcInit(cnTFC * fc);


/* Get the time of a profile.
 * PRE: fc!=NULL; 0<=profile<fc->numProfiles;
 * 'profile' is the index into 'fc->profiles'.
 * POST: returns the number of minutes from 1960 of the profile.
 */
int cnFcGetProfileTime(cnTFC * fc, int profile);


/* Test direction of levels.
 * PRE: fc!=NULL; fc->numLevels>0; (pVar<0)||(0<=pVar<fc->numLevelVars);
 * 'pVar' is the index into the levels for the pressure variable.
 * If 'pVar'<0 then the CN_VAR_PRESSURE ID is searched for in
 * 'fc->levelIds' otherwise 'pVar' is used.
 * POST: returns:
 *   -1 if CN_VAR_PRESSURE was not found in 'levelIds';
 *   1  if the levels from 0..fc->numLevels-1 are moving upwards
 *      in the atmosphere;
 *   0  if they are moving downwards.
 */
int cnFcUp(cnTFC * fc, int pVar);


/* Reverse the order of the levels, fluxes and MWPPS in each profile in
 * 'fc'.
 * PRE: fc!=NULL;
 */
void cnFcReverseLevels(cnTFC * s);


/* Search for the first profile with 'time'.
 * PRE: fc!=NULL;
 * 'time' is minutes from 1960.
 * POST: returns index into 'fc->profiles' or -1 if there is no profile
 * for 'time'.
 */
int cnFcSearch(cnTFC * fc, int time);


/* Search 'fc->levelIds' for 'id'.
 * PRE: fc!=NULL;
 * POST: returns -1 if 'id' not found in 'fc->levelIds[]' otherwise
 * returns the index into 'fc->levelIds[]'.
 */
int cnFcSearchLevelIds(cnTFC * fc, int id);


/* Search 'fc->fluxIds' for 'id'.
 * PRE: fc!=NULL;
 * POST: returns -1 if 'id' not found in 'fc->fluxIds[]' otherwise
 * returns the index into 'fc->fluxIds[]'.
 */
int cnFcSearchFluxIds(cnTFC * fc, int id);


/* Search 'fc->surfaceIds' for 'id'.
 * PRE: fc!=NULL;
 * POST: returns -1 if 'id' not found in 'fc->surfaceIds[]' otherwise
 * returns the index into 'fc->surfaceIds[]'.
 */
int cnFcSearchSurfaceIds(cnTFC * fc, int id);


/* Get all the data values from each level for one variable in a profile.
 * PRE: fc!=NULL; data!=NULL; 0<=profile<fc->numProfiles;
 * 0<='var'<fc->numLevelVars;
 * 'data' must have space for at least 'fc->numLevels' floats.
 * 'profile' is the index into 'fc->profiles' for the profile to copy
 * from.
 * 'var' is the index into the variables for each level.
 * POST: 'data' will contain the data for the 'var'th variable from
 * each level.
 */
void cnFcGetVarProfile(cnTFC * fc, float * data /*OUT*/,
  int profile, int var);


/* Get all the flux data values for one variable in a profile.
 * PRE: fc!=NULL; data!=NULL; 0<=profile<fc->numProfiles;
 * 0<='var'<fc->numFluxVars;
 * 'data' must have space for at least 'fc->numLevels'+1 floats.
 * 'profile' is the index into 'fc->profiles' for the profile to copy
 * from.
 * 'var' is the index into the flux variables for each level.
 * POST: 'data' will contain the data for the 'var'th flux variable from
 * each level.
 */
void cnFcGetFluxProfile(cnTFC * fc, float * data /*OUT*/,
  int profile, int var);


/* Get all the data values for one surface variable from each profile.
 * PRE: fc!=NULL; data!=NULL;
 * 0<='var'<fc->numSurfaceVars;
 * 'var' is the index into the surface variables.
 * POST: 'data' will contain the data for the 'var'th surface variable
 * from each profile.
 */
void cnFcGetSurfaceVar(cnTFC * fc, float * data /*OUT*/, int var);


/************************************************************************/
/************************************************************************/


#endif

