

#include "cnim.h"


/************************************************************************/
/************************************************************************/


/* Read and return an integer global attribute. */
static int cnReadAtt(int infile, char * name)
{
  int retValue;
  NCEXIT(nc_get_att_int(infile, NC_GLOBAL, name, &retValue));
  return retValue;
}


/* Read and return a dimension length.
 * Returns -1 if dimension does not exist.
 */
static int cnReadDim(int infile, char * name)
{
  int    id;
  size_t len;
  int    retValue = -1;

  NCEXIT(nc_inq_dimid(infile, name, &id));
  NCEXIT(nc_inq_dimlen(infile, id, &len));
  retValue = (int) len;

  return retValue;
}


/* Read a variable. */
static void cnReadVarInt(int infile, char * name, int * data /*OUT*/)
{
  int id;

  NCEXIT(nc_inq_varid(infile, name, &id));
  NCEXIT(nc_get_var_int(infile, id, data));
}


/* Read a variable. */
static void cnReadVarFloat(int infile, char * name, float * data /*OUT*/)
{
  int id;

  NCEXIT(nc_inq_varid(infile, name, &id));
  NCEXIT(nc_get_var_float(infile, id, data));
}


void cnImRead(char * filename, cnTIM * im)
{
  int infile;
  int i;
  int nVars;

  NCEXIT(nc_open(filename, 0, &infile));

  /* Read the global attributes. */
  im->ga.site      = cnReadAtt(infile, "site");
  im->ga.institute = cnReadAtt(infile, "institute");
  im->ga.model     = cnReadAtt(infile, "model");
  im->ga.fileType  = cnReadAtt(infile, "file_type");
  im->ga.initTime  = cnReadAtt(infile, "initialization");
  im->ga.utcOffset = cnReadAtt(infile, "utc_offset");

  /* Read the dimensions. */
  im->dims.time   = cnReadDim(infile, "time");
  im->dims.level  = cnReadDim(infile, "level");

  if ( im->dims.time > CN_NC_MAX_TIME_DIM )
  {
    printError(PE_EXIT, "%s: time dimension too big (%d)",
      filename, im->dims.time);
  }

  if ( im->dims.level > CN_NC_MAX_LEVEL_DIM )
  {
    printError(PE_EXIT, "%s: level dimension too big (%d)",
      filename, im->dims.level);
  }


  /* I need to know how many variables there are.
   * There should be "time", "latitude", "longitude".
   * I assume that all the other variables are model parameters.
   */
  NCEXIT(nc_inq_nvars(infile, &nVars));

  /* 'nVars' is the total number of variables in the dataset. */

  /* Read the variables. */
  cnReadVarFloat(infile, "latitude", &im->vars.latitude);
  cnReadVarFloat(infile, "longitude", &im->vars.longitude);
  cnReadVarInt(infile, "time", im->vars.time);

  im->vars.numVars = 0;
  for ( i = 0; i < nVars; i++ )
  {
    /* For each variable in the dataset, get the variable name, check if
     * it is a model parameter variable and read the data.
     */
    static char name[NC_MAX_NAME + 1];
    int         id;   /* one of 'cnEVariableIds' */

    NCEXIT(nc_inq_varname(infile, i, name));
    id = ncSearchNames(name);
    if ( id >= 0 )
    {
      int t;

      im->vars.vars[im->vars.numVars].varId = id;
      for ( t = 0; t < im->dims.time; t++ )
      {
        size_t start[2];
        size_t count[2];

        start[0] = (size_t) t;
        start[1] = 0;
        count[0] = 1;
        count[1] = im->dims.level;
        NCEXIT(nc_get_vara_float(infile, i, start, count,
          im->vars.vars[im->vars.numVars].data[t]));
      }
      im->vars.numVars++;
      if ( im->vars.numVars > CN_NC_MAX_VARS )
      {
        printError(PE_EXIT, "%s: too many variables (%d)",
          filename, im->vars.numVars);
      }
    }
  }


  NCEXIT(nc_close(infile));
}


int cnImSearchVars(cnTIM * im, int varId)
{
  int retValue;
  for ( retValue = im->vars.numVars - 1;
    retValue >= 0 && im->vars.vars[retValue].varId != varId; retValue-- )
  {
  }
  return retValue;
}


/************************************************************************/
/************************************************************************/
