/*
 * NetCDF utilities.
 *
 * HISTORY
 * 2004-02-24 ao started.
 *
 * 2004-05-11  ao  renamed variables in ncVariables, removed the
 *   "var_" and "mwp_" prefixes.
 *   Changed precisions.
 *
 * AUTHORS
 * Andrew Openshaw
 */


#if !defined(CNNETCDF_H)
#define CNNETCDF_H


/************************************************************************/
/************************************************************************/


#include <time.h>
#include <unistd.h>
#include "cnapp.h"
#include "cnmodels.h"
#include "cnsites.h"
#include "cnutils.h"


/************************************************************************/
/************************************************************************/


/* A few limits. */
enum
{
  CN_NC_MAX_TIME_DIM   = 64,
  CN_NC_MAX_LEVEL_DIM  = 96,
  CN_NC_MAX_FREQ_DIM   = 16,
  CN_NC_MAX_LEVEL_VARS = 64,
  CN_NC_MAX_FLUX_VARS  = 64,
  CN_NC_MAX_SFC_VARS   = 64,
  CN_NC_MAX_VARS       = CN_NC_MAX_LEVEL_VARS + CN_NC_MAX_FLUX_VARS
    + CN_NC_MAX_SFC_VARS
};


/************************************************************************/
/************************************************************************/


/* For the 'Conventions' global attribute. */
#define NC_CONVENTIONS "CF-1.0"


/* Check the return value from a netCDF function call.
 * 'fn' is the function call.
 * 'pe' is for 'printError()'.
 * Prints an error message.
 */
#define NCCHK(fn, pe)  ncRetValue = fn; if ( ncRetValue != NC_NOERR )  \
  printError(PE_EXIT, "netCDF error: \"%s\" [file: \"%s\", line: %d]", \
    nc_strerror(ncRetValue), __FILE__, __LINE__)


/* Check the return value from a netCDF function call.
 * 'fn' is the function call.
 * Prints an error message and exits.
 */
#define NCEXIT(fn)     NCCHK(fn, PE_EXIT)


/* Check the return value from a netCDF function call.
 * 'fn' is the function call.
 * Prints an error message but does not exit.
 */
#define NCCONT(fn)     NCCHK(fn, 0)


/************************************************************************/
/************************************************************************/


/* Structure for a netCDF variable. */
typedef struct
{
  char * name;

  /* Number decimal places for each value to be printed to. */
  int    precision;

  /* Attributes: */
  char * units;
  char * longName;
  char * standardName;
  char * comment;
} ncTVariable;


/************************************************************************/
/************************************************************************/


/* Global variable used by 'NCCHK()'. */
extern int ncRetValue;


/* List of netCDF variables.
 * The index into this must be one of 'cnEVariableIds' from 'cnmodels.h'.
 */
extern ncTVariable ncVariables[CN_VARIABLE_ID_COUNT];


/* 'title' global attribute. */
extern char * ncTitles[CN_ML_COUNT];


/* 'source' global attribute. */
extern char * ncSources[CN_ML_COUNT];


/* 'institute' global attribute. */
extern char * ncModelInstitutions[CN_ML_COUNT];


/* The horizontal resolution [km] for each model for each site. */
extern float ncHorizontalResolutions[CN_ML_COUNT][CN_SITE_COUNT];


/************************************************************************/
/************************************************************************/


/*
 * PRE: 'time' is minutes since 1960.
 * 'site' is one of 'cnESites' from 'cnsites.h'.
 * 'model' is one of 'cnEModels' from 'cnmodels.h'.
 * Filename format is:
 *   <model>_<YYYYMMDD>_<HHMM>_<site name>
 *
 * POST: returns a pointer to a static string.
 */
char * ncMakeFilename(int model, int time, int site);


/* Make the 'units' attribute string for the 'time' variable.
 * PRE: 'time' is in seconds from 1960.
 * 'utcTime' is in the format HHMM.
 * POST: returns a pointer to a static string.
 */
char * ncMakeTimeUnits(int time, int utcTime);


/* Make the initialisation time string.
 * PRE: 'time' is in seconds from 1960.
 * 'utcTime' is in the format HHMM.
 * POST: returns a pointer to a static string.
 */
char * ncMakeInitializationTime(int time, int utcTime);


/* Create a string for the "history" global attribute.
 * POST: returns a newly created string for the history attribute.
 * Creates a string with the format:
 *   <current date/time>, <user> on <host>, <progName> <cmdLine>.
 * If 'user==NULL' then the login name is used.
 * <host> is the host computers name.
 * Returns NULL if 'malloc()' failed.
 */
char * ncMakeHistoryLine(char * user, int argc, char * argv[]);


/* Search 'ncVariables' for 'name'.
 * POST: returns -1 if 'name' not found otherwise returns the index
 * into 'ncVariables' (which is one of 'cnEVariables' from "cnmodels.h").
 */
int ncSearchNames(char * name);


/************************************************************************/
/************************************************************************/


#endif
