/*
 * CloudNET Model Conversion Software
 *
 * Convert ECMWF surface file to netCDF/CDL.
 *
 * HISTORY
 * 2004-02-26 ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnnetcdf.h"
#include "cnecifs.h"
#include "version.h"
#include "cdl.c"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS      = 1,

  /* maximum line length for a line read from the file. */
  MAX_LINE_LEN = 4096,

  /* Maximum number of lines that can be read from the file. */
  MAX_LINES    = 16384
};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{
  int utcOffset;

  /* Index into 'argv[]' of the first argument after the options.
   * Or -1 if 'MINARGS=0'.
   */
  int firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Convert ECMWF-IFS model output ASCII files with surface variables"
    " (i.e. 'sfc', 'sfc2' and 'sfce' files) to netCDF-CDL.\n"
    "  Outputs CDL text to stdout. This can be piped through 'ncgen -b'"
    " to produce a netCDF dataset.\n"
  );
  fprintf(outfile,
    "  The output CDL is given a name with the format:\n"
    "    ECIFS_SFC_yyyymmdd_hhmm_<site>\n"
    "  Where \"yyyymmdd\" and \"hhmm\" is the initialization date and"
    " time from the input file. <site> will be the first two characters"
    " of the site the data is for.\n"
    "  This name can be overridden when running 'ncgen' by using the '-o'"
    " option with 'ncgen'.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
    "\n"
    "  -u minutes\n"
    "    UTC offset for all times in the file in minutes "
    "(can be negative). Default: 0\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file' contains the concatenated surface files from the ECMWF-IFS"
    " model.\n"
    "  The 'cnecifscatsfc' program should be used to concatenate the"
    " original files.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;

      case 'u':
        cl.utcOffset = atoi(getOptArg(option, argc, argv, &argn));
	break;

      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = ( MINARGS > 0 ) ? argn : -1;
}


/************************************************************************/
/************************************************************************/


void processFile(char * filename, FILE * infile, int utcOffset)
{
  const int              INSTITUTE               = CN_MI_EC;
  const int              MODEL                   = CN_ML_ECIFS;
  const int              FILETYPE                = CN_FT_SFC;

  /* 'data' contains each line from the file. */
  static cnTECIFSsfcLine data[MAX_LINES];
  static cnTECIFSheader  hdr;
  static char            line[MAX_LINE_LEN + 1];

  int                    numLines;
  int                    numProfiles;
  int                    prof;

  /**********************************************************************/

  cnECIFSGetHeader(filename, infile, line, MAX_LINE_LEN, &hdr);

  /**********************************************************************/

  /* Read each line from the file. */
  numLines    = 0;
  numProfiles = 0;
  while ( fgets(line, MAX_LINE_LEN, infile) != NULL )
  {
    if ( numLines >= MAX_LINES )
    {
      printError(PE_EXIT,
        "%s: the number of lines exceeds %d which is the preset limit.\n"
        "You'll need to edit the program's source code and increase the "
        "value MAX_LINES", filename);
    }

    cnECIFSGetSfcLine(line, &data[numLines]);

    if ( data[numLines].pressure != data[numLines].pressure2 )
    {
      printError(PE_EXIT,
        "%s: the pressures are not equal",
        filename);
    }

    numLines++;
    numProfiles++;
  }

  /**********************************************************************/

  cdlPrint(stdout, INSTITUTE, MODEL, FILETYPE,
    data[0].initTime, data[0].verTime,
    utcOffset, hdr.site, hdr.lat, hdr.lon,
    numProfiles, 1,
    CN_SFC_PRESSURE,
    CN_SFC_NET_SW,
    CN_SFC_NET_LW,
    CN_SFC_DOWN_SW,
    CN_SFC_DOWN_LW,
    CN_SFC_CS_DOWN_SW,
    CN_SFC_CS_DOWN_LW,
    CN_SFC_DOWN_LAT_HEAT_FLX,
    CN_SFC_DOWN_SENS_HEAT_FLX,
    CN_SFC_LS_RAIN,
    CN_SFC_CONV_RAIN,
    CN_SFC_LS_SNOW,
    CN_SFC_CONV_SNOW,
    CN_SFC_LS_PRECIP_FRACTION,
    CN_SFC_TOTAL_CLOUD_FRACTION,
    CN_SFC_BL_HEIGHT,
    CN_SFC_ALBEDO,                /* check units. */
    CN_SFC_TEMP_2M,
    CN_SFC_Q_2M,                  /* check units. */
    CN_SFC_ROUGH_MOM,
    CN_SFC_ROUGH_HEAT,
    CN_SFC_SKIN_TEMP,
    CN_SFC_WIND_U_10M,
    CN_SFC_WIND_V_10M,
    CN_SFC_GEOPOTENTIAL,
    -1
  );

  printf("time = ");
  for ( prof = 0; prof < numProfiles; prof++ )
  {
    if ( prof != 0 ) printf(",\n");

    printf("%d", data[prof].verTime - data[prof].initTime);
  }
  printf(";\n");


#define PRINTVAR(varId, varName)                                         \
  cdlPrintSingleVar(stdout, varId, data, sizeof(cnTECIFSsfcLine),        \
    cnOffsetOf(cnTECIFSsfcLine, varName), numProfiles)


  PRINTVAR(CN_SFC_PRESSURE, pressure);
  PRINTVAR(CN_SFC_NET_SW, swRad);
  PRINTVAR(CN_SFC_NET_LW, lwRad);
  PRINTVAR(CN_SFC_DOWN_SW, downSw);
  PRINTVAR(CN_SFC_DOWN_LW, downLw);
  PRINTVAR(CN_SFC_CS_DOWN_SW, clearSwRad);
  PRINTVAR(CN_SFC_CS_DOWN_LW, clearLwRad);
  PRINTVAR(CN_SFC_DOWN_LAT_HEAT_FLX, latHeatFlx);
  PRINTVAR(CN_SFC_DOWN_SENS_HEAT_FLX, sensHeatFlx);
  PRINTVAR(CN_SFC_LS_RAIN, largeRain);
  PRINTVAR(CN_SFC_CONV_RAIN, convRain);
  PRINTVAR(CN_SFC_LS_SNOW, largeSnow);
  PRINTVAR(CN_SFC_CONV_SNOW, convSnow);
  PRINTVAR(CN_SFC_LS_PRECIP_FRACTION, largeRainFrac);
  PRINTVAR(CN_SFC_TOTAL_CLOUD_FRACTION, totCloud);
  PRINTVAR(CN_SFC_BL_HEIGHT, pblHeight);
  PRINTVAR(CN_SFC_ALBEDO, albedo);
  PRINTVAR(CN_SFC_TEMP_2M, temp2m);
  PRINTVAR(CN_SFC_Q_2M, q2m);
  PRINTVAR(CN_SFC_ROUGH_MOM, roughMom);
  PRINTVAR(CN_SFC_ROUGH_HEAT, roughHeat);
  PRINTVAR(CN_SFC_SKIN_TEMP, skinTemp);
  PRINTVAR(CN_SFC_WIND_U_10M, uWind10m);
  PRINTVAR(CN_SFC_WIND_V_10M, vWind10m);
  PRINTVAR(CN_SFC_GEOPOTENTIAL, gp);

  printf("\n}\n");
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  char * filename;
  FILE * infile;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */
  cl.utcOffset = 0;
  cl.firstArg  = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);

  /**********************************************************************/

  filename = argv[cl.firstArg];
  infile   = fopen(filename, "r");
  if ( infile == NULL )
  {
    printError(PE_EXIT, "unable to open file: %s\n", filename);
  }

  processFile(filename, infile, cl.utcOffset);
  fclose(infile);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

