/*
 * CloudNET Model Conversion Software
 * Determine the model, file type, site, initialization and
 * first forecast times of a file.
 *
 * HISTORY
 * 2004-04-07 ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnapp.h"
#include "cnmodels.h"
#include "cnsites.h"
#include "cnutils.h"
#include "version.h"

#include "getfileinfo.c"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS = 1

};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{
  /* Index into 'argv[]' of the first argument after the options. */
  int firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Determine the model, file type, site, initialization and first"
    " forecast times of a model output file and print one line on"
    " standard output.\n"
    "  The output will consist of \"model file_type site initialization"
    " first_forecast\"\n"
  );
  fprintf(outfile,
    "  For the model, prints one of:\n"
    "    \"MOUMM\"   UK Met Office Unified Model mesoscale.\n"
    "    \"MOUMG\"   UK Met Office Unified Model mesoscale.\n"
    "       (NOTE: in this case the file contains both surface and level"
    " parameters so file type is undetermined.\n"
    "    \"ECIFS\"   ECMWF IFS.\n"
    "    \"KNRAC\"   KNMI RACMO.\n"
    "    \"MFARP\"   Meteo-France ARPEGE.\n"
    "    \"-----\"   Unknown model.\n"
  );
  fprintf(outfile, 
    "\n"
    "  For the file type, prints one of:\n"
    "    \"VAR\"     multi-level variables.\n"
    "    \"FLX\"     multi-level flux variables.\n"
    "    \"SFC\"     surface variables.\n"
    "    \"SF2\"     ECMWF sfc2 file.\n"
    "    \"SFE\"     ECMWF sfce file.\n"
    "    \"---\"     Unknown file type.\n"
  );
  fprintf(outfile,
    "\n"
    "  For the site, prints one of:\n"
    "    \"CH\"      Chilbolton.\n"
    "    \"CA\"      Cabauw.\n"
    "    \"PA\"      Palaiseau.\n"
    "    \"AB\"      Aberystwyth.\n"
    "    \"--\"      Unknown site.\n"
  );
  fprintf(outfile,
    "\n"
    "  For the initialization time and forecast time, prints"
    " \"YYYYMMDDHHMM\" or \"------------\" if a time could not be"
    " determined.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file' is a model output file.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;

      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = argn;
}


/************************************************************************/
/************************************************************************/


void processFile(char * filename)
{
  static char * models[CN_ML_COUNT] = { "MOUMM", "MOUMG", "ECIFS",
    "MFARP", "KNHIR", "KNRAC" };
  static char * fileTypes[CN_FT_COUNT + 2] = { "VAR", "FLX", "SFC",
    "SF2", "SFE" };
  static char * sites[CN_SITE_COUNT] = { "CH", "CA", "PA", "AB" };

  cnTFileInfo info;

  getFileInfo(filename, &info);

  if ( info.model < 0 )
  {
    printf("----- ");
  }
  else
  {
    printf("%s ", models[info.model]);
  }

  if ( info.fileType < 0 )
  {
    printf("--- ");
  }
  else
  {
    printf("%s ", fileTypes[info.fileType]);
  }

  if ( info.site < 0 )
  {
    printf("-- ");
  }
  else
  {
    printf("%s ", sites[info.site]);
  }

  if ( info.initTime < 0 )
  {
    printf("------------ ");
  }
  else
  {
    printf("%08d%04d ",
      utDateFromMins(info.initTime), utTimeFromMins(info.initTime));
  }

  if ( info.forecastTime < 0 )
  {
    printf("------------ ");
  }
  else
  {
    printf("%08d%04d ",
      utDateFromMins(info.forecastTime),
      utTimeFromMins(info.forecastTime));
  }

  printf("\n");
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */

  cl.firstArg = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);
  processFile(argv[cl.firstArg]);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

