/*
 * CloudNET Model Conversion Software
 *
 * Filter for RACMO files.
 * RACMO files have 4 files concatenated into one file. The 4 files
 * represent the 4 grid points around the site.
 *
 * HISTORY
 * 2003-10-13 ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <math.h>
#include "version.h"
#include "cnapp.c"
#include "cnsites.c"
#include "cnutils.c"


/************************************************************************/
/************************************************************************/


enum
{
  /* Minimum number of arguments on the command line. */
  MINARGS                = 2,

  MAX_LINE_LEN           = 4096,

  MAX_NUM_HDR_LINES      = 32,

  MIN_NUM_MAIN_HDR_LINES = 9
};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s infile outfile\n"
    "\n",
    PROGNAME
  );
  fprintf(outfile,
    "DESCRIPTION\n"
    "\n"
    "  The RACMO files consist of the output for 4 grid points"
    " concatenated into one file.\n"
    "  This program will produce a file containing the output for only"
    " one grid point which is the point closest to the site.\n"
    "  It also only outputs profiles that are on the hour.\n"
    "\n"
  );
  fprintf(outfile,
    "ARGUMENTS\n"
    "\n"
    "  infile is a text file with the models' output for four grid "
    "points.\n"
    "  outfile will contain one grid point.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/************************************************************************/
/************************************************************************/


/*
 * PRE: gridCoords must be an array of 4 elements.
 * site must be in the range 0..2.
 * POST: returns the index into gridCoords for the coords closest to
 * the site coords.
 */
int getClosestGridPoint(cnTSiteCoords * gridCoords, int site)
{
  const double RADIUS_EARTH = 6370.0;

  double       pi           = 4.0 * atan(1.0);
  double       siteX;
  double       siteY;
  double       x;
  double       y;
  int          retValue     = 0;
  int          i;
  double       d;
  double       minD         = 100000.0;

  /**********************************************************************/

  /* I assume the 4 grid points are close enough to be approximated by
   * a rectangular grid and we can ignore the divergence/convergence of
   * longitude lines with changing latitude.
   */

  siteX = RADIUS_EARTH * sin(cnSiteCoords[site].lon * pi / 180.0);
  siteY = RADIUS_EARTH * sin(cnSiteCoords[site].lat * pi / 180.0);

  for ( i = 0; i < 4; i++ )
  {
    x = RADIUS_EARTH * sin(gridCoords[i].lon * pi / 180.0) - siteX;
    y = RADIUS_EARTH * sin(gridCoords[i].lat * pi / 180.0) - siteY;
    d = sqrt(x * x + y * y);

    if ( d < minD )
    {
      minD     = d;
      retValue = i;
    }
  }

  return retValue;
}


/* Read a line from 'infile' into 's'.
 * Returns 0 on success, 1 on EOF.
 */
int readLine(char * infilename, FILE * infile, char * s)
{
  int retValue = 0;

  if ( fgets(s, MAX_LINE_LEN, infile) == NULL )
  {
    if ( feof(infile) != 0 )
    {
      retValue = 1;
    }
    else
    {
      printError(PE_EXIT,
        "error reading from file \"%s\"", infilename);
    }
  }

  return retValue;
}


void writeLine(char * outfilename, FILE * outfile, char * s)
{
  if ( fputs(s, outfile) == EOF )
  {
    printError(PE_EXIT,
      "error writing to file \"%s\"", outfilename);
  }
}


void processFile(char * infilename, char * outfilename)
{
  static char   line[MAX_LINE_LEN + 1];
  static char   mainHdr[MAX_NUM_HDR_LINES][MAX_LINE_LEN + 1];

  FILE *        infile;
  FILE *        outfile;
  int           numMainHdrLines = 0;
  cnTSiteCoords gridCoords[4];
  int           site;
  int           i;
  int           closest;

  /**********************************************************************/

  infile = fopen(infilename, "r");
  if ( infile == NULL )
  {
    printError(PE_EXIT, "unable to open file \"%s\"", infilename);
  }


  /* Read the main header. */
  while ( readLine(infilename, infile, mainHdr[numMainHdrLines]) == 0
    && mainHdr[numMainHdrLines][0] == '#' )
  {
    numMainHdrLines++;
    if ( numMainHdrLines > MAX_NUM_HDR_LINES )
    {
      printError(PE_EXIT,
        "too many header lines in file \"%s\"", infilename);
    }
  }

  if ( numMainHdrLines < MIN_NUM_MAIN_HDR_LINES )
  {
    printError(PE_EXIT,
      "expected a minimum of %d header lines in file \"%s\"",
      MIN_NUM_MAIN_HDR_LINES, infilename);
  }


  sscanf(&mainHdr[6][1], "%f%f",
    &gridCoords[0].lon, &gridCoords[0].lat);

  if ( strstr(&mainHdr[5][1], "CH") != NULL )
  {
    site = CN_SITE_CH;
  }
  else
  if ( strstr(&mainHdr[5][1], "CA") != NULL )
  {
    site = CN_SITE_CA;
  }
  else
  if ( strstr(&mainHdr[5][1], "PA") != NULL )
  {
    site = CN_SITE_PA;
  }
  else
  {
    printError(PE_EXIT, "unknown site");
  }


  /* Search for the next 3 headers. */
  i = 1;
  while ( i < 4 && readLine(infilename, infile, line) == 0 )
  {
    if ( line[0] == '#' )
    {
      /* The first line of these headers should be lon/lat. */
      sscanf(&line[1], "%f%f", &gridCoords[i].lon, &gridCoords[i].lat);

      /* Skip the rest of the header. */
      while ( readLine(infilename, infile, line) == 0
        && line[0] == '#' )
      {
      }

      i++;
    }
  }


  /* From the lon/lats decide which grid-point is closest to the site. */
  closest = getClosestGridPoint(gridCoords, site);


  rewind(infile);

  outfile = fopen(outfilename, "w");
  if ( outfile == NULL )
  {
    printError(PE_EXIT, "unable to open file \"%s\"", outfilename);
  }


  /* Copy the first 6 lines of the main header. */
  for ( i = 0; i < 6; i++ )
  {
    readLine(infilename, infile, line);
    writeLine(outfilename, outfile, line);
  }


  /* Now we need to move to the header of the part that we need to
   * copy.
   */
  if ( closest > 0 )
  {
    for ( i = 0; i < closest; i++ )
    {
      /* Skip the rest of the header. */
      while ( readLine(infilename, infile, line) == 0
        && line[0] == '#' )
      {
      }
 
      /* Search for the next header. */
      while ( readLine(infilename, infile, line) == 0
        && line[0] != '#' )
      {
      }

      /* At this point 'line' contains the first line of the next
       * header.
       */
    }
    writeLine(outfilename, outfile, line);
  }


  /* Copy the rest of the header. */
  while ( readLine(infilename, infile, line) == 0 && line[0] == '#' )
  {
    writeLine(outfilename, outfile, line);
  }

  /* Copy all lines upto the EOF or the next header. */
  do
  {
    int date;
    int time;

    /* for the moment only copy profiles on the hour. */
    sscanf(line, "%d%d", &date, &time);
    if ( (time % 100) == 0 )
    {
      writeLine(outfilename, outfile, line);
    }
  } while ( readLine(infilename, infile, line) == 0 && line[0] != '#' );

  fclose(infile);
  fclose(outfile);
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  char * infilename;
  char * outfilename;
  int    argn         = 1;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read the arguments. */
  infilename = argv[argn];
  argn++;
  outfilename = argv[argn];
  processFile(infilename, outfilename);

  return 0;
}


/************************************************************************/
/************************************************************************/
