/*
 * CloudNET Model Conversion Software
 *
 * Convert RACMO single level file to netCDF/CDL.
 *
 * HISTORY
 * 2004-02-27  ao
 *
 * 2004-06-03  ao  changed large scale cloud fraction to total cloud
 *   fraction. Removed convective rain flux.
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnnetcdf.h"
#include "cnknrac.h"
#include "version.h"
#include "cdl.c"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS      = 1,

  /* maximum line length for a line read from the file. */
  MAX_LINE_LEN = 4096,

  /* Maximum number of lines that can be read from the file. */
  MAX_LINES    = 16384
};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{
  int utcOffset;

  /* Index into 'argv[]' of the first argument after the options.
   * Or -1 if 'MINARGS=0'.
   */
  int firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Convert KNMI-RACMO model output ASCII files with surface variables"
    " (i.e. 'sfc' files) to netCDF-CDL.\n"
    "  Outputs CDL text to stdout. This can be piped through 'ncgen -b'"
    " to produce a netCDF dataset.\n"
  );
  fprintf(outfile,
    "  The output CDL is given a name with the format:\n"
    "    KNRAC_SFC_yyyymmdd_hhmm_<site>\n"
    "  Where \"yyyymmdd\" and \"hhmm\" is the initialization date and"
    " time from the input file. <site> will be the first two characters"
    " of the site the data is for.\n"
    "  This name can be overridden when running 'ncgen' by using the '-o'"
    " option with 'ncgen'.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
    "\n"
    "  -u minutes\n"
    "    UTC offset for all times in the file in minutes "
    "(can be negative). Default: 0\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file' contains the concatenated surface files from the KNMI-RACMO"
    " model.\n"
    "  This must be a file produced by the 'cnracmof' program.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;

      case 'u':
        cl.utcOffset = atoi(getOptArg(option, argc, argv, &argn));
	break;

      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = ( MINARGS > 0 ) ? argn : -1;
}


/************************************************************************/
/************************************************************************/


void processFile(char * filename, FILE * infile, int utcOffset)
{
  /*const int              INSTITUTE               = CN_MI_KN;*/
  /*const int              MODEL                   = CN_ML_KNRAC;*/
  const int              FILETYPE                = CN_FT_SFC;

  /* 'data' contains each line from the file. */
  static cnTKNRACsfcLine data[MAX_LINES];
  static cnTKNRACheader  hdr;
  static char            line[MAX_LINE_LEN + 1];

  int                    numLines;
  int                    numProfiles;
  int                    prof;


  /**********************************************************************/

  cnKNRACGetHeader(filename, infile, line, MAX_LINE_LEN, &hdr);

  /**********************************************************************/

  /* Read each line from the file. */
  numLines    = 0;
  numProfiles = 0;
  while ( fgets(line, MAX_LINE_LEN, infile) != NULL )
  {
    if ( numLines >= MAX_LINES )
    {
      printError(PE_EXIT,
        "%s: the number of lines exceeds %d which is the preset limit.\n"
        "You'll need to edit the program's source code and increase the "
        "value MAX_LINES", filename);
    }

    cnKNRACGetSfcLine(line, &data[numLines]);

    numLines++;
    numProfiles++;
  }

  /**********************************************************************/

  if ( hdr.model == CN_ML_KNRAC )
  {
    cdlPrint(stdout, hdr.institute, hdr.model, FILETYPE,
	     hdr.initTime, data[0].verTime, utcOffset, hdr.site, hdr.lat, hdr.lon,
	     numProfiles, 1,
	     CN_SFC_PRESSURE,
	     CN_SFC_DOWN_LAT_HEAT_FLX,
	     CN_SFC_DOWN_SENS_HEAT_FLX,
	     CN_SFC_DOWN_SW,
	     CN_SFC_DOWN_LW,
	     CN_SFC_CONV_RAIN,
	     CN_SFC_LS_RAIN,
	     CN_SFC_SKIN_TEMP,
	     CN_SFC_LS_PRECIP_FRACTION,
	     CN_SFC_TOTAL_CLOUD_FRACTION,
	     -1
	     );    
  }
  else
  if ( hdr.model == CN_ML_SMRCA )
  {
    cdlPrint(stdout, hdr.institute, hdr.model, FILETYPE,
	     hdr.initTime, data[0].verTime, utcOffset, hdr.site, hdr.lat, hdr.lon,
	     numProfiles, 1,
	     CN_SFC_PRESSURE,
	     CN_SFC_DOWN_LAT_HEAT_FLX,
	     CN_SFC_DOWN_SENS_HEAT_FLX,
	     CN_SFC_DOWN_SW,
	     CN_SFC_DOWN_LW,
	     CN_SFC_CONV_RAIN,
	     CN_SFC_LS_RAIN,
	     CN_SFC_SKIN_TEMP,
	     CN_SFC_TOTAL_CLOUD_FRACTION,
	     -1
	     );  
  }


  printf("time = ");
  for ( prof = 0; prof < numProfiles; prof++ )
  {
    if ( prof != 0 ) printf(",\n");

    printf("%d", data[prof].verTime - hdr.initTime);
  }
  printf(";\n");


#define PRINTVAR(varId, varName)                                         \
  cdlPrintSingleVar(stdout, varId, data, sizeof(cnTKNRACsfcLine),        \
    cnOffsetOf(cnTKNRACsfcLine, varName), numProfiles)


  PRINTVAR(CN_SFC_PRESSURE, pressure);
  PRINTVAR(CN_SFC_DOWN_LAT_HEAT_FLX, latHeatFlx);
  PRINTVAR(CN_SFC_DOWN_SENS_HEAT_FLX, sensHeatFlx);
  PRINTVAR(CN_SFC_DOWN_SW, downSwFlx);
  PRINTVAR(CN_SFC_DOWN_LW, downLwFlx);
  PRINTVAR(CN_SFC_CONV_RAIN, convRain);
  PRINTVAR(CN_SFC_LS_RAIN, largeRain); 
  if ( hdr.model == CN_ML_KNRAC )
    {
      PRINTVAR(CN_SFC_SKIN_TEMP, tSkin);
      PRINTVAR(CN_SFC_LS_PRECIP_FRACTION, precipFrac);
      PRINTVAR(CN_SFC_TOTAL_CLOUD_FRACTION, totCloud);
  }
  else
  if ( hdr.model == CN_ML_SMRCA )
  {
      PRINTVAR(CN_SFC_SKIN_TEMP, precipFrac);
      PRINTVAR(CN_SFC_TOTAL_CLOUD_FRACTION, totCloud);
  }
  printf("\n}\n");
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  char * filename;
  FILE * infile;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */
  cl.utcOffset = 0;
  cl.firstArg  = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);

  /**********************************************************************/

  filename = argv[cl.firstArg];
  infile   = fopen(filename, "r");
  if ( infile == NULL )
  {
    printError(PE_EXIT, "unable to open file: %s\n", filename);
  }

  processFile(filename, infile, cl.utcOffset);
  fclose(infile);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

