/*
 * CloudNET Model Conversion Software
 *
 * Convert UKMO global MOLTS file and convert surface variables to
 * netCDF/CDL.
 *
 * HISTORY
 * 2004-03-04 ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnnetcdf.h"
#include "cnmoumg.h"
#include "version.h"
#include "cdl.c"
#include "gribcodes.h"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS      = 1,

  /* maximum line length for a line read from the file. */
  MAX_LINE_LEN = 4096
};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{
  int utcOffset;

  /* Index into 'argv[]' of the first argument after the options.
   * Or -1 if 'MINARGS=0'.
   */
  int firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Reads a UK Met Office MOLTS global output ASCII file and writes"
    " out the surface variables in netCDF-CDL.\n"
    "  Outputs CDL text to stdout. This can be piped through 'ncgen -b'"
    " to produce a netCDF dataset.\n"
  );
  fprintf(outfile,
    "  The output CDL is given a name with the format:\n"
    "    MOUMG_SFC_yyyymmdd_hhmm_<site>\n"
    "  Where \"yyyymmdd\" and \"hhmm\" is the initialization date and"
    " time from the input file. <site> will be the first two characters"
    " of the site the data is for.\n"
    "  This name can be overridden when running 'ncgen' by using the '-o'"
    " option with 'ncgen'.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
    "\n"
    "  -u minutes\n"
    "    UTC offset for all times in the file in minutes "
    "(can be negative). Default: 0\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file' is the MOLTS ASCII output file from the UKMO UM.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;

      case 'u':
        cl.utcOffset = atoi(getOptArg(option, argc, argv, &argn));
	break;

      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = ( MINARGS > 0 ) ? argn : -1;
}


/************************************************************************/
/************************************************************************/


void processFile(char * filename, FILE * infile, int utcOffset)
{
  const int           INSTITUTE               = CN_MI_MO;
  const int           MODEL                   = CN_ML_MOUMG;
  const int           FILETYPE                = CN_FT_SFC;

  /* Time interval between each profile in minutes. */
  const int           TIME_STEP               = 3 * 60;

  static cnTMOUMGfile data;
  static char         line[MAX_LINE_LEN + 1];

  int                 prof;
  int                 i;

  /**********************************************************************/

  cnMOUMGreadFile(filename, infile, line, MAX_LINE_LEN, &data);

  /**********************************************************************/

  cdlPrint(stdout, INSTITUTE, MODEL, FILETYPE,
    data.header.initTime, data.header.initTime,
    utcOffset, data.header.site,
    data.header.lat, data.header.lon,
    data.header.numTimes, 1,
    CN_SFC_PRESSURE,
    CN_SFC_TEMPERATURE,
    CN_SFC_DOWN_SW,
    CN_SFC_DOWN_LW,
    CN_SFC_DOWN_LAT_HEAT_FLX,
    CN_SFC_DOWN_SENS_HEAT_FLX,
    CN_SFC_LS_RAIN,
    CN_SFC_CONV_RAIN,
    CN_SFC_LS_SNOW,
    CN_SFC_CONV_SNOW,
    CN_SFC_LS_CLOUD_FRACTION,
    CN_SFC_BL_HEIGHT,
    CN_SFC_TEMP_2M,
    CN_SFC_Q_2M,
    CN_SFC_WIND_U_10M,
    CN_SFC_WIND_V_10M,
    -1
  );

  printf("time = ");
  for ( prof = 0; prof < data.header.numTimes; prof++ )
  {
    if ( prof != 0 ) printf(",\n");

    printf("%d", prof * TIME_STEP);
  }
  printf(";\n");


#define PRINTVAR(varId, gribCode)                                        \
  i = cnMOUMGsearchSingle(&data, gribCode);                              \
  if ( i >= 0 ) {                                                        \
    printf("%s = ", ncVariables[varId].name);                            \
    for ( prof = 0; prof < data.header.numTimes; prof++ ) {              \
      float f = data.singleLevelVars[i].data[prof];                      \
      if ( prof != 0 ) printf(",\n");                                    \
      cdlPrintValue(stdout, varId, f);                                   \
    }                                                                    \
    printf(";\n");                                                       \
  }


  PRINTVAR(CN_SFC_PRESSURE, GRIB_SFC_PRESSURE);
  PRINTVAR(CN_SFC_TEMPERATURE, GRIB_SFC_TEMPERATURE);
  PRINTVAR(CN_SFC_DOWN_SW, GRIB_SFC_TOT_DOWN_SW);
  PRINTVAR(CN_SFC_DOWN_LW, GRIB_SFC_TOT_DOWN_LW);
  PRINTVAR(CN_SFC_DOWN_LAT_HEAT_FLX, GRIB_NET_LAT_HEAT_FLX);
  PRINTVAR(CN_SFC_DOWN_SENS_HEAT_FLX, GRIB_NET_SENS_HEAT_FLX);
  PRINTVAR(CN_SFC_LS_RAIN, GRIB_LARGE_PRECIP);
  PRINTVAR(CN_SFC_CONV_RAIN, GRIB_CONV_PRECIP);
  PRINTVAR(CN_SFC_LS_SNOW, GRIB_LARGE_SNOW);
  PRINTVAR(CN_SFC_CONV_SNOW, GRIB_CONV_SNOW);
  PRINTVAR(CN_SFC_LS_CLOUD_FRACTION, GRIB_TOT_CLOUD_COVER);
  PRINTVAR(CN_SFC_BL_HEIGHT, GRIB_PBL_HEIGHT);
  PRINTVAR(CN_SFC_TEMP_2M, GRIB_TEMP_1_5M);
  PRINTVAR(CN_SFC_Q_2M, GRIB_SPEC_HUMIDITY_1_5M);
  PRINTVAR(CN_SFC_WIND_U_10M, GRIB_WIND_U_10M);
  PRINTVAR(CN_SFC_WIND_V_10M, GRIB_WIND_V_10M);

  printf("\n}\n");
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  /* 'filename' is the filename of the file to process and 'infile' is
   * 'filename' opened file pointer.
   */
  char * filename;
  FILE * infile;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */
  cl.utcOffset = 0;
  cl.firstArg  = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);

  /**********************************************************************/

  filename = argv[cl.firstArg];
  infile   = fopen(filename, "r");
  if ( infile == NULL )
  {
    printError(PE_EXIT, "unable to open file: %s\n", filename);
  }

  processFile(filename, infile, cl.utcOffset);
  fclose(infile);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

