/*
 * CloudNET Model Conversion Software
 *
 * Convert UKMO global MOLTS file and convert surface variables to
 * netCDF/CDL.
 *
 * HISTORY
 * 2004-03-04 ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnnetcdf.h"
#include "cnfc.h"
#include "cnmoumg.h"
#include "version.h"
#include "cdl.c"
#include "gribcodes.h"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS      = 1,

  /* maximum line length for a line read from the file. */
  MAX_LINE_LEN = 4096
};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{
  int utcOffset;

  /* Index into 'argv[]' of the first argument after the options.
   * Or -1 if 'MINARGS=0'.
   */
  int firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Reads a UK Met Office MOLTS global output ASCII file and writes"
    " out the level variables in netCDF-CDL.\n"
    "  Outputs CDL text to stdout. This can be piped through 'ncgen -b'"
    " to produce a netCDF dataset.\n"
  );
  fprintf(outfile,
    "  The output CDL is given a name with the format:\n"
    "    MOUMG_VAR_yyyymmdd_hhmm_<site>\n"
    "  Where \"yyyymmdd\" and \"hhmm\" is the initialization date and"
    " time from the input file. <site> will be the first two characters"
    " of the site the data is for.\n"
    "  This name can be overridden when running 'ncgen' by using the '-o'"
    " option with 'ncgen'.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
    "\n"
    "  -u minutes\n"
    "    UTC offset for all times in the file in minutes "
    "(can be negative). Default: 0\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file' is the MOLTS ASCII output file from the UKMO UM.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;

      case 'u':
        cl.utcOffset = atoi(getOptArg(option, argc, argv, &argn));
	break;

      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = ( MINARGS > 0 ) ? argn : -1;
}


/************************************************************************/
/************************************************************************/


typedef struct
{
  float h;

  /* On theta levels. */
  float p;
  float t;
  float q;
  float ql;
  float qi;
  float w;
  float lscf;

  /* Interpolated from the rho levels. */
  float u;
  float v;
  float ccf;
} TVars;


void processFile(char * filename, FILE * infile, int utcOffset)
{
#define SEARCH(i, grib, varname)                                         \
  i = cnMOUMGsearchModel(&data, grib);                                   \
  if ( i < 0 ) printError(PE_EXIT, "%s: could not find the %s variable", \
      filename, varname)


  const int           INSTITUTE               = CN_MI_MO;
  const int           MODEL                   = CN_ML_MOUMG;
  const int           FILETYPE                = CN_FT_VAR;

  /* Time interval between each profile in minutes. */
  const int           TIME_STEP               = 3 * 60;

  static cnTMOUMGfile data;
  static char         line[MAX_LINE_LEN + 1];
  static TVars        vars[CN_NC_MAX_TIME_DIM * CN_NC_MAX_LEVEL_DIM];

  int                 numProfiles;
  int                 numLevels;
  int                 prof;
  int                 lev;
  int                 i;
  int                 ti;
  int                 pi;
  int                 qi;
  int                 qli;
  int                 qii;
  int                 ui;
  int                 vi;
  int                 wi;
  int                 lscfi;
  int                 ccfi;

  /**********************************************************************/

  cnMOUMGreadFile(filename, infile, line, MAX_LINE_LEN, &data);

  /**********************************************************************/

  numProfiles = data.header.numTimes;
  numLevels   = data.header.numModelLevels;

  /*
   * The UM global data level variables are available in two sections
   * of the MOLTS files: pressure levels and model levels. The model
   * levels are available on rho levels and theta levels.
   *
   * The model levels have pressure on both rho and theta levels.
   * All the other variables are available on one or the other but not
   * both.
   * The majority of the other variables are available on theta levels
   * so I choose the theta levels for copying.
   * The variables on rho levels are: zonal wind, meridional wind, and
   * convective cloud fraction.
   * These will have to be interpolated onto the theta levels.
   * The variables on the theta levels are: pressure, temperature,
   * specific humidity, large scale cloud fraction, vertical velocity,
   * cloud water content, and cloud ice content.
   */

  /*
   * First problem, how do I know which pressure is which?
   * There are two pressures 'data.modelLevelVars', one at rho levels
   * and one at theta levels, but they both have the same GRIB code.
   * I get the temperature first, this is on theta levels, I then find
   * the pressure with the same heights as the temperature.
   */

  /* 'ti' is the index for temperature. */
  SEARCH(ti, GRIB_TEMPERATURE, "temperature");

  /* Get the heights. */
  for ( i = 0, prof = 0; prof < numProfiles; prof++ )
  {
    for ( lev = 0; lev < numLevels; lev++, i++ )
    {
      vars[i].h = data.modelLevelVars[ti].levels[lev].height;
    }
  }

  /* Now find the pressure on theta levels. */
  pi = -1;
  for ( i = 0; i < data.header.numModelVars; i++ )
  {
    if ( data.modelLevelVars[i].grib == GRIB_PRESSURE
      && data.modelLevelVars[i].levels[0].height == vars[0].h )
    {
      pi = i;
      break;
    }
  }

  if ( pi < 0 )
  {
    printError(PE_EXIT,
      "%s: could not find the pressure variable on theta levels",
      filename);
  }

  SEARCH(qi, GRIB_SPEC_HUMIDITY, "specific humidity");
  SEARCH(qli, GRIB_CLOUD_WATER, "cloud water content");
  SEARCH(qii, GRIB_CLOUD_ICE_CONTENT, "cloud ice content");
  SEARCH(ui, GRIB_WIND_U, "zonal wind");
  SEARCH(vi, GRIB_WIND_V, "meridional wind");
  SEARCH(wi, GRIB_GEOMETRIC_VERTICAL_VEL, "vertical wind");
  SEARCH(lscfi, GRIB_LARGE_CLOUD_FRAC, "large scale cloud fraction");
  SEARCH(ccfi, GRIB_CONV_CLOUD_FRAC, "convective cloud fraction");


  /* Now get all the variables on theta levels. */

  for ( i = 0, prof = 0; prof < numProfiles; prof++ )
  {
    for ( lev = 0; lev < numLevels; lev++, i++ )
    {
      vars[i].t    = data.modelLevelVars[ti].levels[lev].data[prof];
      vars[i].p    = data.modelLevelVars[pi].levels[lev].data[prof];
      vars[i].q    = data.modelLevelVars[qi].levels[lev].data[prof];
      vars[i].ql   = data.modelLevelVars[qli].levels[lev].data[prof];
      vars[i].qi   = data.modelLevelVars[qii].levels[lev].data[prof];
      vars[i].w    = data.modelLevelVars[wi].levels[lev].data[prof];
      vars[i].lscf = data.modelLevelVars[lscfi].levels[lev].data[prof];
      vars[i].u    = cnFcInitialValue;
      vars[i].v    = cnFcInitialValue;
      vars[i].ccf  = cnFcInitialValue;
    }
  }


  /* Interpolate the rho level variables onto the theta levels.
   * I assume here that the theta level for each model level is higher
   * than the corresponding rho level.
   */
  for ( i = 0, prof = 0; prof < numProfiles; prof++ )
  {
    float x1;
    float x2;
    float dx;
    float dx1;
    float dx2;
    float x;
    float y1;
    float y2;

    for ( lev = 0; lev < (numLevels - 1); lev++, i++ )
    {
      x   = vars[i].h;
      /* Use the heights from the zonal wind variable.
       * These are the rho heights and are the same for all the rho
       * level variables.
       */
      x1  = data.modelLevelVars[ui].levels[lev].height;
      x2  = data.modelLevelVars[ui].levels[lev + 1].height;

      dx  = x2 - x1;
      dx1 = x2 - x;
      dx2 = x - x1;

      y1  = data.modelLevelVars[ui].levels[lev].data[prof];
      y2  = data.modelLevelVars[ui].levels[lev + 1].data[prof];
      vars[i].u = (y1 * dx1 + y2 * dx2) / dx;

      y1  = data.modelLevelVars[vi].levels[lev].data[prof];
      y2  = data.modelLevelVars[vi].levels[lev + 1].data[prof];
      vars[i].v = (y1 * dx1 + y2 * dx2) / dx;

      y1  = data.modelLevelVars[ccfi].levels[lev].data[prof];
      y2  = data.modelLevelVars[ccfi].levels[lev + 1].data[prof];
      vars[i].ccf = (y1 * dx1 + y2 * dx2) / dx;
    }

    /* The last level value has to be extrapolated. */
    lev = data.header.numModelLevels - 1;
    x   = vars[i - 1].h;
    x1  = data.modelLevelVars[ui].levels[lev].height;
    x2  = data.modelLevelVars[ui].levels[lev + 1].height;

    dx  = x2 - x1;
    dx1 = x2 - x;
    dx2 = x - x1;

    y1  = data.modelLevelVars[ui].levels[lev].data[prof];
    y2  = data.modelLevelVars[ui].levels[lev + 1].data[prof];
    vars[i].u = (y1 * dx1 + y2 * dx2) / dx;

    y1  = data.modelLevelVars[vi].levels[lev].data[prof];
    y2  = data.modelLevelVars[vi].levels[lev + 1].data[prof];
    vars[i].v = (y1 * dx1 + y2 * dx2) / dx;

    y1  = data.modelLevelVars[ccfi].levels[lev].data[prof];
    y2  = data.modelLevelVars[ccfi].levels[lev + 1].data[prof];
    vars[i].ccf = (y1 * dx1 + y2 * dx2) / dx;

    i++;
  }


  cdlPrint(stdout, INSTITUTE, MODEL, FILETYPE,
    data.header.initTime, data.header.initTime,
    utcOffset, data.header.site,
    data.header.lat, data.header.lon,
    numProfiles, numLevels,
    CN_VAR_PRESSURE,
    CN_VAR_HEIGHT,
    CN_VAR_WIND_U,
    CN_VAR_WIND_V,
    CN_VAR_WIND_W,
    CN_VAR_TEMPERATURE,
    CN_VAR_Q,
    CN_VAR_QL,
    CN_VAR_QI,
    CN_VAR_LS_CLOUD_FRACTION,
    CN_VAR_CONV_CLOUD_FRACTION,
    -1
  );

  printf("time = ");
  for ( prof = 0; prof < numProfiles; prof++ )
  {
    if ( prof != 0 ) printf(",\n");

    printf("%d", prof * TIME_STEP);
  }
  printf(";\n");


#define PRINTVAR(varId, varName)                                         \
  cdlPrintVar(stdout, varId, vars, sizeof(TVars),                        \
    cnOffsetOf(TVars, varName), numProfiles, numLevels)


  PRINTVAR(CN_VAR_HEIGHT, h);
  PRINTVAR(CN_VAR_PRESSURE, p);
  PRINTVAR(CN_VAR_WIND_U, u);
  PRINTVAR(CN_VAR_WIND_V, v);
  PRINTVAR(CN_VAR_WIND_W, w);
  PRINTVAR(CN_VAR_TEMPERATURE, t);
  PRINTVAR(CN_VAR_Q, q);
  PRINTVAR(CN_VAR_QL, ql);
  PRINTVAR(CN_VAR_QI, qi);
  PRINTVAR(CN_VAR_LS_CLOUD_FRACTION, lscf);
  PRINTVAR(CN_VAR_CONV_CLOUD_FRACTION, ccf);

  printf("\n}\n");
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  /* 'filename' is the filename of the file to process and 'infile' is
   * 'filename' opened file pointer.
   */
  char * filename;
  FILE * infile;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */
  cl.utcOffset = 0;
  cl.firstArg  = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);

  /**********************************************************************/

  filename = argv[cl.firstArg];
  infile   = fopen(filename, "r");
  if ( infile == NULL )
  {
    printError(PE_EXIT, "unable to open file: %s\n", filename);
  }

  processFile(filename, infile, cl.utcOffset);
  fclose(infile);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

