/*
 * CloudNET Model Conversion Software
 *
 * Convert UKMO 'profile' file to netCDF/CDL.
 *
 * HISTORY
 * 2004-02-27 ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnnetcdf.h"
#include "cnmoumm.h"
#include "version.h"
#include "cdl.c"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS      = 1,

  /* maximum line length for a line read from the file. */
  MAX_LINE_LEN = 4096,

  /* Maximum number of lines that can be read from the file. */
  MAX_LINES    = 16384
};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{
  int utcOffset;

  /* Index into 'argv[]' of the first argument after the options.
   * Or -1 if 'MINARGS=0'.
   */
  int firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Convert UK Met Office Unified Model (mesoscale) output ASCII files"
    " with level variables (i.e. 'Profile' files) to netCDF-CDL.\n"
    "  Outputs CDL text to stdout. This can be piped through 'ncgen -b'"
    " to produce a netCDF dataset.\n"
  );
  fprintf(outfile,
    "  The output CDL is given a name with the format:\n"
    "    MOUMM_VAR_yyyymmdd_hhmm_<site>\n"
    "  Where \"yyyymmdd\" and \"hhmm\" is the initialization date and"
    " time from the input file. <site> will be the first two characters"
    " of the site the data is for.\n"
    "  This name can be overridden when running 'ncgen' by using the '-o'"
    " option with 'ncgen'.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
    "\n"
    "  -u minutes\n"
    "    UTC offset for all times in the file in minutes "
    "(can be negative). Default: 0\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file' is the ASCII output file from the UKMO UM"
    " containing the variables at each level for each profile.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;

      case 'u':
        cl.utcOffset = atoi(getOptArg(option, argc, argv, &argn));
	break;

      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = ( MINARGS > 0 ) ? argn : -1;
}


/************************************************************************/
/************************************************************************/


void processFile(char * filename, FILE * infile, int utcOffset)
{
  const int              INSTITUTE               = CN_MI_MO;
  const int              FILETYPE                = CN_FT_VAR;
        int              MODEL                   = CN_ML_MOUMM;

  /* 'data' contains each line from the file. */
  static cnTMOUMMvarLine data[MAX_LINES];
  static cnTMOUMMheader  hdr;
  static char            line[MAX_LINE_LEN + 1];

  int                    numLines;
  int                    numProfiles;
  int                    numLevels;
  int                    prof;
  int                    wvarId;

  char                 * p;

  /**********************************************************************/

  /* Check the filename to determine which model */
  /* Default is mesoscale */
  MODEL = CN_ML_MOUMM;
  
  p = strstr(filename, "NAE");
  if ( p != NULL ) {
    MODEL = CN_ML_MOUNA;
  }
  else {
    p = strstr(filename, "HIGHRES");
    if ( p != NULL ) {
      MODEL = CN_ML_MOUHR;
    }
  }

  cnMOUMMGetHeader(filename, infile, line, MAX_LINE_LEN, &hdr);


  /**********************************************************************/

  /* Read each line from the file. */
  numLines    = 0;
  numLevels   = 0;
  numProfiles = 0;
  while ( fgets(line, MAX_LINE_LEN, infile) != NULL )
  {
    if ( numLines >= MAX_LINES )
    {
      printError(PE_EXIT,
        "%s: the number of lines exceeds %d which is the preset limit.\n"
        "You'll need to edit the program's source code and increase the "
        "value MAX_LINES", filename);
    }

    cnMOUMMGetVarLine(line, &data[numLines]);

    if ( data[numLines].level > numLevels )
    {
      numLevels = data[numLines].level;
    }

    /* levels are 0-based. */
    if ( data[numLines].level == 0 )
    {
      numProfiles++;
    }

    numLines++;
  }

  /* levels are 0-based. */
  numLevels++;

  /**********************************************************************/

  wvarId = -1;
  if ( hdr.wOrOmega == 'O' )
  {
    wvarId = CN_VAR_OMEGA;
  }
  else
  if ( hdr.wOrOmega == 'W' )
  {
    wvarId = CN_VAR_WIND_W;
  }

  cdlPrint(stdout, INSTITUTE, MODEL, FILETYPE,
    hdr.initTime, data[0].verTime,
    utcOffset, hdr.site, hdr.lat, hdr.lon,
    numProfiles, numLevels,
    CN_VAR_PRESSURE,
    CN_VAR_HEIGHT,
    CN_VAR_WIND_U,
    CN_VAR_WIND_V,
    CN_VAR_TEMPERATURE,
    CN_VAR_Q,
    CN_VAR_RH,
    CN_VAR_QL,
    CN_VAR_QI,
    CN_VAR_LS_CLOUD_FRACTION,
    CN_VAR_CONV_CLOUD_FRACTION,
    wvarId,
    -1
  );

  printf("time = ");
  for ( prof = 0; prof < numProfiles; prof++ )
  {
    if ( prof != 0 ) printf(",\n");

    printf("%d", data[prof * numLevels].verTime - hdr.initTime);
  }
  printf(";\n");


#define PRINTVAR(varId, varName)                                         \
  cdlPrintVar(stdout, varId, data, sizeof(cnTMOUMMvarLine),              \
    cnOffsetOf(cnTMOUMMvarLine, varName), numProfiles, numLevels)


  PRINTVAR(CN_VAR_PRESSURE, pressure);
  PRINTVAR(CN_VAR_HEIGHT, height);
  PRINTVAR(CN_VAR_WIND_U, uwind);
  PRINTVAR(CN_VAR_WIND_V, vwind);
  PRINTVAR(CN_VAR_TEMPERATURE, temp);
  PRINTVAR(CN_VAR_Q, q);
  PRINTVAR(CN_VAR_RH, rh);
  PRINTVAR(CN_VAR_QL, ql);
  PRINTVAR(CN_VAR_QI, qi);
  PRINTVAR(CN_VAR_LS_CLOUD_FRACTION, largeCloudFrac);
  PRINTVAR(CN_VAR_CONV_CLOUD_FRACTION, convCloudFrac);
  if ( wvarId >= 0 )
  {
    PRINTVAR(wvarId, womega);
  }

  printf("\n}\n");
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  /* 'filename' is the filename of the file to process and 'infile' is
   * 'filename' opened file pointer.
   */
  char * filename;
  FILE * infile;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */
  cl.utcOffset = 0;
  cl.firstArg  = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);

  /**********************************************************************/

  filename = argv[cl.firstArg];
  infile   = fopen(filename, "r");
  if ( infile == NULL )
  {
    printError(PE_EXIT, "unable to open file: %s\n", filename);
  }

  processFile(filename, infile, cl.utcOffset);
  fclose(infile);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

