

/* Returns the byte offset from the beginning of the structure 's' of
 * the field member 'm'.
 * NOTES: this is used when 's' is a structure type or tag.
 */
#define cnOffsetOf(s,m)\
  ((int)((char *)(&(((s *)NULL)->m)) - (char *)NULL))


/* Returns the byte offset from the beginning of the structure 's' of
 * the field member 'm'.
 * NOTES: this is used when 's' is a variable.
 */
#define cnOffset(s,m) ((int)((char *)&s.m - (char *)&s))


const float cnFillValue = -999.0F;


/* Print the first part of the CDL output.
 * PRE: 'outfile' is the text file to write the CDL to.
 * institute==cnEInstitutes from 'cnmodels.h'.
 * model==cnEModels from 'cnmodels.h'.
 * fileType==cnEFileTypes from 'cnmodels.h'.
 * 'initTime' is the initialization time in minutes from 1960.
 * 'firstTime' is the time of the first forecast in minutes from 1960.
 * 'utcOffset' minutes from UTC.
 * site==cnESites from 'cnsites.h'.
 * 'latitude' and 'longitude' are for the model grid point.
 * If 'numLevels'<2 then the level dimension and variable are not
 * defined, otherwise they are defined and the data for the levels
 * is also written.
 * After 'numLevels' is a list of integers terminated with a negative
 * value. The integers should be the 'cnEVariableIds'.
 *
 * POST: everything upto the latitude, longitude and level variables data
 * are written.
 * After a call, all that is required is the profiles and model variables
 * data to be written.
 */
void cdlPrint(FILE * outfile, int institute, int model, int fileType,
  int initTime, int firstTime, int utcOffset,
  int site, float latitude, float longitude,
  int numProfiles, int numLevels, ...)
{
  static char * modelNames[] = 
  {
    "MOUMM",
    "MOUMG",
    "ECIFS",
    "MFARP",
    "KNHIR",
    "KNRAC"
  };

  static char * fileTypes[] = 
  {
    "VAR", "FLX", "SFC"
  };

  char    siteName[3];
  va_list ap;
  int     varId;

  strncpy(siteName, cnGetSiteName(site, 'u'), 2);

  /* Print the header line and the dimensions. */
  fprintf(outfile,
    "netcdf %s_%s_%08d_%04d_%s {\n\n"
    "dimensions:\n"
    "  time  = %d;\n"
    "  level = %d;\n"
    ,
    modelNames[model], fileTypes[fileType],
    utDateFromMins(firstTime), utTimeFromMins(firstTime),
    siteName,
    numProfiles, numLevels
  );

  /* The variables. */
  fprintf(outfile,
    "\n"
    "variables:\n"
  );

  fprintf(outfile,
    "  float latitude;\n"
    "  float longitude;\n"
  );

  fprintf(outfile,
    "  int time(time);\n"
  );

  va_start(ap, numLevels);
  varId = va_arg(ap, int);
  while ( varId >= 0 )
  {
    fprintf(outfile, "  float %s(time", ncVariables[varId].name);

    if ( numLevels > 1 )
    {
      fprintf(outfile, ", level\n");
    }
    fprintf(outfile, ");\n");
    varId = va_arg(ap, int);
  }
  va_end(ap);


  /* Global attributes. */
  fprintf(outfile,
    ":site           = %d;\n"
    ":institute      = %d;\n"
    ":model          = %d;\n"
    ":file_type      = %d;\n"
    ":initialization = %d;\n"
    ":utc_offset     = %d;\n"
    ,
    site,
    institute,
    model,
    fileType,
    initTime,
    utcOffset
  );


  /* The data section. */
  fprintf(outfile,
    "data:\n"
    "latitude  = %.3f;\n"
    "longitude = %.3f;\n"
    ,
    latitude, longitude
  );
}


void cdlPrintValue(FILE * outfile, int varId, float value)
{
  if ( value < cnFillValue )
  {
    value = cnFillValue;
  }

  fprintf(outfile, "%.*f", ncVariables[varId].precision, value);
}


/* Print the data for each level for each profile for a single variable
 * from an array of structures.
 * PRE: 'varId' is the variables' ID.
 * 'data' is a single dimension array of structures. 'sizeObj' is the
 * size of each structure in the 'data' array.
 * 'varOffset' is the offset within the structure of the field to
 * printout. The field must be a float.
 */
void cdlPrintVar(FILE * outfile, int varId, void * data, int sizeObj,
  int varOffset, int numProfiles, int numLevels)
{
  unsigned char * pdata = (unsigned char *) data;
  int             prof;

  fprintf(outfile, "%s = ", ncVariables[varId].name);
  for ( prof = 0; prof < numProfiles; prof++ )
  {
    int lev;

    if ( prof != 0 ) fprintf(outfile, ",\n");

    for ( lev = 0; lev < numLevels; lev++ )
    {
      float f = * ((float *) (pdata + (prof * numLevels + lev)
        * sizeObj + varOffset));

      if ( lev != 0 ) fprintf(outfile, ",\n");

      cdlPrintValue(outfile, varId, f);
    }
  }
  fprintf(outfile, ";\n");
}


/* Print the data for a single level file for each profile for a single
 * variable from an array of structures.
 * PRE: 'varId' is the variables' ID.
 * 'data' is a single dimension array of structures. 'sizeObj' is the
 * size of each structure in the 'data' array.
 * 'varOffset' is the offset within the structure of the field to
 * printout. The field must be a float.
 */
void cdlPrintSingleVar(FILE * outfile, int varId, void * data,
  int sizeObj, int varOffset, int numProfiles)
{
  unsigned char * pdata = (unsigned char *) data;
  int             prof;

  fprintf(outfile, "%s = ", ncVariables[varId].name);
  for ( prof = 0; prof < numProfiles; prof++ )
  {
    float f = * ((float *) (pdata + prof * sizeObj + varOffset));

    if ( prof != 0 ) fprintf(outfile, ",\n");

    cdlPrintValue(outfile, varId, f);
  }
  fprintf(outfile, ";\n");
}
