

#include "cnmoumg.h"


/************************************************************************/
/************************************************************************/


/*
Problem: 001 P can appear twice on rho levels and on theta levels with
the same GRIB code.
*/


enum
{
  GRIB_RH               = 52,
  GRIB_TOT_CLOUD_FRAC   = 71,
  GRIB_LOW_CLOUD_FRAC   = 73,
  GRIB_MED_CLOUD_FRAC   = 74,
  GRIB_HI_CLOUD_FRAC    = 75,
  GRIB_LARGE_CLOUD_FRAC = 149,
  GRIB_CONV_CLOUD_FRAC  = 150,
  GRIB_RH_1_5M          = 154
};


/************************************************************************/
/************************************************************************/


static void scaleArray(float * data, int n, float scale)
{
  int i;
  for ( i = 0; i < n; i++ )
  {
    data[i] *= scale;
  }
}


static void testGribCode(int grib, float * data, int n)
{
  /* These are all percentages in the files. */
  if ( grib == GRIB_RH || grib == GRIB_TOT_CLOUD_FRAC
    || grib == GRIB_LOW_CLOUD_FRAC || grib == GRIB_MED_CLOUD_FRAC
    || grib == GRIB_HI_CLOUD_FRAC || grib == GRIB_LARGE_CLOUD_FRAC
    || grib == GRIB_CONV_CLOUD_FRAC || grib == GRIB_RH_1_5M )
  {
    scaleArray(data, n, 0.01);
  }
}


void cnMOUMGreadFile(char * filename, FILE * infile, char * line,
  int max, cnTMOUMGfile * f/*OUT*/)
{
  int idate;
  int itime;
  int ok;
  int i;

  /* The header should be something like:
13 18 38 41  8 11
Model output time series (MOLTS) from global NWP model for CEOP sites
Initial data time 20021101   0
 
Site                               = CloudNET Cabauw
Longitude/latitude/elevation(m)    =    4.93/  51.97/-999.99
Grid point of NN                   =   6/256
NN longitude/latitude/elevation(m) =    5.00/  52.22/  -6.88
Distance from site (km)            =   28.38
Land(1) or sea(0)                  = 1
Sea-ice cover (%)                  = -999.99
   */

  fgets(line, max, infile);

  sscanf(line, "%d %d %d %d %d %d",
    &f->header.numTimes,
    &f->header.numPressureLevels,
    &f->header.numModelLevels,
    &f->header.numVars,
    &f->header.numPressureVars,
    &f->header.numModelVars
  );


  if( f->header.numTimes > CN_MOUMG_MAX_TIMES )
  {
    printError(PE_EXIT,
      "%s: the number of output times in the header is too big (%d)",
      filename, f->header.numTimes);
  }

  if( f->header.numPressureLevels > CN_MOUMG_MAX_PRESSURE_LEVELS )
  {
    printError(PE_EXIT,
      "%s: the number of pressure levels in the header is too big (%d)",
      filename, f->header.numPressureLevels);
  }

  if( f->header.numModelLevels > CN_MOUMG_MAX_MODEL_LEVELS )
  {
    printError(PE_EXIT,
      "%s: the number of model levels in the header is too big (%d)",
      filename, f->header.numModelLevels);
  }

  if( f->header.numVars > CN_MOUMG_MAX_SINGLE_LEVEL_VARS )
  {
    printError(PE_EXIT,
      "%s: the number of single level variables in the header is too big "
      "(%d)",
      filename, f->header.numVars);
  }

  if( f->header.numPressureVars > CN_MOUMG_MAX_PRESSURE_VARS )
  {
    printError(PE_EXIT,
      "%s: the number of pressure variables in the header is too big (%d)",
      filename, f->header.numPressureVars);
  }

  if( f->header.numModelVars > CN_MOUMG_MAX_MODEL_VARS )
  {
    printError(PE_EXIT,
      "%s: the number of model variables in the header is too big (%d)",
      filename, f->header.numModelVars);
  }

  fgets(line, max, infile);
  fgets(line, max, infile);

  sscanf(line, "Initial data time %d %d", &idate, &itime);

  f->header.initTime = utToMins(idate, itime);

  fgets(line, max, infile);
  fgets(line, max, infile);

  utUpper(line);

  if ( strstr(&line[36], "ABERYSTWYTH") != NULL )
  {
    f->header.site = CN_SITE_AB;
  }
  else
  if ( strstr(&line[36], "CHILBOLTON") != NULL )
  {
    f->header.site = CN_SITE_CH;
  }
  else
  if ( strstr(&line[36], "CABAUW") != NULL )
  {
    f->header.site = CN_SITE_CA;
  }
  else
  if ( strstr(&line[36], "PALAISEAU") != NULL )
  {
    f->header.site = CN_SITE_PA;
  }
  else
  if ( strstr(&line[36], "PARIS") != NULL )
  {
    f->header.site = CN_SITE_PA;
  }
  else
  if ( strstr(&line[36], "LINDENBERG") != NULL )
  {
    f->header.site = CN_SITE_LI;
  }
  else
  if ( strstr(&line[36], "POTENZA") != NULL )
  {
    f->header.site = CN_SITE_PO;
  }
  else
  if ( strstr(&line[36], "SOUTHERN") != NULL )
    {
      f->header.site = CN_SITE_SG;
    }
  else
  if ( strstr(&line[36], "NORTH") != NULL )
    {
      f->header.site = CN_SITE_NS;
    }
  else
  if ( strstr(&line[36], "NAURU") != NULL )
    {
      f->header.site = CN_SITE_T1;
    }
  else
  if ( strstr(&line[36], "MANUS") != NULL )
    {
      f->header.site = CN_SITE_T2;
    }
  else
  if ( strstr(&line[36], "DARWIN") != NULL )
    {
      f->header.site = CN_SITE_T3;
    }
  else
  if ( strstr(&line[36], "NIAMEY") != NULL )
    {
      f->header.site = CN_SITE_NI;
    }
  else
  {
    printError(PE_EXIT, "%s: unrecognized site in header: %s",
      filename, line);
  }


  fgets(line, max, infile);
  sscanf(&line[36], "%f/%f/%f",
    &f->header.lon, &f->header.lat, &f->header.elevation);

  fgets(line, max, infile);
  sscanf(&line[36], "%d/%d",
    &f->header.gridPointNN1, &f->header.gridPointNN2);

  fgets(line, max, infile);
  sscanf(&line[36], "%f/%f/%f",
    &f->header.lonNN, &f->header.latNN, &f->header.elevationNN);

  fgets(line, max, infile);
  sscanf(&line[36], "%f", &f->header.distanceFromSite);
  if ( f->header.distanceFromSite > -999.0F )
  {
    f->header.distanceFromSite *= 1000.0F;
  }

  fgets(line, max, infile);
  sscanf(&line[36], "%d", &f->header.landOrSea);

  fgets(line, max, infile);
  sscanf(&line[36], "%f", &f->header.seaIceFrac);

  if ( f->header.seaIceFrac > -999.0F )
  {
    f->header.seaIceFrac /= 100.0F;
  }


  /* Now the data. */

  /* The first section of data should be the single level data. */
  ok = 0;
  while ( fgets(line, max, infile) != NULL )
  {
    if ( strstr(line, "single level data") != NULL )
    {
      ok = 1;
      break;
    }
  }

  if ( !ok )
  {
    printError(PE_EXIT, "%s: could not find the single level data",
      filename);
  }

  fgets(line, max, infile);
  fgets(line, max, infile);

  for ( i = 0; i < f->header.numVars; i++ )
  {
    char * p;
    int    j;

    fgets(line, max, infile);
    f->singleLevelVars[i].grib = atoi(line);

    p = &line[46];
    for ( j = 0; j < f->header.numTimes; j++, p += 14 )
    {
      f->singleLevelVars[i].data[j] = atof(p);
    }

    testGribCode(f->singleLevelVars[i].grib, f->singleLevelVars[i].data,
      f->header.numTimes);
  }


  /* The second section of data should be the pressure level data. */
  ok = 0;
  while ( fgets(line, max, infile) != NULL )
  {
    if ( strstr(line, "Vertical profiles (pressure levels)") != NULL )
    {
      ok = 1;
      break;
    }
  }

  if ( !ok )
  {
    printError(PE_EXIT, "%s: could not find the pressure level data",
      filename);
  }

  fgets(line, max, infile);
  fgets(line, max, infile);
  fgets(line, max, infile);

  for ( i = 0; i < f->header.numPressureVars; i++ )
  {
    int j;

    fgets(line, max, infile);
    f->pressureLevelVars[i].grib = atoi(line);

    for ( j = 0; j < f->header.numPressureLevels; j++ )
    {
      char * p;
      int    k;

      fgets(line, max, infile);
      f->pressureLevelVars[i].levels[j].pressure = atof(line);

      p = &line[12];
      for ( k = 0; k < f->header.numTimes; k++, p += 14 )
      {
        f->pressureLevelVars[i].levels[j].data[k] = atof(p);
      }

      testGribCode(f->pressureLevelVars[i].grib,
        f->pressureLevelVars[i].levels[j].data,
        f->header.numTimes);
    }
  }


  /* The third section of data should be the pressure level data. */
  ok = 0;
  while ( fgets(line, max, infile) != NULL )
  {
    if ( strstr(line, "Vertical profiles (model levels)") != NULL )
    {
      ok = 1;
      break;
    }
  }

  if ( !ok )
  {
    printError(PE_EXIT, "%s: could not find the model level data",
      filename);
  }

  fgets(line, max, infile);
  fgets(line, max, infile);
  fgets(line, max, infile);

  for ( i = 0; i < f->header.numModelVars; i++ )
  {
    int j;

    fgets(line, max, infile);
    f->modelLevelVars[i].grib = atoi(line);

    for ( j = 0; j < f->header.numModelLevels; j++ )
    {
      char * p;
      int    k;

      fgets(line, max, infile);
      sscanf(line, "%d%f",
        &f->modelLevelVars[i].levels[j].level,
        &f->modelLevelVars[i].levels[j].height
      );

      p = &line[12];
      for ( k = 0; k < f->header.numTimes; k++, p += 14 )
      {
        f->modelLevelVars[i].levels[j].data[k] = atof(p);
      }

      testGribCode(f->modelLevelVars[i].grib,
        f->modelLevelVars[i].levels[j].data,
        f->header.numTimes);
    }
  }
}


int cnMOUMGsearchSingle(cnTMOUMGfile * f, int grib)
{
  int retValue;

  for ( retValue = f->header.numVars - 1;
    retValue >= 0 && f->singleLevelVars[retValue].grib != grib;
    retValue-- )
  {
  }

  return retValue;
}


int cnMOUMGsearchPressure(cnTMOUMGfile * f, int grib)
{
  int retValue;

  for ( retValue = f->header.numPressureVars - 1;
    retValue >= 0 && f->pressureLevelVars[retValue].grib != grib;
    retValue-- )
  {
  }

  return retValue;
}


int cnMOUMGsearchModel(cnTMOUMGfile * f, int grib)
{
  int retValue;

  for ( retValue = f->header.numModelVars - 1;
    retValue >= 0 && f->modelLevelVars[retValue].grib != grib;
    retValue-- )
  {
  }

  return retValue;
}


/************************************************************************/
/************************************************************************/


#ifdef aldjakljf
char * PROGNAME = "test";


void printSynopsis(FILE * outfile, int exitProg)
{
}


void printHeader(cnTMOUMGfile * f)
{
  int i;

  printf(
    "numTimes            : %d\n"
    "numPressureLevels   : %d\n"
    "numModelLevels      : %d\n"
    "numVars             : %d\n"
    "numPressureVars     : %d\n"
    "numModelVars        : %d\n"
    "initTime            : %d\n"
    "site                : %d\n"
    "lon/lat/elevation   : %.3f/%.3f/%.3f\n"
    "NN grid point       : %d/%d\n"
    "NN lat/lon/elevation: %.3f/%.3f/%.3f\n"
    "Distance from site  : %.3f\n"
    "Land or sea         : %d\n"
    "Sea ice fraction    : %.3f\n"
    ,
    f->header.numTimes,
    f->header.numPressureLevels,
    f->header.numModelLevels,
    f->header.numVars,
    f->header.numPressureVars,
    f->header.numModelVars,
    f->header.initTime,
    f->header.site,
    f->header.lon,
    f->header.lat,
    f->header.elevation,
    f->header.gridPointNN1,
    f->header.gridPointNN2,
    f->header.lonNN,
    f->header.latNN,
    f->header.elevationNN,
    f->header.distanceFromSite,
    f->header.landOrSea,
    f->header.seaIceFrac
  );


  for ( i = 0; i < f->header.numVars; i++ )
  {
    int j;
    printf("%4d", f->singleLevelVars[i].grib);
    for ( j = 0; j < f->header.numTimes; j++ )
    {
      printf("%14.6e", f->singleLevelVars[i].data[j]);
    }
    printf("\n");
  }


  for ( i = 0; i < f->header.numPressureVars; i++ )
  {
    int j;
    printf("%4d\n", f->pressureLevelVars[i].grib);
    for ( j = 0; j < f->header.numPressureLevels; j++ )
    {
      int k;
      printf("%14.0f", f->pressureLevelVars[i].levels[j].pressure);
      for ( k = 0; k < f->header.numTimes; k++ )
      {
        printf("%14.6e", f->pressureLevelVars[i].levels[j].data[k]);
      }
      printf("\n");
    }
  }


  for ( i = 0; i < f->header.numModelVars; i++ )
  {
    int j;
    printf("%4d\n", f->modelLevelVars[i].grib);
    for ( j = 0; j < f->header.numModelLevels; j++ )
    {
      int k;
      printf("%4d%14.1f",
        f->modelLevelVars[i].levels[j].level,
        f->modelLevelVars[i].levels[j].height);
      for ( k = 0; k < f->header.numTimes; k++ )
      {
        printf("%14.6e", f->modelLevelVars[i].levels[j].data[k]);
      }
      printf("\n");
    }
  }
}


int main(void)
{
  static cnTMOUMGfile f;
  char * filename = "../../data/um-global/MOLTS_DA2002110100Z_CNET_chi.dat";
  FILE * infile;
  static char line[4096];


  infile = fopen(filename, "r");
  if ( infile== NULL )
    printError(PE_EXIT, "unable to open file: %s", filename);

  cnMOUMGreadFile(filename, infile, line, 4095, &f);

  printHeader(&f);


  return 0;
}

#endif

