/*
 * CloudNET Model Conversion Software
 *
 * Stuff for UKMO UM Global output files.
 *
 * HISTORY
 * 2004-03-03 ao started.
 *
 * AUTHORS
 * Andrew Openshaw
 */


#if !defined(CNMOUMG_H)
#define CNMOUMG_H


/************************************************************************/
/************************************************************************/


#include "cnapp.h"
#include "cnsites.h"
#include "cnutils.h"


/************************************************************************/
/************************************************************************/


/* Constants for static arrays. */
enum
{
  CN_MOUMG_MAX_TIMES             = 32,
  CN_MOUMG_MAX_SINGLE_LEVEL_VARS = 64,
  CN_MOUMG_MAX_PRESSURE_VARS     = 32,
  CN_MOUMG_MAX_PRESSURE_LEVELS   = 64,
  CN_MOUMG_MAX_MODEL_VARS        = 32,
  CN_MOUMG_MAX_MODEL_LEVELS      = 64
};


/************************************************************************/
/************************************************************************/


/* Structure to hold the data for one file. */
typedef struct
{
  struct
  {
    int   numTimes;           /* number of data output times. */
    int   numPressureLevels;
    int   numModelLevels;
    int   numVars;
    int   numPressureVars;
    int   numModelVars;
    int   initTime;           /* Minutes since 1960. */
    int   site;
    float lon;
    float lat;
    float elevation;
    int   gridPointNN1;       /* Grid point of nearest neighbour. */
    int   gridPointNN2;
    float lonNN;
    float latNN;
    float elevationNN;
    float distanceFromSite;   /* metres. */
    int   landOrSea;          /* 0-sea, 1-land. */
    float seaIceFrac;
  } header;

  /* In all of these, the data arrays will have 'numTimes' values. */

  /* There will be 'numVars' of these. */
  struct
  {
    int   grib; /* GRIB code for the variable. */
    float data[CN_MOUMG_MAX_TIMES];
  } singleLevelVars[CN_MOUMG_MAX_SINGLE_LEVEL_VARS];

  /* There will be 'numPressureVars' of these. */
  struct
  {
    int   grib; /* GRIB code for the variable. */

    /* There will be 'numPressureLevels' of these. */
    struct
    {
      float pressure;
      float data[CN_MOUMG_MAX_TIMES];
    } levels[CN_MOUMG_MAX_PRESSURE_LEVELS];
  } pressureLevelVars[CN_MOUMG_MAX_PRESSURE_VARS];

  /* There will be 'numModelVars' of these. */
  struct
  {
    int   grib; /* GRIB code for the variable. */

    /* There will be 'numModelLevels' of these. */
    struct
    {
      int   level;
      float height;
      float data[CN_MOUMG_MAX_TIMES];
    } levels[CN_MOUMG_MAX_MODEL_LEVELS];
  } modelLevelVars[CN_MOUMG_MAX_MODEL_VARS];
} cnTMOUMGfile;


/************************************************************************/
/************************************************************************/


/* Reads the entire global model output file into 'f'. */
void cnMOUMGreadFile(char * filename, FILE * infile, char * line,
  int max, cnTMOUMGfile * f/*OUT*/);


/* Search 'f->singleLevelVars' for matching GRIB code.
 * POST: returns index into 'f->singleLevelVars' for the element with
 * GRIB code matching 'grib' or -1 if not found.
 */
int cnMOUMGsearchSingle(cnTMOUMGfile * f, int grib);


/* Search 'f->pressureLevelVars' for matching GRIB code.
 * POST: returns index into 'f->pressureLevelVars' for the element with
 * GRIB code matching 'grib' or -1 if not found.
 */
int cnMOUMGsearchPressure(cnTMOUMGfile * f, int grib);


/* Search 'f->modelLevelVars' for matching GRIB code.
 * POST: returns index into 'f->modelLevelVars' for the element with
 * GRIB code matching 'grib' or -1 if not found.
 */
int cnMOUMGsearchModel(cnTMOUMGfile * f, int grib);


/************************************************************************/
/************************************************************************/


#endif
