/*
 * CloudNET Model Conversion Software
 *
 * Changes the time variable values in a IM netCDF dataset.
 *
 * HISTORY
 * 2004-03-25 ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <netcdf.h>
#include "cnfc.h"
#include "cnnetcdf.h"
#include "version.h"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS = 1
};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{
  int firstTime;
  int step;

  /* Index into 'argv[]' of the first argument after the options. */
  int firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile, 
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file...\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Modifies the time values in the time variable of IM netCDF"
    " datasets.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
    "\n"
    "  -f minutes\n"
    "    Set the first time value to 'minutes'. Default: don't change"
    " the first value.\n"
    "    Remember - the times are the minutes after initialization.\n"
    "\n"
    "  -s minutes\n"
    "    The number of minutes between each value in the time variable."
    " Default: 60\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file...' are IM netCDF datasets.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;

      case 'f':
        cl.firstTime = atoi(getOptArg(option, argc, argv, &argn));
	break;

      case 's':
        cl.step = atoi(getOptArg(option, argc, argv, &argn));
	break;

      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = argn;
}


/************************************************************************/
/************************************************************************/


void processFile(char * filename, int firstTime, int step)
{
  static int time[CN_NC_MAX_TIME_DIM];
  int        timeDimId;
  size_t     timeDim;
  int        timeVarId;
  int        fileId;
  size_t     i;


  NCEXIT(nc_open(filename, NC_WRITE, &fileId));

  NCEXIT(nc_inq_dimid(fileId, "time", &timeDimId));
  NCEXIT(nc_inq_dimlen(fileId, timeDimId, &timeDim));

  if ( timeDim > CN_NC_MAX_TIME_DIM )
  {
    printError(PE_EXIT, "%s: time dimension too long", filename);
  }

  NCEXIT(nc_inq_varid(fileId, "time", &timeVarId));
  NCEXIT(nc_get_var_int(fileId, timeVarId, time));

  if ( firstTime >= 0 )
  {
    time[0] = firstTime;
  }

  for ( i = 1; i < timeDim; i++ )
  {
    time[i] = time[i - 1] + step;
  }

  NCEXIT(nc_put_var_int(fileId, timeVarId, time));

  NCEXIT(nc_close(fileId));
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */
  cl.firstTime = -1;
  cl.step      = 60;
  cl.firstArg  = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);

  for ( ; cl.firstArg < argc; cl.firstArg++ )
  {
    processFile(argv[cl.firstArg], cl.firstTime, cl.step);
  }

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

