/*
 * CloudNET Model Conversion Software
 *
 * Create a FC dataset from several IM datasets.
 *
 * HISTORY
 * 2004-03-19 ao new version.
 *
 * 2004-05-11  ao  in 'printLineData()' use "*g" instead of "*f" format
 *   in the fprintf() call.
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnutils.h"
#include "cnsites.h"
#include "cnim.h"
#include "cnfc.h"
#include "cnnetcdf.h"
#include "version.h"
#include "cnapp.h"


/************************************************************************/
/************************************************************************/


enum
{
  MINARGS          = 1,

  DEF_START_TIME   = 0,
  DEF_NUM_PROFILES = 25,
  DEF_STEP         = 100     /* HHMM */
};


char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{
  int     startDate;    /* YYYYMMDD */
  int     startTime;    /* HHMM */
  int     step;         /* HHMM */
  int     numProfiles;

  char *  user;

  char ** modelFiles;
  int     numModelFiles;

  char *  outFilename;

  int     forecastMin;  /* minutes. */
  int     forecastMax;  /* minutes. */
} cl;


/************************************************************************/
/************************************************************************/


void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file ...\n"
    ,
    PROGNAME
  );

  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Each file is read in turn in the same order they appear on the"
    " command line. The files must be netCDF files with a models'"
    " output created by one of the stage1 programs.\n"
  );

  fprintf(outfile,
    "\n"
    "  Only one models' data can be processed during a single run of"
    " the program.\n"
    "\n"
    "  If the data within the file falls within the time period"
    " required for the dataset then it will be added to the dataset.\n"
    "\n"
    "  When all the files have been read and if any data has been added"
    " to the dataset then the CDL output is produced.\n"
  );

  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -d date\n"
    "    Start date for the data set. Format is YYYYMMDD. This is"
    " optional and if it is not given then the first date of the"
    " first file read will be used.\n"
  );

  fprintf(outfile,
    "\n"
    "  -o filename\n"
    "    Send CDL output to filename. Default is to send to standard"
    " output.\n"
    "\n"
    "  -p numProfiles\n"
    "    The number of profiles in the data set. Default: %d.\n"
    ,
    DEF_NUM_PROFILES
  );

  fprintf(outfile,
    "\n"
    "  -s step\n"
    "    The time interval between each profile. Format is HHMM."
    " Default: %04d.\n"
    ,
    DEF_STEP
  );

  fprintf(outfile,
    "\n"
    "  -t time\n"
    "    Start time for the data set. Format is HHMM. Default: %04d.\n"
    "\n"
    "  -u userName\n"
    "    User name for the comments in the netCDF file. If not given"
    " then one of the environment variables LOGNAME or USER is used.\n"
    "\n"
    "  -f min max\n"
    "    Only select profiles that have a forecast time in the range"
    " min..max. min and max are hours. They are the times since"
    " initialization of the model. The default is to disregard"
    " forecast times when selecting profiles.\n"
    "\n"
    ,
    DEF_START_TIME
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;

      case 'd':
        cl.startDate = atoi(getOptArg(option, argc, argv, &argn));
        break;

      case 'f':
        cl.forecastMin = atoi(getOptArg(option, argc, argv, &argn));
        cl.forecastMax = atoi(getOptArg(option, argc, argv, &argn));
        break;

      case 'o':
        cl.outFilename = getOptArg(option, argc, argv, &argn);
        break;

      case 'p':
        cl.numProfiles = atoi(getOptArg(option, argc, argv, &argn));
        break;

      case 's':
        cl.step = atoi(getOptArg(option, argc, argv, &argn));
        break;

      case 't':
        cl.startTime = atoi(getOptArg(option, argc, argv, &argn));
        break;

      case 'u':
        cl.user = getOptArg(option, argc, argv, &argn);
        break;

      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.modelFiles    = &argv[argn];
  cl.numModelFiles = argc - argn;

  if ( cl.user == NULL )
  {
    cl.user = utGetUser();
  }
}


/************************************************************************/
/************************************************************************/


/* Stuff for printing the data set as a CDL file. */


char * makeDatasetName(cnTFC * fc)
{
  static char * modelNames[] = 
  {
    "MOUMM",
    "MOUMG",
    "ECIFS",
    "MFARP",
    "KNHIR",
    "KNRAC",
    "SMRCA"
  };

  static char   retValue[FILENAME_MAX + 1];

  char          siteName[32];
  int           date      = utDateFromMins(fc->firstTime);
  int           time      = utTimeFromMins(fc->firstTime);


  strcpy(siteName, cnGetSiteName(fc->site, 'l'));

  sprintf(retValue, "%s_%08d_%04d_%s",
    modelNames[fc->model], date, time, siteName);

  return retValue;
}


char * makeTimeUnits(cnTFC * fc)
{
  static char units[100];
  int         year;
  int         month;
  int         day;
  int         hour;
  int         min;
  int         utcHour;
  int         utcMin;
  int         date    = utDateFromMins(fc->firstTime);
  int         time    = utTimeFromMins(fc->firstTime);

  utExplodeDate(date, &year, &month, &day);
  utExplodeTime(time, &hour, &min);
  utExplodeTime(fc->utcOffset, &utcHour, &utcMin);
  if ( utcMin < 0 )
  {
    utcMin -= utcMin;
  }

  sprintf(units, "hours since %04d-%02d-%02d %02d:%02d:00 %+03d:%02d",
    year, month, day, hour, min, utcHour, utcMin);

  return units;
}


char * makeInitializationTime(cnTFC * fc)
{
  static char initTime[100];
  int         year;
  int         month;
  int         day;
  int         hour;
  int         min;
  int         utcHour;
  int         utcMin;
  int         date    = utDateFromMins(fc->firstTime);
  int         time    = utTimeFromMins(fc->firstTime);

  utExplodeDate(date, &year, &month, &day);
  utExplodeTime(time, &hour, &min);
  utExplodeTime(fc->utcOffset, &utcHour, &utcMin);
  if ( utcMin < 0 )
  {
    utcMin -= utcMin;
  }

  sprintf(initTime, "%04d-%02d-%02d %02d:%02d:00 %+03d:%02d",
    year, month, day, hour, min, utcHour, utcMin);

  return initTime;
}


void printDims(FILE * outfile, cnTFC * fc)
{
  fprintf(outfile,   "dimensions:\n"
                     "  time       = %d;\n", fc->numProfiles);

  if ( fc->numLevelVars > 0 )
  {
    fprintf(outfile, "  level      = %d;\n", fc->numLevels);
  }

  if ( fc->numFluxVars > 0 )
  {
    fprintf(outfile, "  flux_level = %d;\n", fc->numLevels + 1);
  }
}


void printVars(FILE * outfile, cnTFC * fc)
{
  int var;

  fprintf(outfile,
    "\nvariables:\n"
    "  float latitude;\n"
    "    latitude:units          = \"degrees_N\";\n"
    "    latitude:long_name      = \"Latitude of model gridpoint\";\n"
    "    latitude:standard_name  = \"latitude\";\n"
  );
  fprintf(outfile, 
    "  float longitude;\n"
    "    longitude:units         = \"degrees_E\";\n"
    "    longitude:long_name     = \"Longitude of model gridpoint\";\n"
    "    longitude:standard_name = \"longitude\";\n"
  );
  fprintf(outfile,
    "  float horizontal_resolution;\n"
    "    horizontal_resolution:units     = \"km\";\n"
    "    horizontal_resolution:long_name = \"Horizontal resolution of "
      "model\";\n"
  );
  fprintf(outfile,
    "  float time(time);\n"
    "    time:units              = \"%s\";\n"
    "    time:long_name          = \"Hours UTC\";\n"
    "    time:standard_name      = \"time\";\n"
    "    time:axis               = \"T\";\n"
    ,
    makeTimeUnits(fc)
  );
  fprintf(outfile,
    "  float forecast_time(time);\n"
    "    forecast_time:units     = \"hours\";\n"
    "    forecast_time:long_name = \"Time since initialization of "
      "forecast\";\n"
    "    forecast_time:comments  = \"For each profile in the file this"
      " variable contains the time elapsed since the initialization "
      "time of the forecast from which it was taken. Note that the "
      "profiles in this file may be taken from more than one "
      "forecast.\";\n"
  );

  if ( fc->numLevelVars > 0 )
  {
    fprintf(outfile,
      "  short level(level);\n"
      "    level:units             = \"1\";\n"
      "    level:long_name         = \"Model level\";\n"
      "    level:standard_name     = \"model_level_number\";\n"
      "    level:axis              = \"Z\";\n"
      "    level:positive          = \"down\";\n"
    );
  }

  if ( fc->numFluxVars > 0 )
  {
    fprintf(outfile,
      "  short flux_level(flux_level);\n"
      "    flux_level:units        = \"1\";\n"
      "    flux_level:long_name    = \"Model flux level\";\n"
      "    flux_level:axis         = \"Z\";\n"
      "    flux_level:positive     = \"down\";\n"
    );
  }


#define M(i, a, b) if ( ncVariables[i].a != NULL )                     \
  fprintf(outfile, "    %s:" b " \"%s\";\n",                           \
    ncVariables[i].name, ncVariables[i].a)
#define MM(i)                                                          \
  M(i, units,        "units         =");                               \
  M(i, longName,     "long_name     =");                               \
  M(i, standardName, "standard_name =");                               \
  M(i, comment,      "comment       =");                               \
  fprintf(outfile, "    %s:C_format      = \"%%.%df\";\n",             \
    ncVariables[i].name, ncVariables[i].precision);                    \
  fprintf(outfile, "    %s:_FillValue    = %.2fF;\n",                  \
    ncVariables[i].name, cnFcInitialValue);                            \
  fprintf(outfile, "    %s:missing_value = %.2fF;\n",                  \
    ncVariables[i].name, cnFcInitialValue)


  /* The levels vars. */
  for ( var = 0; var < fc->numLevelVars; var++ )
  {
    fprintf(outfile, "  float %s(time, level);\n",
      ncVariables[fc->levelIds[var]].name);
    MM(fc->levelIds[var]);
  }

  /* The flux vars. */
  for ( var = 0; var < fc->numFluxVars; var++ )
  {
    fprintf(outfile, "  float %s(time, flux_level);\n",
      ncVariables[fc->fluxIds[var]].name);
    MM(fc->fluxIds[var]);
  }

  /* The surface vars. */
  for ( var = 0; var < fc->numSurfaceVars; var++ )
  {
    fprintf(outfile, "  float %s(time);\n",
      ncVariables[fc->surfaceIds[var]].name);
    MM(fc->surfaceIds[var]);
  }


#undef M
#undef MM
}


void printGlobalAtts(FILE * outfile, cnTFC * fc, const char * historyLine)
{
  char * location = NULL;
  char * title    = NULL;

  location = utMakeStr(cnGetSiteName(fc->site, 'n'));
  utMixCase(location);
  title    = utMakeStr(ncTitles[fc->model]);
  title    = utAppendStr(title, utMakeStr(" over "), utMakeStr(location),
    NULL);

  fprintf(outfile,
    "\n// Global attributes:\n\n"
    "  :Conventions         = \"%s\";\n"
    "  :title               = \"%s\";\n"
    "  :location            = \"%s\";\n"
    "  :source              = \"%s\";\n"
    "  :institution         = \"%s\";\n"
    "  :initialization_time = \"%s\";\n"
    "  :history             = \"%s\";\n"
    ,
    NC_CONVENTIONS,
    title,
    location,
    ncSources[fc->model],
    ncModelInstitutions[fc->model], 
    makeInitializationTime(fc),
    historyLine
  );

  free(location);
  free(title);
}


void printLineData(FILE * outfile, float * data, int n, int precision,
  int lineLen)
{
  int i;

  for ( i = 0; i < n; i++ )
  {
    fprintf(outfile, "%.*f", precision, data[i]);
    if ( i < (n - 1) )
    {
      fprintf(outfile, ", ");
    }
    if ( ((i + 1) % lineLen) == 0 && i < (n - 1) )
    {
      fprintf(outfile, "\n  ");
    }
  }
}


void printData(FILE * outfile, cnTFC * fc)
{
  static float data[CN_NC_MAX_LEVEL_DIM + 1];
  const int    lineLen = 8;

  int          var;
  int          prof;
  int          lev;
  float        time;
  float        step;


  fprintf(outfile,
    "\ndata:\n"
    "  latitude              = %.2f;\n\n"
    "  longitude             = %.2f;\n\n"
    "  horizontal_resolution = %.1f;\n\n"
    ,
    fc->coords.lat, fc->coords.lon, fc->hres
  );


  /* 'time' and 'forecast_time' are floats with units of hours. */
  fprintf(outfile, "  time = ");
  time = (float) utToDecTime(utTimeFromMins(fc->firstTime));
  step = (float) fc->step / 60.0F;
  for ( prof = 0; prof < fc->numProfiles; prof++, time += step )
  {
    fprintf(outfile, "%.6f", time);
    if ( prof < (fc->numProfiles - 1) )
    {
      fprintf(outfile, ", ");
    }
  }
  fprintf(outfile, ";\n\n");


  /* 'forecast_time' is the number of hours since initialisation. */
  fprintf(outfile, "  forecast_time = ");
  for ( prof = 0; prof < fc->numProfiles; prof++ )
  {
    time = (float) fc->profiles[prof].forecastTime;
    if ( time >= 0.0F )
    {
      time /= 60.0F;
    }

    fprintf(outfile, "%.6f", time);

    if ( prof < (fc->numProfiles - 1) )
    {
      fprintf(outfile, ", ");
    }
  }
  fprintf(outfile, ";\n\n");


  if ( fc->numLevelVars > 0 )
  {
    fprintf(outfile, "  level = ");
    for ( lev = fc->numLevels; lev >= 1; lev-- )
    {
      fprintf(outfile, "%d", lev);
      if ( lev > 1 )
      {
        fprintf(outfile, ", ");
      }
    }
    fprintf(outfile, ";\n\n");
  }


  if ( fc->numFluxVars > 0 )
  {
    fprintf(outfile, "  flux_level = ");
    for ( lev = fc->numLevels + 1; lev >= 1; lev-- )
    {
      fprintf(outfile, "%d", lev);
      if ( lev > 1 )
      {
        fprintf(outfile, ", ");
      }
    }
    fprintf(outfile, ";\n\n");
  }


  /* The levels vars. */
  for ( var = 0; var < fc->numLevelVars; var++ )
  {
    int i = fc->levelIds[var];

    fprintf(outfile, "  %s =\n", ncVariables[i].name);

    for ( prof = 0; prof < fc->numProfiles; prof++ )
    {
      fprintf(outfile, "\n  // %s, profile: %d\n",
        ncVariables[i].name, prof);
      cnFcGetVarProfile(fc, data, prof, var);
      fprintf(outfile, "  ");
      printLineData(outfile, data, fc->numLevels,
        ncVariables[i].precision, lineLen);
      if ( prof < (fc->numProfiles - 1) )
      {
        fprintf(outfile, ",");
      }
      fprintf(outfile, "\n");
    }
    fprintf(outfile, ";\n\n");
  }


  /* The flux vars. */
  for ( var = 0; var < fc->numFluxVars; var++ )
  {
    int i = fc->fluxIds[var];

    fprintf(outfile, "  %s =\n", ncVariables[i].name);

    for ( prof = 0; prof < fc->numProfiles; prof++ )
    {
      fprintf(outfile, "\n  // %s, profile: %d\n",
        ncVariables[i].name, prof);
      cnFcGetFluxProfile(fc, data, prof, var);
      fprintf(outfile, "  ");
      printLineData(outfile, data, fc->numLevels + 1,
        ncVariables[i].precision, lineLen);
      if ( prof < (fc->numProfiles - 1) )
      {
        fprintf(outfile, ",");
      }
      fprintf(outfile, "\n");
    }
    fprintf(outfile, ";\n\n");
  }


  /* The surface vars. */
  for ( var = 0; var < fc->numSurfaceVars; var++ )
  {
    int i = fc->surfaceIds[var];

    fprintf(outfile, "  %s = ", ncVariables[i].name);

    cnFcGetSurfaceVar(fc, data, var);
    printLineData(outfile, data, fc->numProfiles,
      ncVariables[i].precision, lineLen);
    fprintf(outfile, ";\n\n");
  }
}


void printCdl(FILE * outfile, cnTFC * fc, const char * historyLine)
{
assert(outfile != NULL);
assert(fc != NULL);
assert(historyLine != NULL);


  /**********************************************************************/

  fprintf(outfile, "netcdf %s {\n\n", makeDatasetName(fc));

  /**********************************************************************/

  printDims(outfile, fc);
  printVars(outfile, fc);
  printGlobalAtts(outfile, fc, historyLine);
  printData(outfile, fc);

  /**********************************************************************/

  /* Finishing off. */
  fprintf(outfile, "}\n");
}


/************************************************************************/
/************************************************************************/


/* Set the 'fc->numLevels' if it hasn't already been set. */
void getLevels(cnTFC * fc, cnTIM * im, char * filename)
{
  if ( im->ga.fileType != CN_FT_SFC )
  {
    int numLevels = im->dims.level;
    if ( im->ga.fileType == CN_FT_FLX )
    {
      /* 'dims.level' will be for the number of fluxes.
       * 'numLevels' is the number of levels in the model.
       */
      numLevels--;
    }

    if ( fc->numLevels <= 0 )
    {
      fc->numLevels = numLevels;
    }
    else
    if ( fc->numLevels != numLevels )
    {
      printError(PE_EXIT, "%s: the number of levels are not the same as"
        " in previous files", filename);
    }
  }
}


void getVarIds(cnTFC * fc, cnTIM * im, char * filename)
{
  int var;


#define DO(nv, max, ids, srch, msg)                                     \
if ( fc->numLevelVars == 0 ) { /* Add all variables. */                 \
  fc->nv = im->vars.numVars;                                            \
  if ( fc->nv > max ) printError(PE_EXIT, "%s: %s", msg, filename);     \
  for ( var = 0; var < fc->nv; var++ ) {                                \
    fc->ids[var] = im->vars.vars[var].varId;                            \
  }                                                                     \
}                                                                       \
else { /* Add any new variables that there may be. */                   \
  for ( var = 0; var < im->vars.numVars; var++ ) {                      \
    if ( srch(fc, im->vars.vars[var].varId) < 0 ) {                     \
      fc->nv++;                                                         \
      if ( fc->nv > max ) printError(PE_EXIT, "%s: %s", msg, filename); \
      fc->ids[fc->nv - 1] = im->vars.vars[var].varId;                   \
    }                                                                   \
  }                                                                     \
}


  switch ( im->ga.fileType )
  {
    case CN_FT_VAR:
      DO(numLevelVars, CN_NC_MAX_LEVEL_VARS, levelIds,
        cnFcSearchLevelIds, "too many level variables");
      break;

    case CN_FT_FLX:
      DO(numFluxVars, CN_NC_MAX_FLUX_VARS, fluxIds,
        cnFcSearchFluxIds, "too many flux variables");
      break;

    case CN_FT_SFC:
      DO(numSurfaceVars, CN_NC_MAX_SFC_VARS, surfaceIds,
        cnFcSearchSurfaceIds, "too many surface variables");
      break;

    default:
      break;
  }


#undef DO
}


void readFiles(cnTFC * fc)
{
  int fileNum;
  /* The time of the last profile in the dataset. */
  int lastTime = 0;

  /**********************************************************************/

  fc->numProfiles = cl.numProfiles;
  fc->step        = utTimeInMins(cl.step);

  /* If the start date has been given then use it here otherwise get it
   * from the first model data file.
   */
  if ( cl.startDate >= 0 )
  {
    fc->firstTime = utToMins(cl.startDate, cl.startTime);
    lastTime      = cnFcGetProfileTime(fc, fc->numProfiles - 1);
  }

  /**********************************************************************/

  for ( fileNum = 0; fileNum < cl.numModelFiles; fileNum++ )
  {
    static cnTIM im;
    char *       filename = cl.modelFiles[fileNum];

    cnImRead(filename, &im);

    /* If the start date has not been given then use the start date of the
     * first file.
     */
    if ( cl.startDate < 0 )
    {
      cl.startDate  = utDateFromMins(im.ga.initTime + im.vars.time[0]);
      fc->firstTime = utToMins(cl.startDate, cl.startTime);
      lastTime      = cnFcGetProfileTime(fc, fc->numProfiles - 1);
    }

    /* Check if there is any data in the model data that covers the
     * same period as the data set.
     * If the model's first time or last time is within the range of
     * the FC dataset's times then there is some overlap.
     */
    if ( (im.ga.initTime + im.vars.time[0] <= lastTime)
      && (im.ga.initTime + im.vars.time[im.dims.time - 1]
        >= fc->firstTime) )
    {
      int prof;
      int lev;

      /* Check if this is the first valid data file that has been read. */
      if ( fc->model < 0 )
      {
        fc->site       = im.ga.site;
        fc->model      = im.ga.model;
        fc->utcOffset  = im.ga.utcOffset;
        fc->hres       = ncHorizontalResolutions[fc->model][fc->site];
	fc->coords.lat = im.vars.latitude;
	fc->coords.lon = im.vars.longitude;
      }
      else
      if ( fc->model != im.ga.model )
      {
        printError(PE_EXIT, "%s: not the same model as in previous files",
          filename);
      }
      else
	/*      if ( fc->coords.lat != im.vars.latitude
		|| fc->coords.lon != im.vars.longitude )*/
      if ( abs( fc->coords.lat - im.vars.latitude ) > 0.1
        || abs( fc->coords.lon - im.vars.longitude ) > 0.1 )
      {
        printError(PE_EXIT, "%s: not the same grid point as in previous"
          " files", filename);
      }

      getLevels(fc, &im, filename);
      getVarIds(fc, &im, filename);

      /* We want to copy the relevant data from
       *   'im.vars.vars[].data[time][level]'
       * to
       *   'fc->profiles[time].levels[level][]'
       *   or 'fc->profiles[time].fluxes[level][]'
       *   or 'fc->profiles[time].surface[]'
       */

      for ( prof = 0; prof < im.dims.time; prof++ )
      {
        /* 'prof' is the index into 'im.vars.vars[i].data[prof][]'
         * and 'im.vars.time[prof]'.
	 *
         * Does this profile match a profile in 'fc'?
         * The forecast time for this profile is 'im.vars.time[prof]'.
         * The absolute time for this profile is
         * 'im.ga.initTime' + forecast time.
         */
        int forecastTime = im.vars.time[prof];
        int fcProf       = cnFcSearch(fc, im.ga.initTime + forecastTime);

        if ( fcProf >= 0 && forecastTime >= cl.forecastMin*60 && forecastTime <= cl.forecastMax*60)
        {
          /* 'fcProf' is the index into 'fc->profiles[]'. */
          int i;

          fc->profiles[fcProf].forecastTime = forecastTime;

          for ( i = 0; i < im.vars.numVars; i++ )
          {
            int j;

            /* The search functions should always return j>=0 if
             * 'getVarIds()' worked properly.
             */
            switch ( im.ga.fileType )
            {
              case CN_FT_VAR:
                j = cnFcSearchLevelIds(fc, im.vars.vars[i].varId);
assert(j >= 0);
                /* 'j' is the index into
                 *   'fc->profiles[fcProf].levels[][j]'.
                 * 'lev' is the index into
                 *   'im.vars.vars[i].data[prof][lev]'
                 *   and 'fc->profiles[fcProf].levels[lev][j]'.
                 */

                for ( lev = 0; lev < im.dims.level; lev++ )
                {
                  fc->profiles[fcProf].levels[lev][j] =
                    im.vars.vars[i].data[prof][lev];
                }
                break;

              case CN_FT_FLX:
                j = cnFcSearchFluxIds(fc, im.vars.vars[i].varId);
assert(j >= 0);
                for ( lev = 0; lev < im.dims.level; lev++ )
                {
                  fc->profiles[fcProf].fluxes[lev][j] =
                    im.vars.vars[i].data[prof][lev];
                }
                break;

              case CN_FT_SFC:
                j = cnFcSearchSurfaceIds(fc, im.vars.vars[i].varId);
assert(j >= 0);
                fc->profiles[fcProf].surface[j] =
                  im.vars.vars[i].data[prof][0];
                break;

              default:
                break;
            } /* switch ( im.ga.fileType ) */
          } /* for ( i = 0; i < im.numVars; i++ ) */
        } /* if ( fcProf >= 0 ) */
      } /* for ( prof = 0; prof < im.dims.time; prof++ ) */
    } /* if time of 'im' data is within time for 'fc'. */
    else
    {
      printError(0, "%s: no data for the time period required ",
        filename);
    }
  }
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  /* This contains the complete dataset. */
  static cnTFC fc;

  /* The file to write the CDL output to. */
  FILE *       outfile = stdout;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */
  cl.startDate      = -1;
  cl.startTime      = DEF_START_TIME;
  cl.step           = DEF_STEP;
  cl.numProfiles    = DEF_NUM_PROFILES;
  cl.user           = NULL;
  cl.modelFiles     = NULL;
  cl.numModelFiles  = 0;
  cl.outFilename    = NULL;
  cl.forecastMin    = 0;
  cl.forecastMax    = 32000;

  /**********************************************************************/

  readCommandLine(argc, argv);

  /**********************************************************************/

  cnFcInit(&fc);
  readFiles(&fc);

  if ( cnMCheckModel(fc.model) )
  {
    char * historyLine;

    /* Want the levels running upwards. */
    if ( fc.numLevels > 0 && cnFcUp(&fc, -1) == 0 )
    {
      cnFcReverseLevels(&fc);
    }

    if ( cl.outFilename != NULL )
    {
      outfile = fopen(cl.outFilename, "w");
      if ( outfile == NULL )
      {
        printError(PE_EXIT,
          "unable to open output file \"%s\"", cl.outFilename);
      }
    }

    historyLine = ncMakeHistoryLine(cl.user, argc, argv);
    printCdl(outfile, &fc, historyLine);
    free(historyLine);

    if ( cl.outFilename != NULL )
    {
      fclose(outfile);
    }
  }
  else
  {
    printError(0, "no valid data has been read");
  }

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/
