/*
 * CloudNET Model Conversion Software
 *
 * Calculate cloud fraction.
 *
 * HISTORY
 * 2004-03-30 ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnapp.h"
#include "cnfcread.h"
#include "version.h"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS = 1

};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{



  /* Index into 'argv[]' of the first argument after the options. */
  int firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file...\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Calculate the total cloud fraction from the large scale and"
    " convection cloud fractions.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file...' are FC netCDF datasets.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;


      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = argn;
}


/************************************************************************/
/************************************************************************/


void processFile(char * filename, char * historyLine)
{
  static char * comment = "This is the total cloud fraction, calculated "
    "from the large_scale_cloud_fraction and "
    "convective_cloud_fraction held in the model using: "
    "cloud_fraction=convective_cloud_fraction+(1-"
    "convective_cloud_fraction)*large_scale_cloud_fraction";

  static cnTFCread fc;

  int              prof;

  /**********************************************************************/

  cnFcOpen(filename, &fc);

  cnFcVarNotExists(&fc, filename, CN_VAR_TOTAL_CLOUD_FRACTION);
  cnFcVarExists(&fc, filename, CN_VAR_CONV_CLOUD_FRACTION);
  cnFcVarExists(&fc, filename, CN_VAR_LS_CLOUD_FRACTION);

  ncVariables[CN_VAR_TOTAL_CLOUD_FRACTION].comment = comment;

  /* Create the new variable. */
  cnFcAddVar(&fc, CN_VAR_TOTAL_CLOUD_FRACTION);

  for ( prof = 0; prof < fc.dims.time; prof++ )
  {
    static float cf[CN_NC_MAX_LEVEL_DIM];
    static float ccf[CN_NC_MAX_LEVEL_DIM];
    static float lscf[CN_NC_MAX_LEVEL_DIM];

    int          lev;

    cnFcReadVarProfile(&fc, CN_VAR_CONV_CLOUD_FRACTION, prof, ccf);
    cnFcReadVarProfile(&fc, CN_VAR_LS_CLOUD_FRACTION, prof, lscf);

    for ( lev = 0; lev < fc.dims.level; lev++ )
    {
      float c  = ccf[lev];
      float ls = lscf[lev];
      if ( ls >= 0.0 && c >= 0.0 )
      {
        cf[lev] = c + (1.0F - c) * ls;
      }
      else
      {
        cf[lev] = cnFcInitialValue;
      }
    }

    cnFcWriteVarProfile(&fc, CN_VAR_TOTAL_CLOUD_FRACTION, prof, cf);
  }

  /**********************************************************************/

  cnFcAddGlobalAtt(&fc, "history", historyLine);
  NCEXIT(nc_close(fc.id));
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  char * historyLine;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */

  cl.firstArg  = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);

  historyLine = ncMakeHistoryLine(NULL, argc, argv);
  for ( ; cl.firstArg < argc; cl.firstArg++ )
  {
    processFile(argv[cl.firstArg], historyLine);
  }
  free(historyLine);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

