/*
 * CloudNET Model Conversion Software.
 * Calculate fluxes.
 *
 * HISTORY
 * 2004-05-17  ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnapp.h"
#include "cnfcread.h"
#include "lsp_ice.h"
#include "version.h"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS = 1
};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{
  int firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file...\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Calculate the large-scale rainfall and large-scale snowfall"
    " fluxes.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file...' are FC netCDF datasets.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;


      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = argn;
}


/************************************************************************/
/************************************************************************/


void checkFluxes(cnTFCread * fc, char * filename)
{
#define TEST(v) if ( fc->vars.id[v] >= 0 ) \
    printError(PE_EXIT, "%s: %s is already defined", \
      filename, ncVariables[v].name)


  TEST(CN_FLX_LS_RAIN);
  TEST(CN_FLX_LS_SNOW);


#undef TEST
}


void checkVars(cnTFCread * fc, char * filename)
{
#define TEST(v) if ( fc->vars.id[v] < 0 ) \
    printError(PE_EXIT, "%s: %s is not available", \
      filename, ncVariables[v].name)


  TEST(CN_VAR_PRESSURE);
  TEST(CN_VAR_TEMPERATURE);
  TEST(CN_VAR_HEIGHT);
  TEST(CN_VAR_Q);
  TEST(CN_VAR_QL);
  TEST(CN_VAR_QI);
  TEST(CN_VAR_LS_CLOUD_FRACTION);


#undef TEST
}


void checkFluxLevel(cnTFCread * fc)
{
  if ( fc->dimIds.fluxLevel < 0 )
  {
    char * units    = "1";
    char * longName = "Model flux level";
    char * axis     = "Z";
    char * positive = "down";
    int    id;
    int    i;

    fc->dims.fluxLevel   = fc->dims.level + 1;
    fc->dimIds.fluxLevel = cnFcAddDim(fc, "flux_level",
      fc->dims.fluxLevel);

    nc_redef(fc->id);
    NCEXIT(nc_def_var(fc->id, "flux_level", NC_SHORT, 1,
      &fc->dimIds.fluxLevel, &id));

    NCEXIT(nc_put_att_text(fc->id, id, "units", strlen(units), units));
    NCEXIT(nc_put_att_text(fc->id, id, "long_name", strlen(longName),
      longName));
    NCEXIT(nc_put_att_text(fc->id, id, "axis", strlen(axis), axis));
    NCEXIT(nc_put_att_text(fc->id, id, "positive", strlen(positive),
      positive));
    NCEXIT(nc_enddef(fc->id));

    for ( i = 0; i < fc->dims.fluxLevel; i++ )
    {
      fc->vars.fluxLevel[i] = fc->dims.fluxLevel - i;
    }
    NCEXIT(nc_put_var_short(fc->id, id, fc->vars.fluxLevel));
  }
}


void processFile(char * filename, char * historyLine)
{
  static cnTFCread fc;
  static float     rhcrit[CN_NC_MAX_LEVEL_DIM] = {
    0.916F, 0.908F, 0.891F, 0.884F, 0.875F, 0.861F, 0.857F, 0.854F
  };

  float   dt;
  int     prof;
  int     lev;
  int     numIterations;
  LOGICAL bland = 1;

  LSP_INIT();

  cnFcOpen(filename, &fc);

  checkFluxes(&fc, filename);
  checkVars(&fc, filename);
  checkFluxLevel(&fc);

  cnFcAddGlobalAtt(&fc, "history", historyLine);

  cnFcAddVar(&fc, CN_FLX_LS_RAIN);
  cnFcAddVar(&fc, CN_FLX_LS_SNOW);


  for ( lev = 8; lev < fc.dims.level; lev++ )
  {
    rhcrit[lev] = 0.85F;
  }

  /* Get the timestep in seconds. */
  /* dt            = (fc.vars.time[1] - fc.vars.time[0]) * 3600.0F;*/
  /* numIterations = (int) (dt / 600.0F);*/

  dt            = 60.0F * 60.0F;
  numIterations = 1;

  for ( prof = 0; prof < fc.dims.time; prof++ )
  {
    static float dz[CN_NC_MAX_LEVEL_DIM];
    static float r[CN_NC_MAX_LEVEL_DIM];
    static float s[CN_NC_MAX_LEVEL_DIM];
    static float p[CN_NC_MAX_LEVEL_DIM];
    static float t[CN_NC_MAX_LEVEL_DIM];
    static float h[CN_NC_MAX_LEVEL_DIM];
    static float q[CN_NC_MAX_LEVEL_DIM];
    static float ql[CN_NC_MAX_LEVEL_DIM];
    static float qi[CN_NC_MAX_LEVEL_DIM];
    static float lscf[CN_NC_MAX_LEVEL_DIM];
    static float cf[CN_NC_MAX_LEVEL_DIM];
    static float cfl[CN_NC_MAX_LEVEL_DIM];
    static float cff[CN_NC_MAX_LEVEL_DIM];
    static float qcf[CN_NC_MAX_LEVEL_DIM];
    static float qcl[CN_NC_MAX_LEVEL_DIM];

    float vf = 0.0F;
    float rf = 0.0F;
    float sf = 0.0F;

    cnFcReadVarProfile(&fc, CN_VAR_PRESSURE, prof, p);
    cnFcReadVarProfile(&fc, CN_VAR_TEMPERATURE, prof, t);
    cnFcReadVarProfile(&fc, CN_VAR_HEIGHT, prof, h);
    cnFcReadVarProfile(&fc, CN_VAR_Q, prof, q);
    cnFcReadVarProfile(&fc, CN_VAR_QL, prof, ql);
    cnFcReadVarProfile(&fc, CN_VAR_QI, prof, qi);
    cnFcReadVarProfile(&fc, CN_VAR_LS_CLOUD_FRACTION, prof, lscf);

    /* Calculate the thickness of each layer 'dz[]'.
     * This has to be done from the bottom up.
     */
    dz[0] = ( h[0] >= 0.0F ) ? h[0] * 2.0F : cnFcInitialValue;
    for ( lev = 1; lev < fc.dims.level; lev++ )
    {
      dz[lev] = ( h[lev] >= 0.0F && h[lev - 1] >= 0.0F
        && dz[lev - 1] > 0.0F )
        ? 2.0F * (h[lev] - h[lev - 1]) - dz[lev - 1] : cnFcInitialValue;
    }

    /* This alters t and q so may want to re-read them from the file. */
    if ( LS_CLD(p, rhcrit, fc.dims.level, t, cf, q, qcf, qcl, cfl, cff)
      < 0 )
    {
      printError(PE_EXIT, "malloc() failed in LS_CLD()\n");
    }

    /* Calculate the fluxes at the bottom of each layer. This has to be
     * done from the top down.
     */
    for ( lev = fc.dims.level - 1; lev >= 0; lev-- )
    {
      if ( p[lev] >= 0.0F && t[lev] >= 0.0F
        && dz[lev] >= 0.0F && q[lev] >= 0.0F
        && ql[lev] >= 0.0F && qi[lev] >= 0.0F
        && lscf[lev] >= 0.0F
      )
      {
        LSP_ICE(p[lev], dz[lev], dt, rhcrit[lev], qi[lev], ql[lev],
          q[lev], &rf, &sf, &vf, t[lev], cfl[lev], cff[lev], bland,
          numIterations);
        r[lev] = ( rf < 1E-10 ) ? 0.0 : rf;
        s[lev] = ( sf < 1E-10 ) ? 0.0 : sf;
      }
      else
      {
        r[lev] = cnFcInitialValue;
        s[lev] = cnFcInitialValue;
      }
    }

    cnFcWriteVarProfile(&fc, CN_FLX_LS_RAIN, prof, r);
    cnFcWriteVarProfile(&fc, CN_FLX_LS_SNOW, prof, s);
  } /* for each profile. */

  NCEXIT(nc_close(fc.id));
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  char * historyLine;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */

  cl.firstArg = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);

  historyLine = ncMakeHistoryLine(NULL, argc, argv);
  for ( ; cl.firstArg < argc; cl.firstArg++ )
  {
    processFile(argv[cl.firstArg], historyLine);
  }
  free(historyLine);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

