/*
 * CloudNET Model Conversion Software
 *
 * Calculate the level heights.
 *
 * HISTORY
 * 2004-03-30 ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnapp.h"
#include "cnfcread.h"
#include "version.h"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS = 1

};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{
  float sigma0;
  int   calcSigma0;

  /* Index into 'argv[]' of the first argument after the options. */
  int   firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file...\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Calculate the heights using pressure, temperature and specific"
    " humidity.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
  );
  fprintf(outfile,
    "\n"
    "  -s sigma0\n"
    "    The value for sigma0."
    " sigma0 is the ratio of the pressure at the first level above the"
    " surface to the pressure at the surface."
    " Default: 0.998812\n"
  );
  fprintf(outfile,
    "\n"
    "  -S\n"
    "    Calculate sigma0 using the first levels' pressure and the"
    " surface pressure.\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file...' are FC netCDF datasets.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;

      case 's':
        cl.sigma0 = atof(getOptArg(option, argc, argv, &argn));
        break;

      case 'S':
        cl.calcSigma0 = 1;
        break;

      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = argn;
}


/************************************************************************/
/************************************************************************/


void processFile(char * filename, char * historyLine, float sigma0,
  int calcSigma0)
{
  static char * comment = "The heights have been calculated using"
    " pressure, temperature and specific humidity.";

  const double     Rgas          = 287.0; /* Gas constant */
  const double     g             = 9.81;  /* gravity */
  const double     epsilon       = 0.622;
  const double     epsilonFactor = 1.0 - 1.0 / epsilon;

  static float     sfcPressure[CN_NC_MAX_TIME_DIM];
  static cnTFCread fc;

  int              prof;

  /**********************************************************************/

  cnFcOpen(filename, &fc);

  cnFcVarNotExists(&fc, filename, CN_VAR_HEIGHT);
  cnFcVarExists(&fc, filename, CN_VAR_PRESSURE);
  cnFcVarExists(&fc, filename, CN_VAR_TEMPERATURE);
  cnFcVarExists(&fc, filename, CN_VAR_Q);

  /* Need the surface pressure if sigma0 is to be calculated. */
  if ( calcSigma0 )
  {
    if( fc.vars.id[CN_SFC_PRESSURE] < 0 )
    {
      printError(PE_EXIT,
        "%s: unable to calculate sigma0 because %s is not available",
        filename, ncVariables[CN_SFC_PRESSURE].name);
    }
    else
    {
      cnFcReadSfc(&fc, CN_SFC_PRESSURE, sfcPressure);
    }
  }

  ncVariables[CN_VAR_HEIGHT].comment = comment;


  /* Create the new variable. */
  cnFcAddVar(&fc, CN_VAR_HEIGHT);

  for ( prof = 0; prof < fc.dims.time; prof++ )
  {
    static float height[CN_NC_MAX_LEVEL_DIM];
    static float pressure[CN_NC_MAX_LEVEL_DIM];
    static float temp[CN_NC_MAX_LEVEL_DIM];
    static float specificHumidity[CN_NC_MAX_LEVEL_DIM];

    int          lev;
    int          valid;

    cnFcReadVarProfile(&fc, CN_VAR_PRESSURE, prof, pressure);
    cnFcReadVarProfile(&fc, CN_VAR_TEMPERATURE, prof, temp);
    cnFcReadVarProfile(&fc, CN_VAR_Q, prof, specificHumidity);

    valid = 1;
    for ( lev = 0; lev < fc.dims.level; lev++ )
    {
      valid = pressure[lev] > cnFcInitialValue
        && temp[lev] > 0.0F
        && specificHumidity[lev] >= 0.0F;
    }

    if ( !valid )
    {
      for ( lev = 0; lev < fc.dims.level; lev++ )
      {
	height[lev] = cnFcInitialValue;  
      }
    }
    else
    {
      double T;
      double Q;

      lev = 0;
      T   = temp[lev];
      Q   = specificHumidity[lev];

      if ( calcSigma0 && sfcPressure[prof] > cnFcInitialValue )
      {
        sigma0 = pressure[0] / sfcPressure[prof];
      }

      height[lev] = Rgas * (1.0 / sigma0 - 1.0) * T
        / (g * (1.0 + epsilonFactor * Q));

      for ( lev = 1; lev < fc.dims.level; lev++ )
      {
        double H1 = height[lev - 1];
        double P1 = pressure[lev - 1];
        double T1 = temp[lev - 1];
        double Q1 = specificHumidity[lev - 1];
        double P  = pressure[lev];
        T         = temp[lev];
        Q         = specificHumidity[lev];

        height[lev] = H1 + Rgas * (P1 - P) * (T + T1)
          / (g * (P + P1) * (1.0 + epsilonFactor * (Q1 + Q) * 0.5));
      }
    }
    cnFcWriteVarProfile(&fc, CN_VAR_HEIGHT, prof, height);
  }

  /**********************************************************************/

  cnFcAddGlobalAtt(&fc, "history", historyLine);
  NCEXIT(nc_close(fc.id));
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  char * historyLine;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */
  cl.sigma0     = 0.998812F;
  cl.calcSigma0 = 0;
  cl.firstArg   = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);

  historyLine = ncMakeHistoryLine(NULL, argc, argv);
  for ( ; cl.firstArg < argc; cl.firstArg++ )
  {
    processFile(argv[cl.firstArg], historyLine, cl.sigma0, cl.calcSigma0);
  }
  free(historyLine);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

