/*
 * CloudNET Model Conversion Software
 *
 * Calculate MWPP for a FC dataset.
 *
 * HISTORY
 * 2004-03-29  ao
 *
 * 2004-05-14  ao  frequencies given on the command line now override the
 * defaults instead of appending to them.
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnapp.h"
#include "cnfcread.h"
#include "cnmwpp.h"
#include "version.h"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS = 1

};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{
  float frequencies[CN_NC_MAX_FREQ_DIM];
  int   numFreq;
  float sigma0;
  int   calcSigma0;

  /* Index into 'argv[]' of the first argument after the options. */
  int   firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file...\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Calculate the microwave propagation parameters.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
  );
  fprintf(outfile,
    "\n"
    "  -F frequency\n"
    "    Frequency (GHz) to calculate the parameters for."
    " This option may appear up to a maximum of %d times."
    " The default is to calculate parameters for 35 and 94 GHz.\n"
    "    Note: any occurrences of this option will replace"
    " the default settings.\n"
    ,
    CN_NC_MAX_FREQ_DIM
  );
  fprintf(outfile,
    "\n"
    "  -s sigma0\n"
    "    The value for sigma0."
    " sigma0 is the ratio of the pressure at the first level above the"
    " surface to the pressure at the surface."
    " Default: 0.998812\n"
  );
  fprintf(outfile,
    "\n"
    "  -S\n"
    "    Calculate sigma0 using the first levels' pressure and the"
    " surface pressure.\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file...' are FC netCDF datasets.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int fUsed = 0; /* Flag for option -f being used. */
  int argn  = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.numFreq = 0;

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;

      case 'F':
        if ( cl.numFreq >= CN_NC_MAX_FREQ_DIM )
        {
          printError(PE_PRINT_SYNOPSIS | PE_EXIT, "too many frequencies");
        }

        if ( !fUsed )
        {
          cl.numFreq = 0;
          fUsed      = 1;
        }

        cl.frequencies[cl.numFreq] = atof(getOptArg(option, argc, argv,
          &argn));
        cl.numFreq++;
        break;

      case 's':
        cl.sigma0 = atof(getOptArg(option, argc, argv, &argn));
        break;

      case 'S':
        cl.calcSigma0 = 1;
        break;

      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /**********************************************************************/

  /* If no frequencies have been added on the command line use the
   * defaults.
   */
  if ( cl.numFreq == 0 )
  {
    cl.frequencies[0] = 35.0F;
    cl.frequencies[1] = 94.0F;
    cl.numFreq        = 2;
  }

  cl.firstArg = argn;
}


/************************************************************************/
/************************************************************************/


void processFile(char * filename, char * historyLine,
  float * frequencies, int numFreq, float sigma0, int calcSigma0)
{
  static float     sfcPressure[CN_NC_MAX_TIME_DIM];
  static cnTFCread fc;
  static char *    freqUnits     = "GHz";
  static char *    freqLongName  = "Microwave frequency";
  static char *    freqCformat   = "%.6f";
  static float     freqFillValue = -999.0F;
  int              freq;
  int              varId;
  int              i;

  /**********************************************************************/

  cnFcOpen(filename, &fc);

  if ( fc.dims.frequency > 0 )
  {
    printError(PE_EXIT, "%s: frequency dimension is already defined",
      filename);
  }

  /* Need the height, temperature, pressure and specific humidity for
   * each level.
   */
  cnFcVarExists(&fc, filename, CN_VAR_HEIGHT);
  cnFcVarExists(&fc, filename, CN_VAR_TEMPERATURE);
  cnFcVarExists(&fc, filename, CN_VAR_PRESSURE);
  cnFcVarExists(&fc, filename, CN_VAR_Q);

  /* Need the surface pressure if sigma0 is to be calculated. */
  if ( calcSigma0 && fc.vars.id[CN_SFC_PRESSURE] < 0 )
  {
    printError(PE_EXIT,
      "%s: unable to calculate sigma0 because %s is not available",
      filename, ncVariables[CN_SFC_PRESSURE].name);
  }

  /**********************************************************************/

  /* Define the frequency dimension and the frequency variable and
   * the MWPP variables.
   */
  fc.dims.frequency   = numFreq;
  fc.dimIds.frequency = cnFcAddDim(&fc, "frequency", (size_t) numFreq);
  for ( freq = 0; freq < numFreq; freq++ )
  {
    fc.vars.frequency[freq] = frequencies[freq];
  }

  nc_redef(fc.id);
  NCEXIT(nc_def_var(fc.id, "frequency", NC_FLOAT, 1,
    &fc.dimIds.frequency, &varId));

  NCEXIT(nc_put_att_text(fc.id, varId, "units",
    strlen(freqUnits), freqUnits));
  NCEXIT(nc_put_att_text(fc.id, varId, "long_name",
    strlen(freqLongName), freqLongName));
  NCEXIT(nc_put_att_text(fc.id, varId, "C_format",
    strlen(freqCformat), freqCformat));
  NCEXIT(nc_put_att_float(fc.id, varId, "_FillValue", NC_FLOAT,
    1, &freqFillValue));
  NCEXIT(nc_put_att_float(fc.id, varId, "missing_value", NC_FLOAT,
    1, &freqFillValue));

  for ( i = CN_MWP_FIRST; i < CN_MWP_LAST; i++ )
  {
    cnFcAddVar(&fc, i);
  }

  nc_enddef(fc.id);


  if ( calcSigma0 )
  {
    cnFcReadSfc(&fc, CN_SFC_PRESSURE, sfcPressure);
  }


  /* Write the data. */
  NCEXIT(nc_put_var_float(fc.id, varId, frequencies));

  for ( freq = 0; freq < numFreq; freq++ )
  {
    float frequency = frequencies[freq];
    int   prof;
    for ( prof = 0; prof < fc.dims.time; prof++ )
    {
      /* Variables required. */
      static float height[CN_NC_MAX_LEVEL_DIM];
      static float pressure[CN_NC_MAX_LEVEL_DIM];
      static float temp[CN_NC_MAX_LEVEL_DIM];
      static float specificHumidity[CN_NC_MAX_LEVEL_DIM];

      /* Variables calculated. */
      static float gasAtten[CN_NC_MAX_LEVEL_DIM];
      static float specificGasAtten[CN_NC_MAX_LEVEL_DIM];
      static float specificSaturatedGasAtten[CN_NC_MAX_LEVEL_DIM];
      static float specificDryGasAtten[CN_NC_MAX_LEVEL_DIM];
      static float k2[CN_NC_MAX_LEVEL_DIM];
      static float specificLiquidAtten[CN_NC_MAX_LEVEL_DIM];

      /******************************************************************/

      cnFcReadVarProfile(&fc, CN_VAR_HEIGHT, prof, height);
      cnFcReadVarProfile(&fc, CN_VAR_PRESSURE, prof, pressure);
      cnFcReadVarProfile(&fc, CN_VAR_TEMPERATURE, prof, temp);
      cnFcReadVarProfile(&fc, CN_VAR_Q, prof, specificHumidity);

      if ( calcSigma0 && pressure[0] > cnFcInitialValue
        && sfcPressure[prof] > cnFcInitialValue )
      {
        sigma0 = pressure[0] / sfcPressure[prof];
      }

      /* Do saturated first, use 'gasAtten' as a dummy. */
      getAttenuationProfile(height, pressure, temp, specificHumidity,
        fc.dims.level, sigma0, frequency, 1, 0,
        gasAtten, specificSaturatedGasAtten);

      /* Do dry air. */
      getAttenuationProfile(height, pressure, temp, specificHumidity,
        fc.dims.level, sigma0, frequency, 0, 1,
        gasAtten, specificDryGasAtten);

      getAttenuationProfile(height, pressure, temp, specificHumidity,
        fc.dims.level, sigma0, frequency, 0, 0,
        gasAtten, specificGasAtten);

      getLiquidProfile(temp, fc.dims.level, frequency,
        k2, specificLiquidAtten);

      cnFcWriteMwppProfile(&fc, CN_MWP_GAS_ATTEN, freq, prof, gasAtten);
      cnFcWriteMwppProfile(&fc, CN_MWP_SPEC_GAS_ATTEN, freq, prof,
        specificGasAtten);
      cnFcWriteMwppProfile(&fc, CN_MWP_SPEC_SAT_GAS_ATTEN, freq, prof,
        specificSaturatedGasAtten);
      cnFcWriteMwppProfile(&fc, CN_MWP_SPEC_DRY_GAS_ATTEN, freq, prof,
        specificDryGasAtten);
      cnFcWriteMwppProfile(&fc, CN_MWP_K2, freq, prof, k2);
      cnFcWriteMwppProfile(&fc, CN_MWP_SPEC_LIQ_ATTEN, freq, prof,
        specificLiquidAtten);
    }
  }

  /**********************************************************************/

  cnFcAddGlobalAtt(&fc, "history", historyLine);
  NCEXIT(nc_close(fc.id));
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  char * historyLine;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here.
   * The frequencies will be set by 'readCommandLine()'.
   */

  cl.sigma0     = 0.998812F;
  cl.calcSigma0 = 0;
  cl.firstArg   = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);

  historyLine = ncMakeHistoryLine(NULL, argc, argv);
  for ( ; cl.firstArg < argc; cl.firstArg++ )
  {
    processFile(argv[cl.firstArg], historyLine,
      cl.frequencies, cl.numFreq, cl.sigma0, cl.calcSigma0);
  }
  free(historyLine);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

