/*
 * CloudNET Model Conversion Software
 *
 * Calculate the relative humidity.
 *
 * HISTORY
 * 2004-03-30 ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnapp.h"
#include "cnfcread.h"
#include "version.h"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS = 1

};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{



  /* Index into 'argv[]' of the first argument after the options. */
  int firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file...\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Calculate the relative humidity using pressure, temperature and"
    " specific humidity.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file...' are FC netCDF datasets.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;


      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = argn;
}


/************************************************************************/
/************************************************************************/


void processFile(char * filename, char * historyLine)
{
  static char * comment = "With respect to liquid above 0 degrees C and "
    "with respect to ice below 0 degrees C. "
    "Calculated using Goff-Gratch formula.";

  static cnTFCread fc;

  int              prof;

  /**********************************************************************/

  cnFcOpen(filename, &fc);

  cnFcVarNotExists(&fc, filename, CN_VAR_RH);
  cnFcVarExists(&fc, filename, CN_VAR_PRESSURE);
  cnFcVarExists(&fc, filename, CN_VAR_TEMPERATURE);
  cnFcVarExists(&fc, filename, CN_VAR_Q);

  ncVariables[CN_VAR_RH].comment = comment;

  /* Create the new variable. */
  cnFcAddVar(&fc, CN_VAR_RH);

  for ( prof = 0; prof < fc.dims.time; prof++ )
  {
    static float rh[CN_NC_MAX_LEVEL_DIM];
    static float pressure[CN_NC_MAX_LEVEL_DIM];
    static float temp[CN_NC_MAX_LEVEL_DIM];
    static float specificHumidity[CN_NC_MAX_LEVEL_DIM];

    int          lev;
    int          valid;

    cnFcReadVarProfile(&fc, CN_VAR_PRESSURE, prof, pressure);
    cnFcReadVarProfile(&fc, CN_VAR_TEMPERATURE, prof, temp);
    cnFcReadVarProfile(&fc, CN_VAR_Q, prof, specificHumidity);

    valid = 1;
    for ( lev = 0; lev < fc.dims.level; lev++ )
    {
      valid = pressure[lev] > cnFcInitialValue
        && temp[lev] > 0.0F
        && specificHumidity[lev] >= 0.0F;
    }

    if ( !valid )
    {
      for ( lev = 0; lev < fc.dims.level; lev++ )
      {
	rh[lev] = cnFcInitialValue;  
      }
    }
    else
    {
      for ( lev = 0; lev < fc.dims.level; lev++ )
      {
        const double T0 = 273.16;

        double P = pressure[lev];
        double T = temp[lev];
        double Q = specificHumidity[lev];

        /* calculate the saturated vapour pressure. */
        double svp;
        double mr;

        if ( T < T0 )
        {
          /* In ice */
          svp = 611.0 * exp((21.874 * (T - T0)) / (T - 7.66));
        }
        else
        {
          double r1 = T / T0;
          double r2 = T0 / T;
          svp = pow(10.0, 10.79574 * (1.0 - r2) - 5.028 * log10(r1)
            + 1.50475e-4 * (1.0 - pow(10.0, - 8.2969 * (r1 - 1.0)))
            + 0.42873e-3 * pow(10.0, 4.76955 * (1.0 - r2))
            + 0.78614 + 2.0
          );
        }

        /* calculate the mixing ratio. */
        mr = 0.62198 * (svp / (P - svp));

        rh[lev] = Q / mr;
      }
    }

    cnFcWriteVarProfile(&fc, CN_VAR_RH, prof, rh);
  }

  /**********************************************************************/

  cnFcAddGlobalAtt(&fc, "history", historyLine);
  NCEXIT(nc_close(fc.id));
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  char * historyLine;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */

  cl.firstArg  = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);

  historyLine = ncMakeHistoryLine(NULL, argc, argv);
  for ( ; cl.firstArg < argc; cl.firstArg++ )
  {
    processFile(argv[cl.firstArg], historyLine);
  }
  free(historyLine);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

