/*
 * CloudNET Model Conversion Software
 *
 * Calculate vertical velocity.
 *
 * HISTORY
 * 2004-03-30 ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include "cnapp.h"
#include "cnfcread.h"
#include "version.h"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS = 1

};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{



  /* Index into 'argv[]' of the first argument after the options. */
  int firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file...\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Calculate the vertical wind using omega.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file...' are FC netCDF datasets.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;


      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = argn;
}


/************************************************************************/
/************************************************************************/


void processFile(char * filename, char * historyLine)
{
  static char * comment = "The vertical wind has been calculated from "
    "omega (Pa s-1), height and pressure using: w=omega*dz/dp";

  static float     sfcPressure[CN_NC_MAX_TIME_DIM];
  static cnTFCread fc;

  int              prof;

  /**********************************************************************/

  cnFcOpen(filename, &fc);

  cnFcVarNotExists(&fc, filename, CN_VAR_WIND_W);
  cnFcVarExists(&fc, filename, CN_VAR_OMEGA);
  cnFcVarExists(&fc, filename, CN_VAR_HEIGHT);
  cnFcVarExists(&fc, filename, CN_VAR_PRESSURE);

  ncVariables[CN_VAR_WIND_W].comment = comment;

  /* Create the new variable. */
  cnFcAddVar(&fc, CN_VAR_WIND_W);

  /* If the surface pressure is available read it. */
  if ( fc.vars.id[CN_SFC_PRESSURE] >= 0 )
  {
    cnFcReadSfc(&fc, CN_SFC_PRESSURE, sfcPressure);
  }

  for ( prof = 0; prof < fc.dims.time; prof++ )
  {
    static float w[CN_NC_MAX_LEVEL_DIM];
    static float omega[CN_NC_MAX_LEVEL_DIM];
    static float height[CN_NC_MAX_LEVEL_DIM];
    static float pressure[CN_NC_MAX_LEVEL_DIM];

    int          lev;
    int          valid;

    cnFcReadVarProfile(&fc, CN_VAR_OMEGA, prof, omega);
    cnFcReadVarProfile(&fc, CN_VAR_HEIGHT, prof, height);
    cnFcReadVarProfile(&fc, CN_VAR_PRESSURE, prof, pressure);

    valid = 1;
    for ( lev = 0; lev < fc.dims.level; lev++ )
    {
      valid = omega[lev] > cnFcInitialValue
        && height[lev] > cnFcInitialValue
        && pressure[lev] > cnFcInitialValue;
    }

    if ( !valid )
    {
      for ( lev = 0; lev < fc.dims.level; lev++ )
      {
        w[lev] = cnFcInitialValue;
      }
    }
    else
    {
      /* If the surface pressure is available use it to calculate the
       * first value.
       */
      if ( fc.vars.id[CN_SFC_PRESSURE] >= 0 )
      {
        float dz = height[0];
        float dp = pressure[1] - sfcPressure[prof];
        w[0]     = (dz / dp) * omega[0];
      }
      else
      {
        w[0] = cnFcInitialValue;
      }

      for ( lev = 1; lev < fc.dims.level; lev++ )
      {
        float dz = height[lev] - height[lev - 1];
        float dp = pressure[lev] - pressure[lev - 1];
        w[lev]   = (dz / dp) * omega[lev];
      }
    }

    cnFcWriteVarProfile(&fc, CN_VAR_WIND_W, prof, w);
  }

  /**********************************************************************/

  cnFcAddGlobalAtt(&fc, "history", historyLine);
  NCEXIT(nc_close(fc.id));
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  char * historyLine;

  /**********************************************************************/

  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */

  cl.firstArg  = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);

  historyLine = ncMakeHistoryLine(NULL, argc, argv);
  for ( ; cl.firstArg < argc; cl.firstArg++ )
  {
    processFile(argv[cl.firstArg], historyLine);
  }
  free(historyLine);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

