/*
 * CloudNET Model Conversion Software
 * Do all the stages.
 * I'd rather do this in a script but it's too complicated.
 *
 * HISTORY
 * 2004-04-08  ao
 *
 * 2004-05-11  ao  'stage2()' corrected 'executeCommand()' test.
 *
 * 2004-05-19  ao  lots of modifications.
 *
 * 2004-05-24  ao  added the '-r' option.
 *
 * 2004-05-28  ao  added the '-4-' option. In stage 4, added tests so
 *   programs are only run if the model requires it.
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "cnapp.h"
#include "cnmodels.h"
#include "cnsites.h"
#include "cnmwpp.h"
#include "version.h"
#include "getfileinfo.c"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS = 1,

  MAXFILES = 4096,

  DEF_START_TIME   = 0,
  DEF_NUM_PROFILES = 25,
  DEF_STEP         = 100,     /* HHMM */
  DEF_FORECAST_MIN = 0,
  DEF_FORECAST_MAX = 32000
};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Options and arguments from the command line are here. */
struct
{
  int   startDate;    /* YYYYMMDD */
  int   startTime;    /* HHMM */
  int   step;         /* HHMM */
  int   numProfiles;

  int   forecastMin;  /* minutes. */
  int   forecastMax;  /* minutes. */

  int   do4;
  float frequencies[CN_NC_MAX_FREQ_DIM];
  int   numFreq;

  int   recalc;     /* For flux heights in stage 4. */

  char  verbose;     /* Default - don't write to stdout */

  /* Index into 'argv[]' of the first argument after the options. */
  int   firstArg;
} cl;



/* These directory names must not have '/' as the last char. */

static char binDir[FILENAME_MAX + 1] = "./";


/* Temporary directory. */
static char tmpDir[FILENAME_MAX + 1] = "./tmp";


/* Directory for IM netCDF datasets. */
static char imDir[FILENAME_MAX + 1] = "./IM";


/* Directory for FC netCDF datasets. */
static char fcDir[FILENAME_MAX + 1] = "./FC";


/* Scratch array for creating command lines to be executed. */
static char cmdLine[4096];


/* Scratch array. */
static char filename[FILENAME_MAX + 1];


/* Filename for the forecast dataset. */
static char fcFilename[FILENAME_MAX + 1];


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file...\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
    "  Do all stages on the files given.\n"
    "  All the files must be for the same model, site and for only one"
    " days forecast.\n"
    "  All the cloudnet conversion programs must be available in a"
    " directory in the environment variable PATH unless the -bdir option"
    " is used.\n"
  );
  fprintf(outfile,
    "  The stages of conversion are:\n"
    "    1. Convert model output ASCII files into intermediate (IM)"
    " netCDF datasets.\n"
    "    2. Perform any required processing on the IM datasets.\n"
    "    3. Convert the IM datasets into a single forecast (FC) netCDF"
    " dataset.\n"
    "    4. Perform the following processing on the FC dataset:\n"
  );
  fprintf(outfile,
    "      for ECMWF and KNMI RACMO models: calculate the heights of"
    " each level using cnfcht;\n"
    "      for ECMWF model: calculate surface height using cnfcsfcht;\n"
    "      for KNMI RACMO and UK Met Office models: calculate fluxes"
    " using cnfcflx;\n"
    "      for all models: calculate flux heights using cnfcflxht;\n"
    "      for KNMI RACMO and UK Met Office Global models: calculate"
    " relative humidity using cnfcrh;\n"
  );
  fprintf(outfile,
    "      for UK Met Office models: calculate total cloud cover fraction"
    " using cnfccf;\n"
    "      for all models except UK Met Office Global model: calculate"
    " vertical wind velocity from omega using cnfcw;\n"
    "      for all models: calculate the microwave propagation parameters"
    " using cnfcmwpp.\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
  );
  fprintf(outfile,
    "\n"
    "  These options set directories used by the program.\n"
    "  If these directories do not already exist then they will be"
    " created.\n"
    "  NOTES:\n"
    "    1). Do not include the trailing slash '/' in the path.\n"
    "    2). Any directories created and files created within them will"
    " not be removed by the program.\n"
  );
  fprintf(outfile,
    "\n"
    "  -bdir directory\n"
    "    Directory with the cloudnet conversion programs. Default: %s\n"
    , binDir
  );
  fprintf(outfile,
    "\n"
    "  -fdir directory\n"
    "    Directory for the forecast (FC) netCDF file. Default: %s\n"
    , fcDir
  );
  fprintf(outfile,
    "\n"
    "  -idir directory\n"
    "    Directory for the intermediate (IM) netCDF files. Default: %s\n"
    , imDir
  );
  fprintf(outfile,
    "\n"
    "  -tdir directory\n"
    "    Scratch directory for various files. Default: %s\n"
    , tmpDir
  );
  fprintf(outfile,
    "\n"
    "  These options are used in stage 3.\n"
  );
  fprintf(outfile,
    "\n"
    "  -d date\n"
    "    Start date for the data set. Format is YYYYMMDD. This is"
    " optional and if it is not given then the first date of the"
    " first file read will be used.\n"
  );
  fprintf(outfile,
    "\n"
    "  -p numProfiles\n"
    "    The number of profiles in the data set. Default: %d.\n"
    ,
    DEF_NUM_PROFILES
  );
  fprintf(outfile,
    "\n"
    "  -s step\n"
    "    The time interval between each profile. Format is HHMM."
    " Default: %04d.\n"
    ,
    DEF_STEP
  );
  fprintf(outfile,
    "\n"
    "  -t time\n"
    "    Start time for the data set. Format is HHMM. Default: %04d.\n"
    "\n"
    "  -f min max\n"
    "    Only select profiles that have a forecast time in the range"
    " min..max. min and max are hours. They are the times since"
    " initialization of the model. The default is to disregard"
    " forecast times when selecting profiles.\n"
    ,
    DEF_START_TIME
  );
  fprintf(outfile,
    "\n"
    "  These options are used in stage 4.\n"
  );
  fprintf(outfile,
    "\n"
    "  -4-\n"
    "    don't do any stage 4 processing.\n"
  );
  fprintf(outfile,
    "\n"
    "  -F frequency\n"
    "    Frequency (GHz) to calculate the microwave propagation"
    " parameters for."
    " This option may appear up to a maximum of %d times."
    " The default is to calculate parameters for 35 and 94 GHz.\n"
    "    Note: any occurrences of this option will replace"
    " the default settings.\n"
    ,
    CN_NC_MAX_FREQ_DIM
  );
  fprintf(outfile,
    "\n"
    "  -r\n"
    "    If the flux heights already exist then recalculate. Default is"
    " to leave them alone.\n"
  );
  fprintf(outfile,
    "\n"
    "  -v\n"
    "    Verbose.\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file...' are model output files.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    if ( strcmp(option, "bdir") == 0 )
    {
      strcpy(binDir, getOptArg(option, argc, argv, &argn));
    }
    else
    if ( strcmp(option, "tdir") == 0 )
    {
      strcpy(tmpDir, getOptArg(option, argc, argv, &argn));
    }
    else
    if ( strcmp(option, "fdir") == 0 )
    {
      strcpy(fcDir, getOptArg(option, argc, argv, &argn));
    }
    else
    if ( strcmp(option, "idir") == 0 )
    {
      strcpy(imDir, getOptArg(option, argc, argv, &argn));
    }
    else
    if ( strcmp(option, "4-") == 0 )
    {
      cl.do4 = 0;
    }
    else
    {
      switch ( option[0] )
      {
        case 'h':
        case '?':
          printSynopsis(stdout, 1);
          break;

        case 'd':
          cl.startDate = atoi(getOptArg(option, argc, argv, &argn));
          break;

        case 'f':
          cl.forecastMin = atoi(getOptArg(option, argc, argv, &argn));
          cl.forecastMax = atoi(getOptArg(option, argc, argv, &argn));
          break;

        case 'p':
          cl.numProfiles = atoi(getOptArg(option, argc, argv, &argn));
          break;

        case 'r':
          cl.recalc = 1;
          break;

        case 's':
          cl.step = atoi(getOptArg(option, argc, argv, &argn));
          break;

        case 't':
          cl.startTime = atoi(getOptArg(option, argc, argv, &argn));
          break;

        case 'F':
          if ( cl.numFreq >= CN_NC_MAX_FREQ_DIM )
          {
            printError(PE_PRINT_SYNOPSIS | PE_EXIT,
              "too many frequencies");
          }

          cl.frequencies[cl.numFreq] = atof(getOptArg(option, argc, argv,
            &argn));
          cl.numFreq++;
          break;
      
        case 'v':
	  cl.verbose = 1;
	  break;
	
        default:
          printError(PE_PRINT_SYNOPSIS | PE_EXIT,
            "unrecognized option \"%s\"", option);
          break;
      }
    }
    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = argn;
}


/************************************************************************/
/************************************************************************/


/* Array of filenames. */
typedef struct
{
  char        name[MAXFILES][FILENAME_MAX + 1];
  cnTFileInfo info[MAXFILES];
  /* flag for each file. */
  int         valid[MAXFILES];

  /* number of names and info structures. */
  int         n;
} TFilenameList;


int model = -1;
int site  = -1;

/************************************************************************/


/* Get the file info on each file in the command line. */
void getFiles(int argc, char * argv[], int firstArg,
  TFilenameList * files/*OUT*/)
{
  int i;

  files->n = 0;
  for ( i = firstArg; i < argc; i++ )
  {
    cnTFileInfo info;

    getFileInfo(argv[i], &info);

    if ( info.model < 0 )
    {
      printError(PE_EXIT, "%s: unable to determine model source.\n",
        argv[i]);
    }

    if ( info.fileType < 0 && info.model != CN_ML_MOUMG )
    {
      printError(PE_EXIT, "%s: unable to determine file type.\n",
        argv[i]);
    }

    if ( info.site < 0 )
    {
      printError(PE_EXIT, "%s: unable to determine site.\n", argv[i]);
    }

    if ( info.initTime < 0 )
    {
      printError(PE_EXIT,
        "%s: unable to determine initialization time.\n", argv[i]);
    }

    if ( info.forecastTime < 0 )
    {
      printError(PE_EXIT, "%s: unable to determine forecast time.\n",
        argv[i]);
    }

    if ( files->n > 0 )
    {
      if ( info.model != files->info[files->n - 1].model )
      {
        printError(PE_EXIT,
          "%s: model source is different from previous file.\n", argv[i]);
      }

      if ( info.site != files->info[files->n - 1].site )
      {
        printError(PE_EXIT, "%s: site is different from previous file.\n",
          argv[i]);
      }
    }

    strcpy(files->name[files->n], argv[i]);
    files->info[files->n]  = info;
    files->valid[files->n] = 1;
    files->n++;
    model = info.model;
    site  = info.site;
  }

  /**********************************************************************/

  if ( files->n == 0 )
  {
    printError(PE_EXIT, "no valid files to process.");
  }
}


/* Create a directory if it doesn't already exist. */
void createDirectory(char * directory)
{
  if ( mkdir(directory, S_IRWXU) == -1 )
  {
    if ( errno != EEXIST )
    {
      printError(PE_EXIT, "unable to create directory: \"%s\"",
        directory);
    }
  }
}


/* Execute system command in 'cmdLine'.
 * POST: returns zero on success, non-zero on error.
 */
int executeCommand(void)
{
  static char filename[FILENAME_MAX + 1];

  int retValue;

  strcpy(filename, cmdLine);
  sprintf(cmdLine, "%s/%s", binDir, filename);
  if ( cl.verbose ) 
    printf("Executing system command: %s", cmdLine);
  retValue = system(cmdLine);
  if ( cl.verbose ) 
    printf("  Return value: %d ", retValue);
  if ( cl.verbose ) {
    if ( retValue == 0 )
      {
	printf("(success).\n");
      }
    else
      {
	printf("(failure).\n");
      }
  }
  return retValue;
}


int isSurfaceFile(int fileType)
{
  return ( fileType == CN_FT_SFC || fileType == FT_SFC2
    || fileType == FT_SFCE )
    ? 1 : 0;
}


int searchECIFSfiles(TFilenameList * in, int fileType, int site,
  int initTime, int forecastTime)
{
  int retValue = -1;
  int i;
  for ( i = 0; i < in->n; i++ )
  {
    if ( in->info[i].model        == CN_ML_ECIFS
      && in->info[i].fileType     == fileType
      && in->info[i].site         == site
      && in->info[i].initTime     == initTime
      && in->info[i].forecastTime == forecastTime
      && in->valid[i] != 0 )
    {
      retValue = i;
      break;
    }
  }

  return retValue;
}


/************************************************************************/


/* Preprocessing of stage 1. */


/* The ECIFS surface file concatenations.
 * Only call this if the model is ECIFS.
 */
void catECIFSsfcFiles(TFilenameList * files/*IN/OUT*/)
{
  int i;

  for ( i = 0; i < files->n; i++ )
  {
    if ( !files->valid[i] ) continue;

    if ( isSurfaceFile(files->info[i].fileType) )
    {
      int isfc = -1;
      int isf2 = -1;
      int isfe = -1;

      if ( cl.verbose ) 
	printf("\"%s\" - ECMWF surface file, searching for the others...",
	       files->name[i]);

      switch ( files->info[i].fileType )
      {
        case CN_FT_SFC:
          isf2 = searchECIFSfiles(files, FT_SFC2, files->info[i].site,
            files->info[i].initTime, files->info[i].forecastTime);
          if ( isf2 < 0 )
          {
            printf("Unable to find SFC2 file.\n");
            files->valid[i] = 0;
            continue;
          }

          isfe = searchECIFSfiles(files, FT_SFCE, files->info[i].site,
            files->info[i].initTime, files->info[i].forecastTime);
          if ( isfe < 0 )
          {
            printf("Unable to find SFCE file.\n");
            files->valid[i] = 0;
            continue;
          }

          isfc = i;
          break;

        case FT_SFC2:
          isfc = searchECIFSfiles(files, CN_FT_SFC, files->info[i].site,
            files->info[i].initTime, files->info[i].forecastTime);
          if ( isfc < 0 )
          {
            printf("Unable to find SFC file.\n");
            files->valid[i] = 0;
            continue;
          }

          isfe = searchECIFSfiles(files, FT_SFCE, files->info[i].site,
            files->info[i].initTime, files->info[i].forecastTime);
          if ( isfe < 0 )
          {
            printf("Unable to find SFCE file.\n");
            files->valid[i] = 0;
            continue;
          }

          isf2 = i;
          break;

        case FT_SFCE:
          isfc = searchECIFSfiles(files, CN_FT_SFC, files->info[i].site,
            files->info[i].initTime, files->info[i].forecastTime);
          if ( isfc < 0 )
          {
            printf("Unable to find SFC file.\n");
            files->valid[i] = 0;
            continue;
          }

          isf2 = searchECIFSfiles(files, FT_SFC2, files->info[i].site,
            files->info[i].initTime, files->info[i].forecastTime);
          if ( isf2 < 0 )
          {
            printf("Unable to find SFC2 file.\n");
            files->valid[i] = 0;
            continue;
          }

          isfe = i;
          break;

        default: break;
      }

      if ( isfc >= 0 && isf2 >= 0 && isfe >= 0 )
      {
	if ( cl.verbose ) 
	  printf("success.\n");

        /* We now have the indeces into 'files->name[]' for the three
         * files required for concatenating into one surface file.
         */
        /* Make the filename for the concatenated file. */
        sprintf(filename, "%s/ECIFS-%02d-%02d-%010d-%010d",
          tmpDir,
          CN_FT_SFC,
          files->info[isfc].site,
          files->info[isfc].initTime,
          files->info[isfc].forecastTime
        );

        sprintf(cmdLine, "cnecifscatsfc %s %s %s > %s\n",
          files->name[isfc], files->name[isf2], files->name[isfe],
          filename);


        if ( executeCommand() != 0 )
        {
          files->valid[i] = 0;
        }
        else
        {
          strcpy(files->name[i], filename);
          files->info[i].fileType = CN_FT_SFC;
        }

        if ( isfc != i ) files->valid[isfc] = 0;
        if ( isf2 != i ) files->valid[isf2] = 0;
        if ( isfe != i ) files->valid[isfe] = 0;
      }
    }
  }
}


/* Filtering of the KNRAC files.
 * Only call this if the model is KNRAC.
 */
void filterKNRACfiles(TFilenameList * files/*IN/OUT*/)
{
  int i;
  char fstem[5];
 
  for ( i = 0; i < files->n; i++ )
  {
    if ( !files->valid[i] ) continue;

    if ( cl.verbose ) 
      printf("Filtering \"%s\" to extract the relevant grid point.\n",
	     files->name[i]);


    if ( files->info[i].model == CN_ML_SMRCA )
      sprintf(fstem, "SMRCA");
    else 
      sprintf(fstem, "KNRAC");

      /* Make the filename for the filtered file. */
    sprintf(filename, "%s/%s-%02d-%02d-%010d-%010d",
      tmpDir, fstem,
      files->info[i].fileType,
      files->info[i].site,
      files->info[i].initTime,
      files->info[i].forecastTime
    );
    sprintf(cmdLine, "cnknracf %s %s\n", files->name[i], filename);

    if ( executeCommand() != 0 )
    {
      files->valid[i] = 0;
    }
    else
    {
      /* Set the new filename. */
      strcpy(files->name[i], filename);
    }
  }
}


/************************************************************************/


/* Do stage 1. */
void stage1(TFilenameList * files/*IN/OUT*/)
{
  int n;
  int i;

  /* For UKMO MOLTS files 'files->n' can be incremented so must use
   * 'n' before it is changed.
   */
  n = files->n;
  for ( i = 0; i < n; i++ )
  {
    static char progName[FILENAME_MAX + 1];

    if ( !files->valid[i] ) continue;

    /* Special case if it is a Met Office MOLTS file. */
    if ( files->info[i].model == CN_ML_MOUMG )
    {
      sprintf(filename, "%s/%02d-%02d-%02d-%010d-%010d.nc",
        imDir,
        files->info[i].model,
        CN_FT_VAR,
        files->info[i].site,
        files->info[i].initTime,
        files->info[i].forecastTime
      );

      sprintf(cmdLine, "cnmoumgvar %s | ncgen -o %s\n",
        files->name[i], filename);
      if ( executeCommand() == 0 )
      {
        /* Add the IM filename to the list of files. */
        strcpy(files->name[files->n], filename);
        files->info[files->n]          = files->info[i];
        files->info[files->n].fileType = CN_FT_VAR;
        files->valid[files->n]         = 1;
        files->n++;
      }

      sprintf(filename, "%s/%02d-%02d-%02d-%010d-%010d.nc",
        imDir,
        files->info[i].model,
        CN_FT_SFC,
        files->info[i].site,
        files->info[i].initTime,
        files->info[i].forecastTime
      );

      sprintf(cmdLine, "cnmoumgsfc %s | ncgen -o %s\n",
        files->name[i], filename);
      if ( executeCommand() == 0 )
      {
        /* Add the IM filename to the list of files. */
        strcpy(files->name[files->n], filename);
        files->info[files->n]          = files->info[i];
        files->info[files->n].fileType = CN_FT_VAR;
        files->valid[files->n]         = 1;
        files->n++;
      }
      files->valid[i] = 0;
      continue;
    }


    sprintf(filename, "%s/%02d-%02d-%02d-%010d-%010d.nc",
      imDir,
      files->info[i].model,
      files->info[i].fileType,
      files->info[i].site,
      files->info[i].initTime,
      files->info[i].forecastTime
    );

    progName[0] = '\0';
    switch ( files->info[i].model )
    {
      case CN_ML_ECIFS:
        switch ( files->info[i].fileType )
        {
          case CN_FT_VAR: strcpy(progName, "cnecifsvar"); break;
          case CN_FT_FLX: strcpy(progName, "cnecifsflx"); break;
          case CN_FT_SFC: strcpy(progName, "cnecifssfc"); break;
          default: break;
        }
        break;

      case CN_ML_MFARP:
        switch ( files->info[i].fileType )
        {
          case CN_FT_VAR: strcpy(progName, "cnmfarpvar"); break;
          case CN_FT_FLX: strcpy(progName, "cnmfarpflx"); break;
          case CN_FT_SFC: strcpy(progName, "cnmfarpsfc"); break;
          default: break;
        }
        break;

      case CN_ML_KNRAC:
        switch ( files->info[i].fileType )
        {
          case CN_FT_VAR: strcpy(progName, "cnknracvar"); break;
          case CN_FT_SFC: strcpy(progName, "cnknracsfc"); break;
          default: break;
        }
        break;      

      case CN_ML_SMRCA:
        switch ( files->info[i].fileType )
        {
          case CN_FT_VAR: strcpy(progName, "cnsmrcavar"); break;
          case CN_FT_SFC: strcpy(progName, "cnsmrcasfc"); break;
          default: break;
        }
        break;

      case CN_ML_MOUMM:
      case CN_ML_MOUNA:
      case CN_ML_MOUHR:
        switch ( files->info[i].fileType )
        {
          case CN_FT_VAR: strcpy(progName, "cnmoummvar"); break;
          case CN_FT_SFC: strcpy(progName, "cnmoummsfc"); break;
          default: break;
        }
        break;

      case CN_ML_DWDLM:
        switch ( files->info[i].fileType )
        {
          case CN_FT_VAR: strcpy(progName, "cndwdlmvar"); break;
          case CN_FT_SFC: strcpy(progName, "cndwdlmsfc"); break;
          default: break;
        }
        break;

      default: break;
    }

    if ( strlen(progName) > 0 )
    {
      sprintf(cmdLine, "%s %s | ncgen -o %s\n", progName, files->name[i],
        filename);
      if ( executeCommand() != 0 )
      {
        files->valid[i] = 0;
      }
      else
      {
        strcpy(files->name[i], filename);
      }
    }
    else
    {
      files->valid[i] = 0;
    }
  }
}


/************************************************************************/


/* Do stage 2. */
void stage2(TFilenameList * files/*IN/OUT*/)
{
  int i;

  for ( i = 0; i < files->n; i++ )
  {
    if ( !files->valid[i] ) continue;

    if ( files->info[i].model == CN_ML_MFARP
      && files->info[i].fileType == CN_FT_VAR )
    {
      sprintf(cmdLine, "cnimsettimes %s\n", files->name[i]);
      if ( executeCommand() != 0 )
      {
        files->valid[i] = 0;
      }
    }
  }
}


/************************************************************************/


/* Do stage 3. */
void stage3(TFilenameList * files/*IN/OUT*/)
{
  static char * modelNames[] = 
  {
    "MOUMM",
    "MOUMG",
    "ECIFS",
    "MFARP",
    "KNHIR",
    "KNRAC",
    "SMRCA",
    "DWDLM",
    "MOUNA",
    "MOUHR"
  };

  char   siteName[32];
  char * p;
  int    i;
  int    startDate = -1;
  int    startTime = DEF_START_TIME;

  /**********************************************************************/

  sprintf(cmdLine, "cnfcmk ");

  p = &cmdLine[strlen(cmdLine)];

  if ( cl.startDate > 0 )
  {
    sprintf(p, "-d %d ", cl.startDate);
    p = &cmdLine[strlen(cmdLine)];

    startDate = cl.startDate;
  }

  if ( cl.startTime != DEF_START_TIME )
  {
    sprintf(p, "-t %d ", cl.startTime);
    p = &cmdLine[strlen(cmdLine)];

    startTime = cl.startTime;
  }

  if ( cl.step != DEF_STEP )
  {
    sprintf(p, "-s %d ", cl.step);
    p = &cmdLine[strlen(cmdLine)];
  }

  if ( cl.numProfiles != DEF_NUM_PROFILES )
  {
    sprintf(p, "-p %d ", cl.numProfiles);
    p = &cmdLine[strlen(cmdLine)];
  }

  if ( cl.forecastMin != DEF_FORECAST_MIN
    || cl.forecastMax != DEF_FORECAST_MAX )
  {
    sprintf(p, "-f %d %d ", cl.forecastMin, cl.forecastMax);
    p = &cmdLine[strlen(cmdLine)];
  }

  /**********************************************************************/

  /* get first valid file. */
  for ( i = 0; i < files->n && !files->valid[i]; i++ )
  {
  }

  /* 'i' is now the index of the first valid file. */

  /**********************************************************************/

  /* Create the filename for the forecast dataset. */
  strcpy(siteName, cnGetSiteName(files->info[i].site, 'l'));
  if ( startDate < 0 )
  {
    startDate = utDateFromMins(files->info[i].forecastTime);
  }

  sprintf(fcFilename, "%s/%s_%08d_%04d_%s.nc",
    fcDir, modelNames[files->info[i].model],
    startDate, startTime, siteName
  );

  /**********************************************************************/

  /* Put all the valid files into the command line. */
  for ( i = 0; i < files->n; i++ )
  {
    if ( !files->valid[i] ) continue;

    sprintf(p, "%s ", files->name[i]);
    p = &cmdLine[strlen(cmdLine)];
  }

  sprintf(p, " | ncgen -o %s\n", fcFilename);
  if ( executeCommand() != 0 )
  {
    printError(PE_EXIT, "I can't do any more, sorry. Goodbye.\n");
  }
}


/************************************************************************/


void stage4(void)
{
  if ( model == CN_ML_ECIFS && (site == CN_SITE_T1 || site == CN_SITE_T2 || site == CN_SITE_T3 ))
  {
    if ( cl.verbose ) 
      printf("Calculate level heights.\n");
    sprintf(cmdLine, "cnfcht %s\n", fcFilename);
    executeCommand();
  }

  if ( (model == CN_ML_ECIFS  && site != CN_SITE_T1 && site != CN_SITE_T2 && site != CN_SITE_T3 ) || model == CN_ML_KNRAC )
  {
    if ( cl.verbose ) 
      printf("Calculate level heights.\n");
    sprintf(cmdLine, "cnfcht -S %s\n", fcFilename);
    executeCommand();
  }


  if ( model == CN_ML_ECIFS && site != CN_SITE_T1 && site != CN_SITE_T2 && site != CN_SITE_T3 )
  {
    if ( cl.verbose ) 
      printf("Calculate surface height.\n");
    sprintf(cmdLine, "cnfcsfcht %s\n", fcFilename);
    executeCommand();
  }


  /*  if ( model == CN_ML_KNRAC
      || model == CN_ML_MOUMM || model == CN_ML_MOUMG )*/
  if ( model == CN_ML_MOUMM || model == CN_ML_MOUNA || model == CN_ML_MOUHR || model == CN_ML_MOUMG )
  {
    if ( cl.verbose ) 
      printf("Calculate fluxes.\n");
    sprintf(cmdLine, "cnfcflx %s\n", fcFilename);
    executeCommand();
  }


  /*  if ( model == CN_ML_ECIFS || model == CN_ML_KNRAC
      || model == CN_ML_MOUMM || model == CN_ML_MOUMG )*/
  if ( model == CN_ML_ECIFS
    || model == CN_ML_MOUMM || model == CN_ML_MOUNA || model == CN_ML_MOUHR || model == CN_ML_MOUMG )
  {
    if ( cl.verbose ) 
      printf("Calculate flux heights.\n");
    if ( cl.recalc )
    {
      sprintf(cmdLine, "cnfcflxht -r %s\n", fcFilename);
    }
    else
    {
      sprintf(cmdLine, "cnfcflxht %s\n", fcFilename);
    }
    executeCommand();
  }


  if ( model == CN_ML_KNRAC || model == CN_ML_SMRCA || model == CN_ML_MOUMG )
  {
    if ( cl.verbose ) 
      printf("Calculate relative humidity.\n");
    sprintf(cmdLine, "cnfcrh %s\n", fcFilename);
    executeCommand();
  }


  if ( model == CN_ML_MOUMM || model == CN_ML_MOUNA || model == CN_ML_MOUHR || model == CN_ML_MOUMG )
  {
    if ( cl.verbose ) 
      printf("Calculate total cloud cover fraction.\n");
    sprintf(cmdLine, "cnfccf %s\n", fcFilename);
    executeCommand();
  }


  if ( model == CN_ML_ECIFS || model == CN_ML_MFARP
    || model == CN_ML_KNRAC || model == CN_ML_MOUMM || model == CN_ML_MOUNA || model == CN_ML_MOUHR )
  {
    if ( cl.verbose ) 
      printf("Calculate vertical wind velocity from omega.\n");
    sprintf(cmdLine, "cnfcw %s\n", fcFilename);
    executeCommand();
  }


  if ( cl.verbose ) 
    printf("Calculate the microwave propagation parameters.\n");
  if ( cl.numFreq > 0 )
  {
    char * p;
    int    i;

    if ( model == CN_ML_MOUMM && cl.startDate < 20020815 )
      sprintf(cmdLine, "cnfcmwpp ");
    else
      sprintf(cmdLine, "cnfcmwpp -S ");
    
    p = &cmdLine[strlen(cmdLine)];

    for ( i = 0; i < cl.numFreq; i++ )
    {
      sprintf(p, " -F %.2f ", cl.frequencies[i]);
      p = &cmdLine[strlen(cmdLine)];
    }

    sprintf(p, " %s\n", fcFilename);
  }
  else
  {
    if ( (model == CN_ML_MOUMM && cl.startDate < 20020815) || (model == CN_ML_ECIFS && (site == CN_SITE_T1 || site == CN_SITE_T2  || site == CN_SITE_T3 )) )
      sprintf(cmdLine, "cnfcmwpp %s\n", fcFilename);
    else
      sprintf(cmdLine, "cnfcmwpp -S %s\n", fcFilename);
  }
  executeCommand();
}


/************************************************************************/


void processFiles(int argc, char * argv[], int firstArg)
{
  static TFilenameList files;

  int i;

  /**********************************************************************/

  if ( (argc - firstArg) >= MAXFILES )
  {
    printError(PE_EXIT, "too many filenames, maximum is: %d", MAXFILES);
  }

  /**********************************************************************/

  if ( cl.verbose ) 
    printf("Scanning files...\n");
  getFiles(argc, argv, firstArg, &files);

  /**********************************************************************/

  if ( cl.verbose ) 
    printf("\nCreating directory \"%s\"...\n", tmpDir);
  createDirectory(tmpDir);

  /**********************************************************************/

  if ( model == CN_ML_ECIFS )
  {
    catECIFSsfcFiles(&files);
  }
  else
  if ( model == CN_ML_KNRAC )
  {
    filterKNRACfiles(&files);
  }

  /**********************************************************************/

  if ( cl.verbose )
    printf("\nCreating directory \"%s\"...\n", imDir);
  createDirectory(imDir);

  if ( cl.verbose )
    printf("\nStage 1 processing.\n");
  stage1(&files);

  /**********************************************************************/

  if ( files.info[0].model == CN_ML_MFARP )
  {
    if ( cl.verbose )
      printf("\nStage 2 processing.\n");
    stage2(&files);
  }

  /**********************************************************************/

  for ( i = 0; i < files.n && !files.valid[i]; i++ )
  {
  }

  if ( i == files.n )
  {
    printError(PE_EXIT, "no valid files to process.");
  }

  /**********************************************************************/

  if ( cl.verbose )
    printf("\nCreating directory \"%s\"...\n", fcDir);
  createDirectory(fcDir);

  if ( cl.verbose )
    printf("\nStage 3 processing.\n");
  stage3(&files);

  /**********************************************************************/

  if ( cl.do4 )
  {
    if ( cl.verbose )
      printf("\nStage 4 processing.\n");
    stage4();
  }

  /**********************************************************************/

  if ( cl.verbose )
    printf("\nThe forecast dataset \"%s\" is now complete.\n", fcFilename);
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */
  cl.startDate   = -1;
  cl.startTime   = DEF_START_TIME;
  cl.step        = DEF_STEP;
  cl.numProfiles = DEF_NUM_PROFILES;
  cl.forecastMin = DEF_FORECAST_MIN;
  cl.forecastMax = DEF_FORECAST_MAX;
  cl.do4         = 1;
  cl.numFreq     = 0;
  cl.recalc      = 0;
  cl.verbose     = 0;
  cl.firstArg    = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);
  processFiles(argc, argv, cl.firstArg);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

