/*
 * Perform various tests on a netCDF file.
 *
 * HISTORY
 * 2004-03-29 ao
 *
 * AUTHORS
 * Andrew Openshaw
 */


/************************************************************************/
/************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <netcdf.h>
#include "cnapp.h"
#include "cnnetcdf.h"
#include "version.h"


/************************************************************************/
/************************************************************************/


enum
{
  /* minimum number of arguments on command line after the options. */
  MINARGS = 1

};


/* Program name.
 * This is initialised in the 'main()' function.
 */
char * PROGNAME;


/* Command line options. */
enum
{
  DIM_NAME,
  VAR_NAME,
  ATT_NAME
};


/* Options and arguments from the command line are here. */
struct
{
  int    option;
  char * name;

  /* Index into 'argv[]' of the first argument after the options. */
  int firstArg;
} cl;


/************************************************************************/
/************************************************************************/


/* Print the program synopsis on 'outfile'.
 * If 'exitProg'!=0 then 'exit(EXIT_FAILURE)' is called.
 */
void printSynopsis(FILE * outfile, int exitProg)
{
  fprintf(outfile,
    "\n"
    "CloudNET Model Conversion Software\n"
    "\n"
    "NAME\n"
    "\n"
    "  %s [options] file\n"
    ,
    PROGNAME
  );
  fprintf(outfile,
    "\n"
    "DESCRIPTION\n"
    "\n"
  );
  fprintf(outfile,
    "\n"
    "OPTIONS\n"
    "\n"
    "  -h\n"
    "  -?\n"
    "    This help display.\n"
    "\n"
    "  -a name\n"
    "    Checks if the global attribute 'name' is in the file. "
    "Prints 'name' if it is in the file or nothing if it is not.\n"
    "\n"
    "  -d name\n"
    "    Checks if the dimension 'name' is in the file. "
    "Prints 'name' if it is in the file or nothing if it is not.\n"
    "\n"
    "  -v name\n"
    "    Checks if the variable 'name' is in the file. "
    "Prints 'name' if it is in the file or nothing if it is not.\n"
  );
  fprintf(outfile,
    "\n"
    "ARGUMENTS\n"
    "\n"
    "  'file' is a netCDF dataset.\n"
    "\n"
  );

  if ( exitProg )
  {
    exit(EXIT_FAILURE);
  }
}


/* Read the command line options and arguments. */
void readCommandLine(int argc, char * argv[])
{
  int argn = 1;

  if ( (argc - 1) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  /* Read options. */
  while ( argn < argc && argv[argn][0] == '-' )
  {
    /* 'option' is set to the command line option after the
     * leading hyphen.
     */
    char * option = &argv[argn][1];

    /* For single character options use 'switch ( option[0] )' and a
     * 'case' statement for each option.
     * For multi-character options use 'if ( strcmp(option, "xxx") == 0)'
     * for each option.
     *
     * The 'getOptArg()' function is declared in 'cnapp.h' and is called
     * as 'getOptArg(option, argc, argv, &argn)'.
     */
    switch ( option[0] )
    {
      case 'h':
      case '?':
        printSynopsis(stdout, 1);
        break;

     case 'a':
       cl.option = ATT_NAME;
       cl.name = getOptArg(option, argc, argv, &argn);
       break;

     case 'd':
       cl.option = DIM_NAME;
       cl.name = getOptArg(option, argc, argv, &argn);
       break;

     case 'v':
       cl.option = VAR_NAME;
       cl.name = getOptArg(option, argc, argv, &argn);
       break;

      default:
        printError(PE_PRINT_SYNOPSIS | PE_EXIT,
          "unrecognized option \"%s\"", option);
        break;
    }

    argn++;
  }

  /**********************************************************************/

  if ( (argc - argn) < MINARGS )
  {
    printError(PE_PRINT_SYNOPSIS | PE_EXIT, "not enough arguments");
  }

  cl.firstArg = argn;
}


/************************************************************************/
/************************************************************************/


void processFile(char * filename, int option, char * name)
{
  int fileId;
  int id;

  NCEXIT(nc_open(filename, NC_NOWRITE, &fileId));

  switch ( option )
  {
    case DIM_NAME:
      if ( nc_inq_dimid(fileId, name, &id) == NC_NOERR )
      {
        printf("%s\n", name);
      }
      break;

    case VAR_NAME:
      if ( nc_inq_varid(fileId, name, &id) == NC_NOERR )
      {
        printf("%s\n", name);
      }
      break;

    case ATT_NAME:
      if ( nc_inq_attid(fileId, NC_GLOBAL, name, &id) == NC_NOERR )
      {
        printf("%s\n", name);
      }
      break;

    default: break;
  }

  NCEXIT(nc_close(fileId));
}


/************************************************************************/
/************************************************************************/


int main(int argc, char * argv[])
{
  /* Set PROGNAME to the name of the program.
   * Remove any leading directory path.
   * Assumes UNIX directory separator.
   */
  PROGNAME = strrchr(argv[0], '/');
  PROGNAME = ( PROGNAME == NULL ) ? argv[0] : PROGNAME + 1;

  /**********************************************************************/

  /* Initialise 'cl' with default values here. */

  cl.option    = -1;
  cl.name      = NULL;
  cl.firstArg  = -1;

  /**********************************************************************/

  readCommandLine(argc, argv);
  processFile(argv[cl.firstArg], cl.option, cl.name);

  /**********************************************************************/

  return 0;
}


/************************************************************************/
/************************************************************************/

