

#include "cnfcread.h"


/************************************************************************/
/************************************************************************/


/* Returns the dimension ID or -1 if 'name' does not exist in the dataset.
 */
static int getDimId(int fileId, char * name)
{
  int i;
  int retValue = -1;

  if ( nc_inq_dimid(fileId, name, &i) == NC_NOERR )
  {
    retValue = i;
  }

  return retValue;
}


/* Returns the current dimension length or -1 if 'name' does not exist
 * in the dataset.
 */
static int getDimLen(int fileId, int dimId)
{
  int    retValue = -1;
  size_t len;

  if ( nc_inq_dimlen(fileId, dimId, &len) == NC_NOERR )
  {
    retValue = (int) len;
  }

  return retValue;
}


/* Read an entire variable.
 * Returns -1 if variable is not in the dataset otherwise returns 0.
 */
static int readVar(int fileId, char * name, void * p)
{
  int i;
  int retValue = -1;

  if ( nc_inq_varid(fileId, name, &i) == NC_NOERR )
  {
    nc_type t;
    NCEXIT(nc_inq_vartype(fileId, i, &t));

    switch ( t )
    {
      case NC_BYTE:
        NCEXIT(nc_get_var_uchar(fileId, i, (unsigned char *) p));
        break;

      case NC_CHAR:
        NCEXIT(nc_get_var_schar(fileId, i, (signed char *) p));
        break;

      case NC_SHORT:
        NCEXIT(nc_get_var_short(fileId, i, (short *) p));
        break;

      case NC_INT:
        NCEXIT(nc_get_var_int(fileId, i, (int *) p));
        break;

      case NC_FLOAT:
        NCEXIT(nc_get_var_float(fileId, i, (float *) p));
        break;

      default: break;
    }

    retValue = 0;
  }

  return retValue;
}


void cnFcOpen(char * filename, cnTFCread * fc/*OUT*/)
{
  int nVars;
  int i;

  NCEXIT(nc_open(filename, NC_WRITE, &fc->id));

  fc->dimIds.time      = getDimId(fc->id, "time");
  fc->dimIds.level     = getDimId(fc->id, "level");
  fc->dimIds.fluxLevel = getDimId(fc->id, "flux_level");
  fc->dimIds.frequency = getDimId(fc->id, "frequency");
  fc->dims.time        = getDimLen(fc->id, fc->dimIds.time);
  fc->dims.level       = getDimLen(fc->id, fc->dimIds.level);
  fc->dims.fluxLevel   = getDimLen(fc->id, fc->dimIds.fluxLevel);
  fc->dims.frequency   = getDimLen(fc->id, fc->dimIds.frequency);

  readVar(fc->id, "latitude", &fc->vars.latitude);
  readVar(fc->id, "longitude", &fc->vars.longitude);
  readVar(fc->id, "horizontal_resolution", &fc->vars.horRes);
  readVar(fc->id, "time", fc->vars.time);
  readVar(fc->id, "forecast_time", fc->vars.forecastTime);
  readVar(fc->id, "level", fc->vars.level);
  readVar(fc->id, "flux_level", fc->vars.fluxLevel);
  readVar(fc->id, "fequency", fc->vars.frequency);

  NCEXIT(nc_inq_nvars(fc->id, &nVars));

  for ( i = 0; i < CN_VARIABLE_ID_COUNT; i++ )
  {
    fc->vars.id[i] = -1;
  }

  for ( i = 0; i < nVars; i++ )
  {
    /* For each variable in the dataset, get the variable name and check
     * if it is a model parameter variable.
     */
    static char name[NC_MAX_NAME + 1];
    int         id;   /* one of 'cnEVariableIds' */

    NCEXIT(nc_inq_varname(fc->id, i, name));
    id = ncSearchNames(name);
    if ( id >= 0 )
    {
      fc->vars.id[id] = i;
    }
  }
}


void cnFcVarExists(cnTFCread * fc, char * filename, int varId)
{
  if ( fc->vars.id[varId] < 0 )
  {
    printError(PE_EXIT, "%s: the %s variable is not available",
      filename, ncVariables[varId].name);
  }
}


void cnFcVarNotExists(cnTFCread * fc, char * filename, int varId)
{
  if ( fc->vars.id[varId] >= 0 )
  {
    printError(PE_EXIT, "%s: the %s variable is already available",
      filename, ncVariables[varId].name);
  }
}


int cnFcAddDim(cnTFCread * fc, char * name, size_t len)
{
  int retValue = -1;

  nc_redef(fc->id);
  NCEXIT(nc_def_dim(fc->id, name, len, &retValue));
  NCEXIT(nc_enddef(fc->id));

  return retValue;
}


void cnFcAddVar(cnTFCread * fc, int id)
{
  char cformat[10];
  int  dimIds[3];
  int  n;

  if ( cnMCheckRange(CN_VAR_FIRST, CN_VAR_LAST, id) )
  {
    dimIds[0] = fc->dimIds.time;
    dimIds[1] = fc->dimIds.level;
    n = 2;
  }
  else
  if ( cnMCheckRange(CN_FLX_FIRST, CN_FLX_LAST, id) )
  {
    dimIds[0] = fc->dimIds.time;
    dimIds[1] = fc->dimIds.fluxLevel;
    n = 2;
  }
  else
  if ( cnMCheckRange(CN_SFC_FIRST, CN_SFC_LAST, id) )
  {
    dimIds[0] = fc->dimIds.time;
    n = 1;
  }
  else
  if ( cnMCheckRange(CN_MWP_FIRST, CN_MWP_LAST, id) )
  {
    dimIds[0] = fc->dimIds.frequency;
    dimIds[1] = fc->dimIds.time;
    dimIds[2] = fc->dimIds.level;
    n = 3;
  }

  nc_redef(fc->id);
  NCEXIT(nc_def_var(fc->id, ncVariables[id].name, NC_FLOAT, n, dimIds,
    &fc->vars.id[id]));

  if ( ncVariables[id].units != NULL )
  {
    NCEXIT(nc_put_att_text(fc->id, fc->vars.id[id], "units",
      strlen(ncVariables[id].units), ncVariables[id].units));
  }

  if ( ncVariables[id].longName != NULL )
  {
    NCEXIT(nc_put_att_text(fc->id, fc->vars.id[id], "long_name",
      strlen(ncVariables[id].longName), ncVariables[id].longName));
  }

  if ( ncVariables[id].standardName != NULL )
  {
    NCEXIT(nc_put_att_text(fc->id, fc->vars.id[id], "standard_name",
      strlen(ncVariables[id].standardName),
      ncVariables[id].standardName));
  }

  if ( ncVariables[id].comment != NULL )
  {
    NCEXIT(nc_put_att_text(fc->id, fc->vars.id[id], "comment",
      strlen(ncVariables[id].comment), ncVariables[id].comment));
  }

  sprintf(cformat, "%%.%df", ncVariables[id].precision);

  NCEXIT(nc_put_att_text(fc->id, fc->vars.id[id], "C_format",
    strlen(cformat), cformat));

  NCEXIT(nc_put_att_float(fc->id, fc->vars.id[id], "_FillValue",
    NC_FLOAT, 1, &cnFcInitialValue));

  NCEXIT(nc_put_att_float(fc->id, fc->vars.id[id], "missing_value",
    NC_FLOAT, 1, &cnFcInitialValue));

  NCEXIT(nc_enddef(fc->id));
}


void cnFcAddGlobalAtt(cnTFCread * fc, char * name, char * att)
{
  size_t len;

  /* Check if attribute already exists and if it does then append. */
  if ( nc_inq_attlen(fc->id, NC_GLOBAL, name, &len) == NC_NOERR )
  {
    char * p = (char *) malloc(len + 1);
    if ( p == NULL )
    {
      printError(PE_EXIT, "malloc() failed");
    }

    NCEXIT(nc_get_att_text(fc->id, NC_GLOBAL, name, p));
    p[len] = '\0';
    p      = utAppendStr(p, utMakeStr("\n"), utMakeStr(att), NULL);
    if ( p == NULL )
    {
      printError(PE_EXIT, "malloc() failed");
    }

    nc_redef(fc->id);
    NCEXIT(nc_put_att_text(fc->id, NC_GLOBAL, name, strlen(p), p));
    NCEXIT(nc_enddef(fc->id));
    free(p);
  }
  else
  {
    nc_redef(fc->id);
    NCEXIT(nc_put_att_text(fc->id, NC_GLOBAL, name, strlen(att), att));
    NCEXIT(nc_enddef(fc->id));
  }
}


char * cnFcReadGlobalAtt(cnTFCread * fc, char * name)
{
  size_t len;
  char * retValue = NULL;

  NCEXIT(nc_inq_attlen(fc->id, NC_GLOBAL, name, &len));

  retValue = (char *) malloc(len + 1);
  if ( retValue == NULL )
  {
    printError(PE_EXIT, "malloc() failed");
  }

  NCEXIT(nc_get_att_text(fc->id, NC_GLOBAL, name, retValue));
  retValue[len] = '\0';

  return retValue;
}


void cnFcReadSfc(cnTFCread * fc, int varId, float * data/*OUT*/)
{
  if ( fc->vars.id[varId] >= 0 )
  {
    NCEXIT(nc_get_var_float(fc->id, fc->vars.id[varId], data));
  }
}


void cnFcWriteSfc(cnTFCread * fc, int varId, float * data)
{
  if ( fc->vars.id[varId] >= 0 )
  {
    NCEXIT(nc_put_var_float(fc->id, fc->vars.id[varId], data));
  }
}


void cnFcReadVarProfile(cnTFCread * fc, int varId, int profileNum,
  float * data/*OUT*/)
{
  if ( fc->vars.id[varId] >= 0 )
  {
    size_t start[2];
    size_t count[2];

    start[0] = (size_t) profileNum;
    start[1] = 0;
    count[0] = 1;
    count[1] = ( cnMCheckRange(CN_VAR_FIRST, CN_VAR_LAST, varId) )
      ? fc->dims.level : fc->dims.fluxLevel;
    NCEXIT(nc_get_vara_float(fc->id, fc->vars.id[varId], start, count,
      data));
  }
}


void cnFcWriteVarProfile(cnTFCread * fc, int varId, int profileNum,
  float * data)
{
  if ( fc->vars.id[varId] >= 0 )
  {
    size_t start[2];
    size_t count[2];

    start[0] = (size_t) profileNum;
    start[1] = 0;
    count[0] = 1;
    count[1] = ( cnMCheckRange(CN_VAR_FIRST, CN_VAR_LAST, varId) )
      ? fc->dims.level : fc->dims.fluxLevel;
    NCEXIT(nc_put_vara_float(fc->id, fc->vars.id[varId], start, count,
      data));
  }
}


void cnFcReadMwppProfile(cnTFCread * fc, int varId, int frequencyNum,
  int profileNum, float * data/*OUT*/)
{
  if ( fc->vars.id[varId] >= 0 )
  {
    size_t start[3];
    size_t count[3];

    start[0] = (size_t) frequencyNum;
    start[1] = (size_t) profileNum;
    start[2] = 0;
    count[0] = 1;
    count[1] = 1;
    count[2] = fc->dims.level;
    NCEXIT(nc_get_vara_float(fc->id, fc->vars.id[varId], start, count,
      data));
  }
}


void cnFcWriteMwppProfile(cnTFCread * fc, int varId, int frequencyNum,
  int profileNum, float * data)
{
  if ( fc->vars.id[varId] >= 0 )
  {
    size_t start[3];
    size_t count[3];

    start[0] = (size_t) frequencyNum;
    start[1] = (size_t) profileNum;
    start[2] = 0;
    count[0] = 1;
    count[1] = 1;
    count[2] = fc->dims.level;
    NCEXIT(nc_put_vara_float(fc->id, fc->vars.id[varId], start, count,
      data));
  }
}


/************************************************************************/
/************************************************************************/


