/*
 * CloudNET Model Conversion Software
 *
 * Stuff for accessing FC netCDF datasets.
 *
 * HISTORY
 * 2004-03-26  ao
 *
 * 2004-05-18  ao  added 'cnFcReadGlobalAtt()'.
 *
 * AUTHORS
 * Andrew Openshaw
 */


#if !defined(CNFCREAD_H)
#define CNFCREAD_H


/************************************************************************/
/************************************************************************/


#include <netcdf.h>
#include "cnmodels.h"
#include "cnfc.h"


/************************************************************************/
/************************************************************************/


/* For reading a FC dataset. */
typedef struct
{
  /* netCDF ID for the dataset. */
  int id;


  /* The dimension IDs. If not available then set to -1. */
  struct
  {
    int time;
    int level;
    int fluxLevel;
    int frequency;
  } dimIds;


  /* The dimension lengths. If not available then set to -1. */
  struct
  {
    int time;
    int level;
    int fluxLevel;
    int frequency;
  } dims;


  struct
  {
    float latitude;
    float longitude;
    float horRes;
    float time[CN_NC_MAX_TIME_DIM];
    float forecastTime[CN_NC_MAX_TIME_DIM];
    short level[CN_NC_MAX_LEVEL_DIM];
    short fluxLevel[CN_NC_MAX_LEVEL_DIM];
    float frequency[CN_NC_MAX_FREQ_DIM];

    /* The index into here is one of 'cnEVariableIds' from "cnmodels.h".
     * The value stored will be the netCDF ID for that variable or -1
     * if not in the file.
     */
    int   id[CN_VARIABLE_ID_COUNT];
  } vars;
} cnTFCread;


/************************************************************************/
/************************************************************************/


/* Open a FC dataset for read/write access.
 * Sets up 'fc'.
 */
void cnFcOpen(char * filename, cnTFCread * fc/*OUT*/);


/* Returns if 'varId' exists in the dataset.
 * 'varId' is the index into 'fc->vars.id[]'.
 */
void cnFcVarExists(cnTFCread * fc, char * filename, int varId);


/* Returns if 'varId' does not exist in the dataset.
 * 'varId' is the index into 'fc->vars.id[]'.
 */
void cnFcVarNotExists(cnTFCread * fc, char * filename, int varId);


/* Create a new dimension in the dataset.
 * PRE: 'name' is the name of the dimension and 'len' is it's length
 * (can be NC_UNLIMITED).
 * POST: returns the ID for the dimension.
 */
int cnFcAddDim(cnTFCread * fc, char * name, size_t len);


/* Create a new variable in the dataset.
 * PRE: 'id' is one of 'cnEVariableIds' in "cnmodels.h".
 * Enters define mode, defines the variable with correct dimensions
 * and its attributes then exits define mode.
 */
void cnFcAddVar(cnTFCread * fc, int id);


/* Add a text global attribute to the dataset.
 * PRE: 'name' is the name of the attribute and 'att' is it's value.
 * POST: if the global attribute already exists then 'att' is appended
 * to it otherwise the global attribute is created.
 */
void cnFcAddGlobalAtt(cnTFCread * fc, char * name, char * att);


/* Read the value of a text global attribute.
 * POST: returns allocated memory containing the zero-terminated text for
 * the attribute.
 */
char * cnFcReadGlobalAtt(cnTFCread * fc, char * name);


/* Read an entire surface variable with dimensions (time).
 * PRE: 'varId' is one of 'cnEVariables' and 'fc->vars.id[varId]' must
 * be >=0 otherwise nothing is done.
 * 'data' must have space for at least 'fc->dims.time' values.
 */
void cnFcReadSfc(cnTFCread * fc, int varId, float * data/*OUT*/);


/* Write an entire surface variabe with dimensions (time).
 * PRE: 'varId' is one of 'cnEVariables' and 'fc->vars.id[varId]' must
 * be >=0 otherwise nothing is done.
 * 'data' must have at least 'fc->dims.level' values.
 */
void cnFcWriteSfc(cnTFCread * fc, int varId, float * data);


/* Read an entire profile for a level or flux variable with dimensions
 * (time, level) or (time, flux_level).
 * PRE: 'varId' is one of 'cnEVariables' and 'fc->vars.id[varId]' must
 * be >=0 otherwise nothing is done.
 * 'data' must have space for at least 'fc->dims.level' or
 * 'fc->dims.fluxLevel' values.
 */
void cnFcReadVarProfile(cnTFCread * fc, int varId, int profileNum,
  float * data/*OUT*/);


/* Write an entire profile for a level or flux variable with dimensions
 * (time, level) or (time, flux_level).
 * PRE: 'varId' is one of 'cnEVariables' and 'fc->vars.id[varId]' must
 * be >=0 otherwise nothing is done.
 * 'data' must have at least 'fc->dims.level' or
 * 'fc->dims.fluxLevel' values.
 */
void cnFcWriteVarProfile(cnTFCread * fc, int varId, int profileNum,
  float * data);


/* Read an entire profile for a MWPP variable with dimensions
 * (frequency, time, level).
 * PRE: 'varId' is one of 'cnEVariables' and 'fc->vars.id[varId]' must
 * be >=0 otherwise nothing is done.
 * 'data' must have space for at least 'fc->dims.level' values.
 */
void cnFcReadMwppProfile(cnTFCread * fc, int varId, int frequencyNum,
  int profileNum, float * data/*OUT*/);


/* Write an entire profile for a MWPP variable with dimensions
 * (frequency, time, level).
 * PRE: 'varId' is one of 'cnEVariables' and 'fc->vars.id[varId]' must
 * be >=0 otherwise nothing is done.
 * 'data' must have at least 'fc->dims.level' values.
 */
void cnFcWriteMwppProfile(cnTFCread * fc, int varId, int frequencyNum,
  int profileNum, float * data);


/************************************************************************/
/************************************************************************/


#endif
