

#include "lsp_ice.h"


/************************************************************************/
/************************************************************************/

#define SQR(x)    ((x) * (x))
#define CUBE(x)   ((x) * SQR(x))

/************************************************************************/

const REAL    PI             = 3.14159265358979323846;


/* Conversion between Celsius and Kelvin. */
const REAL    ZERODEGC       = 273.15;

/* Gas constant for dry air. */
const REAL    R              = 287.05;

/* Specific of dry air at constant pressure. */
const REAL    CP             = 1005.0;

/* Ratio of molecular weights of water and dry air. */
const REAL    EPSILON        = 0.62198;

REAL          C_VIRTUAL;

/* Latent heat of condensation of water at 0degc. */
const REAL    LC             = 2.501E6;
                                                                          
/* Latent heat of fusion at 0degc. */
const REAL    LF             = 0.334E6;

/* The APB and TW parameters represent diffusional growth constants
 * and wet bulb temperature parameters. Do not change them.
 */
/* Terms in deposition and sublimation. */
const REAL    APB1           = 7.14E11;
const REAL    APB2           = 1.16E8;
const REAL    APB3           = 2.416E2;
                                                                         
/* Terms in evap of melting snow and rain. */
const REAL    APB4           = 5.57E11;
const REAL    APB5           = 1.03E8;
const REAL    APB6           = 2.04E2;
                                                                         
/* Numerical fit to wet bulb temperature. */
const REAL    TW1            = 1329.31;
const REAL    TW2            = 0.0074615;
const REAL    TW3            = 0.85E5;
                                                                         
/* Numerical fit to wet bulb temperature. */                            
const REAL    TW4            = 40.637;
const REAL    TW5            = 275.0;


/* Inhomogeneity factor for autoconversion rate */
const REAL    INHOMOG_RATE   = 1.0;

/* Inhomogeneity factor for autoconversion limit */
const REAL    INHOMOG_LIM    = 1.0;

/* Collision collection coefficient */
const REAL    EC_AUTO        = 0.55;

/* Droplet concentration over land */
const REAL    N_DROP_LAND    = 6.0E8;

/* (N_DROP_LAND)^(-1/3) */
const REAL    N_DROP_LAND_CR = 1.18563E-3;

/* Droplet concentration over sea */
const REAL    N_DROP_SEA     = 1.5E8;

/* (N_DROP_SEA)^(-1/3) */
const REAL    N_DROP_SEA_CR  = 1.88207E-3;

/* Threshold droplet radius for autoconversion */
const REAL    R_THRESH       = 7.0E-6;


/* These are initialised in LSP_INIT(). */

/* Liquid Water limit for autoconversion over sea */
REAL          AUTORATE_LAND;

/* Liquid Water limit for autoconversion over land */
REAL          AUTORATE_SEA;

/* Rate constant for autoconversion over sea */
REAL          AUTOLIM_LAND;

/* Rate constant for autoconversion over land */
REAL          AUTOLIM_SEA;

/* These are 1-based arrays. Index must be 1<=I<=16. */
REAL          CX[17];
REAL          CONSTP[17];

/* Vertical advection method.
 * ADV_TYPE=1: Original formulation
 * ADV_TYPE=2: Revised formulation has better fall through layers
 */
const INTEGER ADV_TYPE       = 2;


/* NUCLEATION OF ICE */

/* Note that the assimilation scheme uses temperature thresholds
 * in its calculation of qsat.
 */

/* Nucleation mass */
const REAL    M0             = 1.0E-12;

/* Maximum Temp for ice nuclei nucleation (deg C) */
const REAL    TNUC           = -10.0;

/* Maximum Temp for homogenous nucleation (deg C) */
const REAL    THOMO          = -40.0;


/* HALLETT MOSSOP PROCESS */

/* Min temp for production of Hallett Mossop splinters (deg C) */
const REAL    HM_T_MIN       = -8.0;

/* Max temp for production of Hallett Mossop splinters (deg C).
 * Switch off Hallett Mossop in this version but allow functionality
 */
const REAL    HM_T_MAX       = -273.0;

/* Residence distance for Hallett Mossop splinters (1/deg C) */
const REAL    HM_DECAY       = 1.0 / 7.0;

/* Reciprocal of scaling liquid water content for HM process */
const REAL    HM_RQCL        = 1.0 / 0.1E-3;


/* Minimum allowed QCF after microphysics */
const REAL    QCFMIN         = 1.0E-8;

/************************************************************************/

/* Prototypes. These are defined after LSP_ICE(). */

void    QSAT(REAL QS[], REAL T[], REAL P[], INTEGER POINTS);
void    QSAT_WAT(REAL QSL[], REAL T[], REAL P[], INTEGER POINTS);
REAL    ABS(REAL x);
REAL    MAX(REAL x, REAL y);
REAL    MIN(REAL x, REAL y);
INTEGER NINT(REAL x);


/************************************************************************/
/************************************************************************/


/* Calculates complete gammaf function by polynomial approximation. */
static REAL GAMMAF(REAL Y)
{
  INTEGER I;
  INTEGER M;
  REAL    GG;
  REAL    G;
  REAL    PARE;
  REAL    X;
  REAL    X_2;
  REAL    X_4;

  GG = 1.0;
  M  = Y;
  X  = Y - M;
  if ( M != 1.0 ) {
    for ( I = 1; I <= M - 1; I++ ) {
      G  = Y - I;
      GG = GG * G;
    }
  }

  X_2 = SQR(X);
  X_4 = SQR(X_2);
  PARE = -0.5748646 * X + 0.9512363 * X_2 - 0.6998588 * X_2 * X
    + 0.4245549 * X_4 - 0.1010678 * X_4 * X + 1.0;

  return PARE * GG;
}


/* Initialise the arrays CX and CONSTP. */
static void LSPCON(REAL CX[16], REAL CONSTP[16])
{
  REAL    TEMP;
  REAL    GBD1;
  REAL    GB1;
  REAL    GD3;
  REAL    GD52;
  REAL    GDR3;
  REAL    GDR4;
  REAL    GDR52;
  REAL    G1;
  REAL    G2;
  REAL    G3;

  /* Drop size distribution for rain: N(D) = N0 exp(-lambda D)
   * where N0 = X1R lambda^X2R.
   */
  REAL    X1R = 8.E6;
  REAL    X2R = 0.0;

  /* Particle size distribution for ice: N(D) = N0 D^m exp(-lambda D)
   * where N0 = X1I exp( -X3I T[deg C]/8.18 ) lambda^X2I and m=X4I.
   * Ice distn coeff.
   */
  REAL    X1I = 2.E6;
  REAL    X2I = 0.0;

  /* Whether there is a temperature dependence to N0snow. */
  REAL    X3I = 1.0;

  /* m parameter in gamma distribution. */
  REAL    X4I = 0.0;

  /* Mass diameter relationship for ice:  m(D) = AI D^BI. */
  REAL    AI = 6.9E-2;
  REAL    BI = 2.0;                                             

  /* Fall speed diameter relationship for ice: vt(D) = CI D^DI. */
  REAL    CI = 25.2;
  REAL    DI = 0.527;
                                            
  /* Fall speed diameter relationship for rain: vt(D) = CR D^DR. */
  REAL    CR = 386.8;
  REAL    DR = 0.67;

  CX[1]      = NINT(1000.0 * (2.0 + X4I - X2I) / (BI + 1.0 - X2I + X4I))
    / 1000.0;
  CX[2]      = NINT(1000.0 * (5.0 + DI - 2.0 * X2I + 2.0 * X4I)
    / (2.0 * (BI + 1.0 - X2I + X4I))) / 1000.0;
  CX[3]      = NINT(1000.0 * DI / (BI + 1.0 - X2I + X4I)) / 1000.0;
  CX[4]      = NINT(1000.0 * (3.0 + DI -X2I + X4I)
    / (BI + 1.0 - X2I + X4I)) / 1000.0;
  CX[5]      = NINT(1000.0 * DR / (4.0 + DR - X2R)) / 1000.0;
  CX[6]      = NINT(1000.0 / (X2I - X4I - 1.0 - BI)) / 1000.0;
  CX[7]      = NINT(1000.0 * (3.0 + DR - X2R) / (4.0 + DR - X2R))
    / 1000.0;  CX[8]      = NINT(1000.0 * X2I) / 1000.0;
  CX[9]      = NINT(1000.0 * X2R) / 1000.0;
  CX[10]     = NINT(1000.0 / (4.0 + DR - X2R)) / 1000.0;
  CX[11]     = NINT(1000.0 * ((DR + 5.0) / 2.0 - X2R)) / 1000.0;
  CX[12]     = NINT(1000.0 * (2.0 - X2R)) / 1000.0;
  CX[13]     = X3I;
  CX[14]     = 3.0 + X4I;
  CX[15]     = 2.0 + X4I;
  CX[16]     = 1.0 + X4I;
  TEMP       = BI + DI + 1.0 + X4I;
  GBD1       = GAMMAF(TEMP);
  TEMP       = BI + 1.0 + X4I;
  GB1        = GAMMAF(TEMP);
  TEMP       = 3.0 + DI + X4I;
  GD3        = GAMMAF(TEMP);
  TEMP       = 2.5 + DI / 2.0 + X4I;
  GD52       = GAMMAF(TEMP);
  TEMP       = 4.0 + DR;
  GDR4       = GAMMAF(TEMP);
  TEMP       = 3.0 + DR;
  GDR3       = GAMMAF(TEMP);
  TEMP       = 2.5 + DR / 2.0;
  GDR52      = GAMMAF(TEMP);
  TEMP       = 1.0 + X4I;
  G1         = GAMMAF(TEMP);
  TEMP       = 2.0 + X4I;
  G2         = GAMMAF(TEMP);
  TEMP       = 3.0 + X4I;
  G3         = GAMMAF(TEMP);
  CONSTP[1]  = 1.0 / (AI * X1I * GB1);
  CONSTP[2]  = 6.2832 * X1R;
  CONSTP[3]  = CI * GBD1 / GB1;
  CONSTP[4]  = 0.7854 * X1I * CI * GD3;
  CONSTP[5]  = 1.0 * 6.2832 * X1I;
  CONSTP[6]  = 89.48 * sqrt(CI) * GD52;
  CONSTP[7]  = 4.57E-7 * X1I;
  CONSTP[8]  = 523.6 * X1R * GDR4 * CR;
  CONSTP[9]  = 9869.6 * X1I * X1R;
  CONSTP[10] = 0.7854 * X1R * CR * GDR3;
  CONSTP[11] = CR * GDR4;
  CONSTP[12] = 63.100 * GDR52 * sqrt(CR);
  CONSTP[13] = G2;
  CONSTP[14] = G3 * 0.25;
  CONSTP[15] = G2 * 2.0;
  CONSTP[16] = G1 * 5.0;
}


void LSP_INIT(void)
{
  REAL R_THRESH_3 = CUBE(R_THRESH);

  /* The numbers 5907.24 and 4188.79 represent combinations of
   * physical constants. Do NOT change them.
   * PLEASE REFER TO UMDP 26 EQUATIONS P26.129 TO P26.136 FOR
   * AN EXPLANATION OF THE AUTOCONVERSION PARAMETERIZATION.
   * REMEMBER THAT THE PARAMETERIZATION IS VERY ROUGH.
   */
  C_VIRTUAL     = 1.0 / EPSILON - 1.0;
  AUTORATE_LAND = 5907.24 * EC_AUTO * INHOMOG_RATE * N_DROP_LAND_CR;
  AUTORATE_SEA  = 5907.24 * EC_AUTO * INHOMOG_RATE * N_DROP_SEA_CR;
  AUTOLIM_LAND  = 4188.79 * R_THRESH_3 * N_DROP_LAND * INHOMOG_LIM;
  AUTOLIM_SEA   = 4188.79 * R_THRESH_3 * N_DROP_SEA  * INHOMOG_LIM;

  LSPCON(CX, CONSTP);
}


void LSP_ICE(
  REAL    P,
  REAL    DZ,
  REAL    TIMESTEPFIXED,
  REAL    RHCRIT,
  REAL    QCF,
  REAL    QCL,
  REAL    Q,
  REAL *  RAIN,
  REAL *  SNOW,
  REAL *  VF,
  REAL    T,
  REAL    CFLIQ,
  REAL    CFICE,
  LOGICAL BLAND,
  INTEGER LS_ITER
)
{
  /* Latent heat of condensation / Cp (K). */
  REAL    LCRCP;

  /* Latent heat of fusion / Cp (K). */
  REAL    LFRCP;

  /* Sum of the above (S for Sublimation). */
  REAL    LSRCP;

  /* Constant in wet bulb temperature calculation. */
  REAL    CONW;

  /* Counter for the iterations */
  INTEGER J;

  /* Virtual temperature as at start of loop. */
  REAL    VTEMP;

  /* temperature degree C as at start of loop. */
  REAL    TEMPC;

  /* saturation vapour pressure (wrt ice below zero) */
  REAL    ESI;

  /* saturation vapour pressure (wrt water at all T) */
  REAL    ESW;

  /* increment to/from ice/snow */
  REAL    DQI;

  /* increment to/from cloud water */
  REAL    DQIL;

  /* increment to/from rain */
  REAL    DPR;

  /* fraction of ice inferred for fall speed calculations. */
  REAL    CFICETEMP;

  /* fallspeed for ice */
  REAL    FQI;

  /* 1.0/(DHI*LS_ITER) */
  REAL    DHILSITERR;

  /* saved flux of ice out of layer */
  REAL    FQIRQI;

  /* updated ice for long timestep */
  REAL    QUP;

  /* updated liquid cloud in implicit calculations */
  REAL    QCLNEW;

  /* term in melting */
  REAL    TEMP7;

  /* term in evaporation of rain */
  REAL    PR02;

  /* square of pr02 */
  REAL    PR04;

  /* term in autoconversion of cloud to rain */
  REAL    QC;

  /* denominator in deposition or evaporation of ice */
  REAL    APLUSB;

  /* density correction for fall speed */
  REAL    CORR;

  /* density correction for fall speed */
  REAL    ROCOR;

  /* surface air density */
  REAL    RHO1;

  /* Mean fall speed of rain */
  REAL    VR1;

  /* Mean fall speed of snow */
  REAL    VS1;

  /* Inverse lambda in rain exponential distribution */
  REAL    LAMR1;

  /* Expression containing calculations with lambda */
  REAL    LAMFAC1;

  /* lambda in snow exponential distribution */
  REAL    LAMS1;

  /* Mean velocity difference between rain and snow */
  REAL    FV1;

  /* Timestep of each iteration */
  REAL    TIMESTEP;

  /* Temperature correction of viscosity etc. */
  REAL    CORR2;

  /* Relative humidity required for nucleation */
  REAL    RHNUC;

  /* Temperature Factor for X1I in Cox Golding calculation */
  REAL    TCG;

  /* Inverse of TCG */
  REAL    TCGI;

  /* Constant effecting rate of deposition/sublimation of ice */
  REAL    RATEQ;

  /* Constant representing effect of sub grid distribution of ice */
  REAL    RATEQCF;

  /* Critical humidity for rain evaporation */
  REAL    RATEQSL;

  REAL    QS;
  REAL    QSL;
  REAL    SNOWT;
  REAL    RHO;
  REAL    RHOR;
  REAL    DHI;
  REAL    DHIR;
  REAL    RATEQS;

  /**********************************************************************/

  LCRCP = LC / CP;
  LFRCP = LF / CP;
  LSRCP = LCRCP + LFRCP;
  CONW  = R / (EPSILON * LC);

  /**********************************************************************/

  /* ------------------------------------------------------------------ */
  /* 2.1 Start the microphysics calculations                            */
  /* ------------------------------------------------------------------ */
  /* Set up the iterations */

  TIMESTEP = TIMESTEPFIXED / (float) LS_ITER;

  /* Set up sub grid scale constants. For no sub grid scale variability
   * use the set up RATEQ=1.0, RATEQCF=0.0, RATEQS=1.0.
   * Ideally represent these in a comdeck but require RHCRIT
   */

  RATEQ   = 1.0;
  RATEQCF = 0.0;
  SNOWT   = 0.0;
  RATEQS  = RHCRIT;

  /* ------------------------------------------------------------------ */
  /* 2.2 Start iterating.                                               */
  /* ------------------------------------------------------------------ */

  for ( J = 0; J < LS_ITER; J++ ) {
    REAL RHODZ;

    /* ---------------------------------------------------------------- */
    /* 2.3 Calculate sat humidity mixing ratio(respect to ice below     */
    /*     zero)                                                        */
    /* ---------------------------------------------------------------- */
    QSAT(&QS, &T, &P, 1);

    /* ---------------------------------------------------------------- */
    /* 2.4 Calculate sat humidity mixing ratio(respect to water at all  */
    /*     temps)                                                       */
    /* ---------------------------------------------------------------- */
    QSAT_WAT(&QSL, &T, &P, 1);

    /* -------------------------------------------------------------- */
    /* 3.1 Calculate density of air, RHO, via virtual temperature.    */
    /* -------------------------------------------------------------- */
    VTEMP   = T * (1.0 + C_VIRTUAL * Q - QCL - QCF);
    RHO     = P / (R * VTEMP);
    RHODZ   = RHO * DZ;
    RHOR    = 1.0 / RHO;
    RHO1    = 1.0;

    /* Correction factor of fall speeds etc. due to density. */
    CORR    = pow(RHO1 * RHOR, 0.4);

    /* Correction factor in viscosity etc. due to temperature. */
    CORR2   = pow(T / 273.0, 1.5) * (393.0 / (T + 120.0));

    /* Combined correction factor */
    ROCOR   = (REAL) sqrt((double) (RHO * CORR * CORR2));

    /* Calculate RATEQS as a function of RHCRIT and CFICE */
    RATEQS  = RHCRIT + CFICE * (1.0 - RHCRIT);
    RATEQSL = RHCRIT + CFLIQ * (1.0 - RHCRIT);

    /* -------------------------------------------------------------- */
    /* 3.2 set T deg C sat vapour pressures in N/m2                   */
    /* -------------------------------------------------------------- */
    TEMPC   = T - ZERODEGC;
    ESI     = QS * P / EPSILON;
    ESW     = QSL * P / EPSILON;

    /* Calculate a temperature factor for N0snow. CX[13]=1.0 if there
     * is a temperature dependence, and 0.0 if there is not.
     */
    TCG   = exp(- CX[13] * TEMPC / 8.18);
    TCGI  = 1.0 / TCG;

    /* -------------------------------------------------------------- */
    /* 4 Check that ice cloud fraction is sensible.                   */
    /* -------------------------------------------------------------- */
    CFICETEMP = CFICE;

    /* The possibility exists in multiple iterations that ice cloud
     * fraction is equal to zero but nucleation and deposition from the
     * last iteration has produced a finite ice content. Hence this
     * section produces a fix which will stop the scheme crashing.
     * Only need to use for more than 1 iteration
     */
    if ( LS_ITER > 1 ) {
      if ( QCF > 0.0 && CFICE < 0.1) {
        CFICETEMP = MAX(CFLIQ, 0.1);
      }                                                          
    }                                                            

    /* -------------------------------------------------------------- */
    /* 5.1 Calculate fallspeeds for ice and water and limit           */
    /*     to CFL criteria for time to fall through layer.            */
    /* -------------------------------------------------------------- */

    /* Estimate fall speed out of this layer. We want to avoid advecting
     * very small amounts of snow between layers, as this can cause
     * numerical problems in other routines, so if QCF is smaller than
     * a single nucleation mass per metre cubed don't advect it.
     */
    if ( QCF > M0 ) {
      /* Estimate the mean fall speed across the entire gridbox.
       * Use a top hat distrubution within the gridbox
       */
      FQI = CONSTP[3] * CORR * pow(RHO * QCF * CONSTP[1] * TCGI
        / CFICETEMP, CX[3]);
    }
    else {                                                              
      /* QCF is smaller than zero so set fall speed to zero */
      FQI = 0.0;
    }                                                            

    /* Calculate CFL quantity of timestep over level separation. */
    DHI  = TIMESTEP / DZ;

    /* Define DHIR and DHILSITERR to speed up calculations. */
    DHIR    = 1.0 / DHI;
    DHILSITERR = 1.0 / (DHI * LS_ITER);

    /* See if fall speed is small enough that not all the ice falls out
     * of the layer.
     */
    if ( *VF < DHIR ) {
      /* short timestep solution */
      *VF = FQI;

      if ( *VF < DHIR ) {
        /* flux out is just represented by the fall speed estimated
         * above
         */
        FQIRQI = FQI * RHO * QCF;
      }
      else {
        /* cannot allow more ice to leave than was already there */
        FQIRQI = RHO * QCF * DHIR;
      }

      /* calculate new ice content in this layer by flux divergence */
      QCF = QCF + (*SNOW - FQIRQI) * DHI * RHOR;
    }
    else {                                                              
      /* long timestep case */
      QUP    = *SNOW * RHOR / *VF;
      FQIRQI = *SNOW - (RHODZ * (QUP - QCF) / TIMESTEP);
      QCF    = QUP;

      /* VF must be as least as great as the fall velocity of the
       * current layer 
       */
      if ( *VF < FQI ) *VF = FQI;
    }                                                            

    /* -------------------------------------------------------------- */
    /* 5.2 Snow is used to save fall out of layer                     */
    /*     for calculation of fall into next layer                    */
    /* -------------------------------------------------------------- */
    SNOWT = SNOWT + FQIRQI / LS_ITER;

    /* -------------------------------------------------------------- */
    /* Transfer processes only active at T less than 0 deg C          */
    /* -------------------------------------------------------------- */
    if ( T < ZERODEGC ) {
      /* ------------------------------------------------------------ */
      /* 6   Nucleation of snow - explicit                            */
      /* ------------------------------------------------------------ */

      /* ------------------------------------------------------------ */
      /* 6.1 Homogenous nucleation takes place at temperatures less   */
      /*     than THOMO                                               */
      /* ------------------------------------------------------------ */
      if ( T < (ZERODEGC + THOMO) ) {
        /* Turn all liquid to ice */
        QCF = QCF + QCL;
        T   = T + LFRCP * QCL;
        QCL = 0.0;
      }                                                         

      /* ------------------------------------------------------------ */
      /* 6.2 Heteorgenous nucleation occurs for temps less than TNUC  */
      /*     deg C                                                    */
      /* ------------------------------------------------------------ */
      if ( T < (ZERODEGC + TNUC) ) {
        REAL TC = T - ZERODEGC;

        /* Calculate number of active ice nucleii */
        DQI     = MIN(0.01 * exp(-0.6 * TEMPC), 1.0E5);

        /* Each nucleus can grow to arbitary mass of M0 kg */
        DQI     = M0 * DQI * RHOR;

        /* RHNUC represents how much moisture is available for ice
         * formation.
         */
        RHNUC   = (188.92 +2.81 * TC + 0.013336 * SQR(TC) - 10.0) * 0.01;
        RHNUC   = MIN(RHNUC, 1.0);

        /* Predict transfer of mass to ice. */
        DQI     = MAX(MIN(DQI, Q + QCL - RATEQS
          * MAX(QSL * RHNUC, QS)), 0.0);
        QCF     = QCF + DQI;

        /* This comes initially from liquid water */
        DQIL    = MIN(DQI, QCL);
        QCL     = QCL - DQIL;
        T       = T + LFRCP * DQIL;

        /* If more moisture is required then nucleation removes from
         * vapour.
         */
        DQI     = DQI - DQIL;
        T       = T + LSRCP * DQI;
        Q       = Q - DQI;
      }

      /* ------------------------------------------------------------ */
      /* 7 Deposition/Sublimation of snow - explicit                  */
      /* ------------------------------------------------------------ */
      if ( QCF > M0 ) {
        REAL T_2 = SQR(T);
        REAL T_3 = T * T_2;

        /* Calculate transfer rate as a function of QCF and T */
        PR02   = RHO * QCF * CONSTP[1] * TCGI;
        APLUSB = (APB1 - APB2 * T) * ESI;
        APLUSB = APLUSB + T_3 * P * APB3;
        DQI    = TCG * CONSTP[5] * T_2 * ESI * RATEQ
          * (MIN(Q + QCL, QSL) - RATEQCF * QCF - RATEQS * QS)
          * (0.65 * CONSTP[13] * CORR2 * pow(PR02, CX[1])
            + CONSTP[6] * ROCOR * pow(PR02, CX[2])) / (QS * APLUSB * RHO);

        /* Limits depend on whether deposition or sublimation occurs */
        if ( DQI > 0.0 ) {
          /* Deposition is occuring. Limit is available moisture. */
          DQI = MIN(DQI * TIMESTEP, (Q + QCL - RATEQCF * QCF
            - RATEQS * QS) / (1.0 + RATEQCF));
        }
        else {                                                          
          /* Sublimation is occuring. Limits are spare moisture capacity
           * and QCF
           */
          DQI = MAX(MAX(DQI * TIMESTEP, (Q + QCL - RATEQCF * QCF
            - RATEQS * QS) / (1.0 + RATEQCF)), - QCF);
        }                                                        

        /* Adjust ice content */
        QCF    = QCF + DQI;
        DQIL   = MAX(MIN(DQI, QCL), 0.0);

        /* Adjust liquid content (deposits before vapour by Bergeron
         * Findeison process).
         */
        QCL    = QCL - DQIL;
        T      = T + LFRCP * DQIL;
        DQI    = DQI - DQIL;

        /* Adjust vapour content */
        Q      = Q - DQI;
        T      = T + LSRCP * DQI;
      }                                                          

      /* ------------------------------------------------------------ */
      /* 8 Riming of snow by cloud water -implicit in QCL             */
      /* ------------------------------------------------------------ */
      if ( QCF > M0 && QCL > 0.0 ) {                        
        QCLNEW = QCL / (1.0 + CONSTP[4] * TCG * CORR * TIMESTEP
          * pow(RHO * QCF  * CONSTP[1] * TCGI, CX[4]));

        /* Recalculate water contents */
        QCF = QCF + (QCL - QCLNEW);
        T   = T + LFRCP * (QCL - QCLNEW);
        QCL = QCLNEW;
      }                                                          

      /* ------------------------------------------------------------ */
      /* 9 Capture of rain by snow - implicit in rain                 */
      /*   This is a candidate to be removed to speed things up.      */
      /* ------------------------------------------------------------ */
      if ( *RAIN > 0.0 && QCF > M0 ) {
        REAL L_2;
        REAL L_3;
        REAL L_4;
        REAL L_5;
        REAL L_6;

        /* Calculate velocities */
        VR1 = CORR * CONSTP[11] / 6.0
          * pow(*RAIN / (CONSTP[8] * CORR), CX[5]);
        VS1 = CONSTP[3] * CORR
          * pow(RHO * QCF * CONSTP[1] * TCGI, CX[3]);

        /* Estimate the mean absolute differences in velocities. */
        FV1 = MAX(ABS(VR1 - VS1), (VR1 + VS1) / 8.0);

        /* Calculate functions of slope parameter lambda */
        /* MOD fdm2f404 changes next 3 active lines */
        /* LAMR1 and LAMS1 are INVERSE lambda values */
        LAMR1   = pow(*RAIN / (CONSTP[8] * CORR), CX[10]);
        LAMS1   = pow(RHO * QCF * CONSTP[1] * TCGI, -CX[6]);

        L_2     = SQR(LAMR1);
        L_3     = L_2 * LAMR1;
        L_4     = SQR(L_2);
        L_5     = L_2 * L_3;
        L_6     = SQR(L_3);

        LAMFAC1 = CONSTP[16] * (L_6 * pow(LAMS1, CX[16]))
                + CONSTP[15] * (L_5 * pow(LAMS1, CX[15]))
                + CONSTP[14] * (L_4 * pow(LAMS1, CX[14]));

        /* Calculate transfer */
        DPR = TCG * CONSTP[9] * pow(LAMS1, -CX[8]) * pow(LAMR1, -CX[9])
          * FV1 * LAMFAC1 * TIMESTEP * RHOR;
        DPR = MIN(DPR, *RAIN * (DHI * LS_ITER) * RHOR);

        /* Adjust ice and rain contents */
        QCF  = QCF + DPR;
        *RAIN = *RAIN - DPR * RHO * DHILSITERR;
        T    = T + LFRCP * DPR;
      }                                                          

      /* ------------------------------------------------------------ */
      /* End of transfer processes only active at T less than 0 deg C */
      /* ------------------------------------------------------------ */
    }




    /* -------------------------------------------------------------- */
    /* 10 Evaporate melting snow - implicit in subsaturation          */
    /* -------------------------------------------------------------- */
    if ( QCF > M0 && T > ZERODEGC ) {
      /* Calculate transfer as a function of QCF, T and specific
       * humidity
       */
      PR02 = RHO * QCF * CONSTP[1] * TCGI;
      PR04 = (APB4 - APB5 * T) * ESW + APB6 * P * CUBE(T);
      DPR  = TCG * RATEQ * CONSTP[5] * SQR(T) * ESW * TIMESTEP
        * (0.65 * CONSTP[13] * CORR2 * pow(PR02, CX[1])
        + CONSTP[6] * ROCOR * pow(PR02, CX[2])) / (QSL * RHO * PR04);
      DPR  = DPR * MAX((RATEQS * QSL+ RATEQCF
          * QCF - Q - QCL), 0.0) / (1.0 + DPR * (1.0 + RATEQCF));

      /* Extra check to see we don't get a negative QCF */
      DPR  = MIN(DPR, QCF);

      /* Update values of ice and vapour */
      QCF = QCF - DPR;
      Q   = Q + DPR;
      T   = T- DPR * LSRCP;
    }                                                            

    /* -------------------------------------------------------------- */
    /* 11 Melting of snow - explicit                                  */
    /*    USE WET BULB TEMP (DEG.C) IN SNOW MELT CALC.                */
    /*    Use a numerical approximation.                              */
    /* -------------------------------------------------------------- */
    if ( QCF > M0 && T > ZERODEGC ) {
      TEMPC = T - ZERODEGC;

      /* An approximate calculation of wet bulb temperature */
      TEMP7 = TEMPC - RATEQ
        * MAX((RATEQS * QSL + RATEQCF * QCF - Q), 0.0)
        * (TW1 + TW2 * (P - TW3) - TW4 * (T - TW5) );
      TEMP7 = MAX(TEMP7, 0.0);
      /* End of wet bulb temp formulations. */

      PR02 = RHO * QCF * CONSTP[1] * TCGI;
      DPR  = TCG * CONSTP[7] * TIMESTEP
        * (0.65 * CONSTP[13] * CORR2 * pow(PR02, CX[1])
          + CONSTP[6] * ROCOR * pow(PR02, CX[2])) * RHOR;

      /* Solve implicitly in terms of temperature */
      DPR  = TEMP7 * (1.0 - 1.0 / (1.0 + DPR * LFRCP)) / LFRCP;
      DPR  = MIN(DPR, QCF);

      /* Update values of ice and Rain */
      QCF  = QCF - DPR;
      *RAIN = *RAIN + DPR * RHO * DHILSITERR;
      T    = T - LFRCP * DPR;
    }                                                            

    /* -------------------------------------------------------------- */
    /* 12 Evaporation of rain - implicit in subsaturation             */
    /* -------------------------------------------------------------- */
    if ( *RAIN > 0.0 ) {
      PR04 = (APB4 - APB5 * T) * ESW + APB6 * P * CUBE(T);

      /* New, consistent evaporation method, with rain fall speed
       * relationship.
       * LAMR1 is now an INVERSE lambda value
       */
      LAMR1 = pow(*RAIN / (CONSTP[8] * CORR), CX[10]);
      DPR   = CONSTP[2] * SQR(T) * ESW * TIMESTEP;
      DPR   = DPR * ( (0.78 * CORR2 * pow(LAMR1, CX[12]))
        + (CONSTP[12] * ROCOR * (pow(LAMR1, CX[11]))) );
      DPR   = DPR * RATEQ / (QSL * RHO * PR04);

      /* Calculate transfers. */
      DPR   = DPR * MAX((RATEQSL * QSL - Q - QCL), 0.0) / (1.0 + DPR);
      DPR   = DPR * RHO * DHILSITERR;
      DPR   = MIN(DPR, *RAIN);

      /* Update values of rain and vapour */
      *RAIN = *RAIN - DPR;
      Q    = Q + DPR * DHI * LS_ITER * RHOR;
      T    = T - DPR * LCRCP * DHI * LS_ITER * RHOR;
    }                                                            

    /* -------------------------------------------------------------- */
    /* 13 Accretion of cloud on rain - implicit in liquid water       */
    /*    content                                                     */
    /* -------------------------------------------------------------- */
    if ( *RAIN > 0.0 && QCL > 0.0 ) {
      /* New accretion formulation. */
      PR02   = *RAIN / (CONSTP[8] * CORR);
      QCLNEW = QCL / ((1.0 + CONSTP[10] * CORR * TIMESTEP
        * pow(PR02, CX[7])));

      /* Now calculate increments to rain. */
      *RAIN = *RAIN + (QCL - QCLNEW) * RHO * DHILSITERR;
      QCL  = QCL - (QCL - QCLNEW);
    }                                                          

    /* -------------------------------------------------------------- */
    /* 14 Autoconversion of cloud to rain - explicit                  */
    /* -------------------------------------------------------------- */
    if ( QCL > 0.0 && CFLIQ > 0.0 ) {
      /* Use a liquid cloud fraction here as this term is very
       * non-linear
       * The section below is a simple way of proceeding.
       */
      if ( BLAND ) {
        /* Land point */
        QC  = MIN(AUTOLIM_LAND * CFLIQ * RHOR, QCL);
        DPR = MIN(AUTORATE_LAND * pow(RHO * QCL / CFLIQ, 1.333)
          * TIMESTEP * QCL / CORR2, QCL - QC);
      }
      else {
        /* Sea point */
        QC  = MIN(AUTOLIM_SEA * CFLIQ * RHOR, QCL);
        DPR = MIN(AUTORATE_SEA * pow(RHO * QCL / CFLIQ, 1.333)
          * TIMESTEP * QCL / CORR2, QCL - QC);
      }                                                        
      /* End of calculation of autoconversion amount DPR */

      QCL  = QCL - DPR;
      *RAIN = *RAIN + DPR * RHO * DHILSITERR;
    }
  }


  /* Copy contents of SNOWT to SNOW, to fall into next layer down */
  *SNOW = SNOWT;

  /* ---------------------------------------------------------------- */
  /* 16 Melt any SNOW which has reached here, as long as T is large   */
  /*    enough                                                        */
  /* ---------------------------------------------------------------- */

  /* Only use if long timestep case. In which case melt the excess
   * snow which falls straight through a layer. Use DQI variable to
   * save space.
   * DQI APPROXIMATELY represents the excess SNOW.
   */

  DQI = MIN(*VF * DHI - 1.0, 1.0);

  if ( DQI > 0.0 ) {
    /* Long timestep case */
    TEMPC = T - ZERODEGC;

    if ( *SNOW > 0.0 && T > ZERODEGC ) {
      /* Numerical approximation of wet bulb temperature. */
      TEMP7 = TEMPC - RATEQ * MAX((RATEQS * QSL + RATEQCF
        * QCF - Q), 0.0) * (TW1 + TW2 * (P -TW3 ) - TW4 * (T - TW5));
      TEMP7 = MAX(TEMP7, 0.0);
      /* End of wet bulb calculation */

      DPR = TEMP7 / (LFRCP * LS_ITER);
      DPR = MIN(DPR, *SNOW * DHI * RHOR * DQI);

      /* Update values of snow and rain */
      *SNOW = *SNOW - DPR * RHO * DHIR;
      *RAIN = *RAIN + DPR * RHO * DHIR;
      T    = T - LFRCP * DPR * LS_ITER;
    }
  }

  /* ---------------------------------------------------------------- */
  /* 17 Remove any small amount of QCF which is left over to be tidy. */
  /*    If QCF is less than 1e-10 kg/kg (hardwired) and isn't growing */
  /*    by deposition (assumed to be given by RHCRIT) then remove it. */
  /* ---------------------------------------------------------------- */
  if ( QCF < 1.0E-10
    && (T > ZERODEGC || (Q + QCL < RHCRIT * QS) || QCF < 0.0) ) {
    Q   = Q + QCF;
    T   = T - LSRCP * QCF;
    QCF = 0.0;
  }
}


/* Large-scale Cloud Scheme Compression routine (Cloud points only).
 * Purpose: Calculates liquid cloud water amounts and cloud amounts,
 *          temperature and specific humidity from cloud-conserved and
 *          other model variables. This is done for one model level.
 * Documentation: UMDP No.29
 */
static void LS_CLD_C(
  /* pressure (Pa). */
  REAL   P_F,

  /* Critical relative humidity.
   * See the paragraph incorporating eqs P292.11 to P292.14.
   */
  REAL   RHCRIT,

  /* saturated humidity at temperature TL, and pressure P_F */
  REAL   QSL_F,

  /* Normalised super/subsaturation ( = QC/BS). */
  REAL   QN_F,

  /* On input : Vapour + liquid water content (QW) (kg per kg air).
   * On output: Specific humidity at processed levels
   * (kg water per kg air).
   */
  REAL * Q_F,

  /* On input : Liquid water temperature (TL) (K).
   * On output: Temperature at processed levels (K).
   */
  REAL * T_F,

  /*OUT Cloud liquid water content at processed levels (kg per kg air). */
  REAL * QCL_F,

  /*OUT Liquid cloud fraction at processed levels. */
  REAL * CF_F,

  /*OUT Super/subsaturation on processed levels. Input initially RMDI. */
  REAL * GRID_QC_F,

  /* OUT Value of bs at processed levels. Input initialized to RMDI. */
  REAL * BS_F
)
{
  /* For liquid AlphaL calculation. */
  REAL    ALPHL     = EPSILON * LC / R;

  /* Lat ht of condensation/Cp. */
  REAL    LCRCP     = LC / CP;

  /* Weighting for ALPHAL iteration */
  REAL    WTN       = 0.75;

  /* Total number of iterations */
  INTEGER ITS       = 5;

  /* LOCAL AL (see equation P292.6). */
  REAL    AL;

  /* LOCAL ALPHAL (see equation P292.5). */
  REAL    ALPHAL;

  /* N - iteration counter. */
  INTEGER N;

  /* Pressure (Pa). */
  REAL    P;

  /* Saturated spec humidity for temp T. */
  REAL    QS;

  /* Cloud water normalised with BS. */
  REAL    QCN;

  /* temperature. */
  REAL    T;

  /* specific humidity. */
  REAL    Q;

  /* Sigmas*sqrt(6): sigmas the parametric standard deviation of local
   * cloud water content fluctuations.
   */
  REAL    BS;

  /* ALPHAL at previous iteration. */
  REAL    ALPHAL_NM1;

  /**********************************************************************/

  P          = P_F;
  QCN        = QN_F;

  /* ---------------------------------------------------------------- */
  /* 1. Calculate ALPHAL (eq P292.5) and AL (P292.6).                 */
  /*    CAUTION: T_F acts as TL (input value) until update in final   */
  /*             section                                              */
  /*    CAUTION: Q_F acts as QW (input value) until update in final   */
  /*             section                                              */
  /* ---------------------------------------------------------------- */

  ALPHAL     = ALPHL * QSL_F / SQR(*T_F);
  AL         = 1.0 / (1.0 + (LCRCP * ALPHAL));
  ALPHAL_NM1 = ALPHAL;

  if ( RHCRIT < 1.0 ) {

    /* -------------------------------------------------------------- */
    /* 2. Calculate cloud fraction CF, BS (ie. sigma*sqrt(6), where   */
    /*    sigma is as in P292.14) and normalised cloud water          */
    /*    QCN=qc/BS, using eqs P292.15 & 16 if RHcrit < 1.            */
    /* N.B. QN (input) is initially in QCN                            */
    /* N.B. QN does not depend on AL and so CF and QCN can be         */
    /*      calculated outside the iteration (which is performed in   */
    /*      LS_CLD_C).                                                */
    /*      QN is > -1 for all points processed so CF > 0.            */
    /* -------------------------------------------------------------- */

    BS = (1.0 - RHCRIT) * AL * QSL_F;
    if ( QCN <= 0.0 ) {
      REAL rtmp = 1.0 + QCN;
      rtmp      = SQR(rtmp);

      *CF_F     = 0.5 * rtmp;
      QCN       = (1.0 + QCN) * rtmp / 6.0;
    }
    else if ( QCN < 1.0 ) {
      REAL rtmp = 1.0 - QCN;
      rtmp      = SQR(rtmp);

      *CF_F     = 1.0 - 0.5 * rtmp;
      QCN       = QCN + (1.0 - QCN) * rtmp / 6.0;
    }
    else {
      *CF_F     = 1.0;
    }
  }
  else {
    /* -------------------------------------------------------------- */
    /* 3.a If RHcrit = 1., all points processed have QN > 0 and       */
    /*     CF = 1.                                                    */
    /* -------------------------------------------------------------- */
    BS          = AL;
    *CF_F       = 1.0;
  }

  /* ---------------------------------------------------------------- */
  /* 3.1 Calculate 1st approx. to qc (store in QCL)                   */
  /* ---------------------------------------------------------------- */
  *QCL_F = QCN * BS;

  /* ---------------------------------------------------------------- */
  /* 3.2 Calculate 1st approx. specific humidity (total minus cloud   */
  /*     water)                                                       */
  /* ---------------------------------------------------------------- */
  Q      = *Q_F - *QCL_F;

  /* ---------------------------------------------------------------- */
  /* 3.3 Calculate 1st approx. to temperature, adjusting for latent   */
  /*     heating                                                      */
  /* ---------------------------------------------------------------- */
  T      = *T_F + LCRCP * *QCL_F;

  /* ------------------------------------------------------------------ */
  /* 4. Iteration to find better cloud water values.                    */
  /* ------------------------------------------------------------------ */
  if ( ITS >= 2 ) {
    for ( N = 2; N <= ITS; N++ ) {
      QSAT_WAT(&QS, &T, &P, 1);

      if ( T > *T_F ) {
        /* NB. T > TL implies cloud fraction > 0. */
        ALPHAL     = (QS - QSL_F) / (T - *T_F);
        ALPHAL     = WTN * ALPHAL + (1.0 - WTN) * ALPHAL_NM1;
        ALPHAL_NM1 = ALPHAL;
        AL         = 1.0 / (1.0 + (LCRCP * ALPHAL));

        if ( RHCRIT < 1.0 ) {
          BS = (1.0 - RHCRIT) * AL * QSL_F;
        }
        else {
          BS = AL;
        }

        /* ---------------------------------------------------------- */
        /* 4.1 Calculate Nth approx. to qc (store in QCL).            */
        /* ---------------------------------------------------------- */
        *QCL_F = QCN * BS;

        /* ---------------------------------------------------------- */
        /* 4.2 Calculate Nth approx. spec. humidity (total minus      */
        /*     cloud water).                                          */
        /* ---------------------------------------------------------- */
        Q     = *Q_F - *QCL_F;

        /* ---------------------------------------------------------- */
        /* 4.3 Calculate Nth approx. to temperature, adjusting for    */
        /*     latent heating                                         */
        /* ---------------------------------------------------------- */
        T     = *T_F + LCRCP * *QCL_F;
      }
    }
  }

  /* ------------------------------------------------------------------ */
  /* 5. Finally scatter back cloud point results to full field arrays.  */
  /*    CAUTION: T_F updated from TL (input) to T (output)              */
  /*    CAUTION: Q_F updated from QW (input) to Q (output)              */
  /* ------------------------------------------------------------------ */
  *Q_F       = Q;
  *T_F       = T;
  *GRID_QC_F = BS * QN_F;
  *BS_F      = BS;
}


int LS_CLD(
  REAL       P[],
  REAL       RHCRIT[],
  INTEGER    LEVELS,
  REAL       T[],
  REAL       CF[],
  REAL       Q[],
  REAL       QCF[],
  REAL       QCL[],
  REAL       CFL[],
  REAL       CFF[]
)
{
#define alloc_real(p, size) p = malloc(sizeof(REAL) * size);\
  if ( p == NULL ) return -1


  /* New real missing data indicator (-2**30) */
  REAL    RMDI                  = -32768.0 * 32768.0;

  /* Selects different methods for calculating the ice cloud
   * fraction and the total layer cloud fraction (overlapping of liquid
   * and ice cloud).
   *
   * Which method is to be used to calculate ice cloud fraction
   *
   * 1: Based on Smith relationship between condensate and cloud fraction
   * 2: Like 1 but uses cloud top temperature to modify the qsat value.
   */
  INTEGER ICE_FRACTION_METHOD   = 1;

  /* Which method is to be used to calculate overlap between ice and
   * liquid cloud
   * 1: Use a minimum overlap method in a bit comparable way to vn5.1
   * 2: Uses a generalized degree of overlap between ice and liquid clouds
   *    given in file C_CLDSGS
   */
  INTEGER CLOUD_FRACTION_METHOD = 1;

  /* Weight to apply to cloud top temperature rather than local
   * temperature.
   * For a Smith like scheme set CTT_WEIGHT=0.0 and T_WEIGHT=1.0
   */
  REAL    CTT_WEIGHT            = 0.333;

  /* Weight to apply to local temperature rather than cloud top
   * temperature
   */
  REAL    T_WEIGHT              = 0.333;

  /* Fixed reference value of saturation wrt liquid. */
  REAL    QSAT_FIXED            = 0.1E-3;

  /* Overlap of ice and liquid cloud in the SAME layer
   * Set OVERLAP_ICE_LIQUID to 1.0 for maximum overlap, 0.0 for random
   * overlap, -1.0 for minimum overlap, or any value between -1.0 and 1.0
   * for something inbetween these fixed methods.
   */
  REAL    OVERLAP_ICE_LIQUID    = 0.0;

  /* Cloud fraction tuning constant
   * Advise SUB_CLD to be around 0.225 for 12km resolution, 0.225 for 60km
   * resolution and 0.15 for 300km resolution.
   */
  REAL    SUB_CLD               = 0.225;

  /* Minimum ice content to perform calculations */
  REAL    QCFMIN                = 1.0E-8;


  REAL    ROOTWO                = sqrt(2.0);

  /* Subgrid parameter in ice cloud calculation */
  REAL    SUBGRID;

  /* qCF / bs */
  REAL    QCFRBS;

  /* Arc-cosine term in Cloud ice fraction calc. */
  REAL    PHIQCF;

  /* Cosine term in Cloud ice fraction calc. */
  REAL    COSQCF;

  /* Maximum possible overlap */
  REAL    OVERLAP_MAX;

  /* Minimum possible overlap */
  REAL    OVERLAP_MIN;

  /* Random overlap */
  REAL    OVERLAP_RANDOM;

  /* Temporaries for combinations of the overlap parameters */
  REAL    TEMP0;
  REAL    TEMP1;
  REAL    TEMP2;

  /* Loop counters: K - vertical level index. */
  INTEGER K;

  /* Saturated specific humidity for temp TL or T. */
  REAL    QSL;

  /* Saturated specific humidity wrt liquid at cloud
   * top temperature
   */
  REAL    QSL_CTT;

  /* Cloud water normalised with BS. */
  REAL    QN;

  /* [LEVELS] Gridbox mean saturation excess at
   * processed levels (kg per kg air).
   * Set to RMDI when cloud is absent.
   */
  REAL *  GRID_QC;

  /* [LEVELS] Maximum moisture fluctuation /6*sigma at
   * processed levels (kg per kg air). Set to RMDI when cloud is absent.
   */
  REAL *  BS;

  /* Ice cloud top temperature (K) - as coded it
   * is really TL
   */
  REAL    CTT;

  /* True for points with non-zero cloud */
  LOGICAL LQC;

  /* Last Level With Ice Cloud */
  INTEGER LLWIC;

  /**********************************************************************/

  alloc_real(GRID_QC, LEVELS);
  alloc_real(BS, LEVELS);

  /**********************************************************************/

  /* Initialize cloud-top-temperature and last-level-with-ice-cloud
   * arrays
   */
  CTT   = 0.0;
  LLWIC = 0;

  for ( K = LEVELS - 1; K >= 0; K-- ) {
    /* ---------------------------------------------------------------- */
    /* 1. Calculate QSAT at liquid/ice water temperature, TL, and       */
    /*    initialize cloud water, sub-grid distribution and fraction    */
    /*    arrays.                                                       */
    /*    This requires a preliminary calculation of the pressure.      */
    /*    NB: On entry to the subroutine 'T' is TL and 'Q' is QW.       */
    /* ---------------------------------------------------------------- */
    QCL[K]     = 0.0;
    CFL[K]     = 0.0;
    GRID_QC[K] = RMDI;
    BS[K]      = RMDI;

    QSAT_WAT(&QSL, &T[K], &P[K], 1);

    if ( RHCRIT[K] < 1.0 ) {
      /* -------------------------------------------------------- */
      /* 2. Calculate the quantity QN = QC/BS = (QW/QSL-1)        */
      /*    /(1-RHcrit)                                           */
      /*    if RHcrit is less than 1                              */
      /* -------------------------------------------------------- */
      QN = (Q[K] / QSL - 1.0) / (1.0 - RHCRIT[K]);

      /* -------------------------------------------------------- */
      /* 3. Set logical variable for cloud, LQC, for the case     */
      /*    RHcrit < 1;                                           */
      /*    where QN > -1, i.e. qW/qSAT(TL,P) > RHcrit, there is  */
      /*    cloud                                                 */
      /* -------------------------------------------------------- */
      LQC = QN > -1.0;
    }
    else {
      /* -------------------------------------------------------- */
      /* 2.a Calculate QN = QW - QSL if RHcrit equals 1           */
      /* -------------------------------------------------------- */
      QN = Q[K] - QSL;

      /* -------------------------------------------------------- */
      /* 3.a Set logical variable for cloud, LQC, for the case    */
      /*     RHcrit = 1;                                          */
      /*     where QN > 0, i.e. qW > qSAT(TL,P), there is cloud   */
      /* -------------------------------------------------------- */
      LQC = QN > 0.0;
    }

    /* ---------------------------------------------------------------- */
    /* 4. Form index of points where non-zero liquid cloud fraction     */
    /* ---------------------------------------------------------------- */


    /* ---------------------------------------------------------------- */
    /* 5. Call LS_CLD_C to calculate cloud water content, specific      */
    /*    humidity, water cloud fraction and determine temperature.     */
    /* ---------------------------------------------------------------- */
    if ( LQC ) {
      LS_CLD_C(P[K], RHCRIT[K], QSL, QN,
        &Q[K], &T[K], &QCL[K], &CFL[K], &GRID_QC[K], &BS[K]);
    }

    /* ---------------------------------------------------------------- */
    /* 6. Calculate cloud fractions for ice clouds.                     */
    /*    THIS IS STILL HIGHLY EXPERIMENTAL.                            */
    /*    Begin by calculating Qsat_wat(T,P), at Temp. T, for estimate  */
    /*    of bs.                                                        */
    /* ---------------------------------------------------------------- */

    /* Check for last level with cloud and update cloud top temperature */
    if ( LLWIC != K + 1 ) {
      CTT = T[K];
    }

    /* Check for significant ice content and update last level
     * with ice cloud
     */
    if ( QCF[K] > QCFMIN ) {
      LLWIC = K;
    }

    QSAT_WAT(&QSL, &T[K], &P[K], 1);

    if ( ICE_FRACTION_METHOD == 2 ) {
      /* Use cloud top temperature and a fixed qsat to give QCFRBS */
      QSAT_WAT(&QSL_CTT, &CTT, &P[K], 1);

      SUBGRID = pow(SUB_CLD, (1.0 - T_WEIGHT))
        / pow(QSAT_FIXED, (1.0 - T_WEIGHT - CTT_WEIGHT));
    }


    /* ------------------------------------------------------------ */
    /* 6a Calculate qCF/bs.                                         */
    /* ------------------------------------------------------------ */
    if ( RHCRIT[K] < 1.0 ) {
      if ( ICE_FRACTION_METHOD == 1 ) {
        QCFRBS =  QCF[K] / ((1.0 - RHCRIT[K]) * QSL);
      }
      else if ( ICE_FRACTION_METHOD == 2 ) {
        QCFRBS = SUBGRID * QCF[K]
          / ((1.0 - RHCRIT[K]) * pow(QSL_CTT, CTT_WEIGHT)
            * pow(QSL, T_WEIGHT));
      }

      /* ---------------------------------------------------------- */
      /* 6b Calculate frozen cloud fraction from frozen cloud water */
      /*    content.                                                */
      /* ---------------------------------------------------------- */
      if ( QCFRBS <= 0.0 ) {
        CFF[K] = 0.0;
      }
      else if ( 0.0 < QCFRBS && (6.0 * QCFRBS) <= 1.0 ) {
        CFF[K] = 0.5 * pow(6.0 * QCFRBS, 2.0 / 3.0);
      }
      else if ( 1.0 < (6.0 * QCFRBS ) && QCFRBS < 1.0 ) {
        PHIQCF = acos(ROOTWO * 0.75 * (1.0 - QCFRBS));
        COSQCF = cos((PHIQCF + (4.0 * PI)) / 3.0);
        CFF[K] = 1.0 - (4.0 * COSQCF * COSQCF);
      }
      else if ( QCFRBS >= 1.0 ) {
        CFF[K] = 1.0;
      }
    }
    else { /* RHcrit = 1, set cloud fraction to 1 or 0 */
      if ( QCF[K] > 0.0 ) {
        CFF[K] = 1.0;
      }
      else {
        CFF[K] = 0.0;
      }
    }
                                                                          
    /* ------------------------------------------------------------ */
    /* 6c Calculate combined cloud fraction.                        */
    /* ------------------------------------------------------------ */
    if ( CLOUD_FRACTION_METHOD == 1 ) {
      /* Use minimum overlap condition */
      CF[K] = MIN(CFL[K] + CFF[K], 1.0);
    }
    else if ( CLOUD_FRACTION_METHOD == 2 ) {
      /* Calculate possible overlaps between ice and liquid in
       * THIS layer
       */
      OVERLAP_MAX    = MIN(CFL[K], CFF[K]);
      OVERLAP_MIN    = MAX(CFL[K] + CFF[K] - 1.0, 0.0);
      OVERLAP_RANDOM = CFL[K] * CFF[K];

      /* Now use the specified degree of overlap to calculate the
       * total cloud fraction (= cfice + cfliq - overlap)
       */
      TEMP0 = OVERLAP_RANDOM;
      TEMP1 = 0.5 * (OVERLAP_MAX - OVERLAP_MIN);
      TEMP2 = 0.5 * (OVERLAP_MAX + OVERLAP_MIN) - OVERLAP_RANDOM;
      CF[K] = CFL[K] + CFF[K] - (TEMP0 + TEMP1 * OVERLAP_ICE_LIQUID
        + TEMP2 * OVERLAP_ICE_LIQUID * OVERLAP_ICE_LIQUID);

      /* Check that the overlap wasnt negative */
      CF[K] = MIN(CF[K], CFL[K] + CFF[K]);
    }
  }

  /**********************************************************************/

  free(GRID_QC);
  free(BS);

  return 0;


#undef alloc_real
}


/************************************************************************/
/************************************************************************/


REAL ABS(REAL x)
{
  return fabs(x);
}

REAL MAX(REAL x, REAL y)
{
  return ( x > y ) ? x : y;
}

REAL MIN(REAL x, REAL y)
{
  return ( x < y ) ? x : y;
}

/* Returns nearest integer to x. */
INTEGER NINT(REAL x)
{
  double fp;  /* fractional result from modf function. */
  double ip;  /* integer result from modf function. */

  fp = modf(x, &ip);
  return (INTEGER) (( fabs(fp) < 0.5 )
    ? ip
    : ( x < 0.0 )
      ? ip - 1.0
      : ip + 1.0);
}

/************************************************************************/

/* Note: 0 element is a repeat of 1st element to cater for special case
 * of low temperatures (<= T_LOW) for which the array index is
 * rounded down due to machine precision.
 */
static REAL QSAT_LUT[1553] = {
  0.966483E-02,
  0.966483E-02,0.984279E-02,0.100240E-01,0.102082E-01,0.103957E-01,
  0.105865E-01,0.107803E-01,0.109777E-01,0.111784E-01,0.113825E-01,
  0.115902E-01,0.118016E-01,0.120164E-01,0.122348E-01,0.124572E-01,
  0.126831E-01,0.129132E-01,0.131470E-01,0.133846E-01,0.136264E-01,
  0.138724E-01,0.141225E-01,0.143771E-01,0.146356E-01,0.148985E-01,
  0.151661E-01,0.154379E-01,0.157145E-01,0.159958E-01,0.162817E-01,
  0.165725E-01,0.168680E-01,0.171684E-01,0.174742E-01,0.177847E-01,
  0.181008E-01,0.184216E-01,0.187481E-01,0.190801E-01,0.194175E-01,
  0.197608E-01,0.201094E-01,0.204637E-01,0.208242E-01,0.211906E-01,
  0.215631E-01,0.219416E-01,0.223263E-01,0.227172E-01,0.231146E-01,
  0.235188E-01,0.239296E-01,0.243465E-01,0.247708E-01,0.252019E-01,
  0.256405E-01,0.260857E-01,0.265385E-01,0.269979E-01,0.274656E-01,
  0.279405E-01,0.284232E-01,0.289142E-01,0.294124E-01,0.299192E-01,
  0.304341E-01,0.309571E-01,0.314886E-01,0.320285E-01,0.325769E-01,
  0.331348E-01,0.337014E-01,0.342771E-01,0.348618E-01,0.354557E-01,
  0.360598E-01,0.366727E-01,0.372958E-01,0.379289E-01,0.385717E-01,
  0.392248E-01,0.398889E-01,0.405633E-01,0.412474E-01,0.419430E-01,
  0.426505E-01,0.433678E-01,0.440974E-01,0.448374E-01,0.455896E-01,
  0.463545E-01,0.471303E-01,0.479191E-01,0.487190E-01,0.495322E-01,
  0.503591E-01,0.511977E-01,0.520490E-01,0.529145E-01,0.537931E-01,
  0.546854E-01,0.555924E-01,0.565119E-01,0.574467E-01,0.583959E-01,
  0.593592E-01,0.603387E-01,0.613316E-01,0.623409E-01,0.633655E-01,
  0.644053E-01,0.654624E-01,0.665358E-01,0.676233E-01,0.687302E-01,
  0.698524E-01,0.709929E-01,0.721490E-01,0.733238E-01,0.745180E-01,
  0.757281E-01,0.769578E-01,0.782061E-01,0.794728E-01,0.807583E-01,
  0.820647E-01,0.833905E-01,0.847358E-01,0.861028E-01,0.874882E-01,
  0.888957E-01,0.903243E-01,0.917736E-01,0.932464E-01,0.947407E-01,
  0.962571E-01,0.977955E-01,0.993584E-01,0.100942E+00,0.102551E+00,
  0.104186E+00,0.105842E+00,0.107524E+00,0.109231E+00,0.110963E+00,
  0.112722E+00,0.114506E+00,0.116317E+00,0.118153E+00,0.120019E+00,
  0.121911E+00,0.123831E+00,0.125778E+00,0.127755E+00,0.129761E+00,
  0.131796E+00,0.133863E+00,0.135956E+00,0.138082E+00,0.140241E+00,
  0.142428E+00,0.144649E+00,0.146902E+00,0.149190E+00,0.151506E+00,
  0.153859E+00,0.156245E+00,0.158669E+00,0.161126E+00,0.163618E+00,
  0.166145E+00,0.168711E+00,0.171313E+00,0.173951E+00,0.176626E+00,
  0.179342E+00,0.182096E+00,0.184893E+00,0.187724E+00,0.190600E+00,
  0.193518E+00,0.196473E+00,0.199474E+00,0.202516E+00,0.205604E+00,
  0.208730E+00,0.211905E+00,0.215127E+00,0.218389E+00,0.221701E+00,
  0.225063E+00,0.228466E+00,0.231920E+00,0.235421E+00,0.238976E+00,
  0.242580E+00,0.246232E+00,0.249933E+00,0.253691E+00,0.257499E+00,
  0.261359E+00,0.265278E+00,0.269249E+00,0.273274E+00,0.277358E+00,
  0.281498E+00,0.285694E+00,0.289952E+00,0.294268E+00,0.298641E+00,
  0.303078E+00,0.307577E+00,0.312135E+00,0.316753E+00,0.321440E+00,
  0.326196E+00,0.331009E+00,0.335893E+00,0.340842E+00,0.345863E+00,
  0.350951E+00,0.356106E+00,0.361337E+00,0.366636E+00,0.372006E+00,
  0.377447E+00,0.382966E+00,0.388567E+00,0.394233E+00,0.399981E+00,
  0.405806E+00,0.411714E+00,0.417699E+00,0.423772E+00,0.429914E+00,
  0.436145E+00,0.442468E+00,0.448862E+00,0.455359E+00,0.461930E+00,
  0.468596E+00,0.475348E+00,0.482186E+00,0.489124E+00,0.496160E+00,
  0.503278E+00,0.510497E+00,0.517808E+00,0.525224E+00,0.532737E+00,
  0.540355E+00,0.548059E+00,0.555886E+00,0.563797E+00,0.571825E+00,
  0.579952E+00,0.588198E+00,0.596545E+00,0.605000E+00,0.613572E+00,
  0.622255E+00,0.631059E+00,0.639962E+00,0.649003E+00,0.658144E+00,
  0.667414E+00,0.676815E+00,0.686317E+00,0.695956E+00,0.705728E+00,
  0.715622E+00,0.725641E+00,0.735799E+00,0.746082E+00,0.756495E+00,
  0.767052E+00,0.777741E+00,0.788576E+00,0.799549E+00,0.810656E+00,
  0.821914E+00,0.833314E+00,0.844854E+00,0.856555E+00,0.868415E+00,
  0.880404E+00,0.892575E+00,0.904877E+00,0.917350E+00,0.929974E+00,
  0.942771E+00,0.955724E+00,0.968837E+00,0.982127E+00,0.995600E+00,
  0.100921E+01,0.102304E+01,0.103700E+01,0.105116E+01,0.106549E+01,
  0.108002E+01,0.109471E+01,0.110962E+01,0.112469E+01,0.113995E+01,
  0.115542E+01,0.117107E+01,0.118693E+01,0.120298E+01,0.121923E+01,
  0.123569E+01,0.125234E+01,0.126923E+01,0.128631E+01,0.130362E+01,
  0.132114E+01,0.133887E+01,0.135683E+01,0.137500E+01,0.139342E+01,
  0.141205E+01,0.143091E+01,0.145000E+01,0.146933E+01,0.148892E+01,
  0.150874E+01,0.152881E+01,0.154912E+01,0.156970E+01,0.159049E+01,
  0.161159E+01,0.163293E+01,0.165452E+01,0.167640E+01,0.169852E+01,
  0.172091E+01,0.174359E+01,0.176653E+01,0.178977E+01,0.181332E+01,
  0.183709E+01,0.186119E+01,0.188559E+01,0.191028E+01,0.193524E+01,
  0.196054E+01,0.198616E+01,0.201208E+01,0.203829E+01,0.206485E+01,
  0.209170E+01,0.211885E+01,0.214637E+01,0.217424E+01,0.220242E+01,
  0.223092E+01,0.225979E+01,0.228899E+01,0.231855E+01,0.234845E+01,
  0.237874E+01,0.240937E+01,0.244040E+01,0.247176E+01,0.250349E+01,
  0.253560E+01,0.256814E+01,0.260099E+01,0.263431E+01,0.266800E+01,
  0.270207E+01,0.273656E+01,0.277145E+01,0.280671E+01,0.284248E+01,
  0.287859E+01,0.291516E+01,0.295219E+01,0.298962E+01,0.302746E+01,
  0.306579E+01,0.310454E+01,0.314377E+01,0.318351E+01,0.322360E+01,
  0.326427E+01,0.330538E+01,0.334694E+01,0.338894E+01,0.343155E+01,
  0.347456E+01,0.351809E+01,0.356216E+01,0.360673E+01,0.365184E+01,
  0.369744E+01,0.374352E+01,0.379018E+01,0.383743E+01,0.388518E+01,
  0.393344E+01,0.398230E+01,0.403177E+01,0.408175E+01,0.413229E+01,
  0.418343E+01,0.423514E+01,0.428746E+01,0.434034E+01,0.439389E+01,
  0.444808E+01,0.450276E+01,0.455820E+01,0.461423E+01,0.467084E+01,
  0.472816E+01,0.478607E+01,0.484468E+01,0.490393E+01,0.496389E+01,
  0.502446E+01,0.508580E+01,0.514776E+01,0.521047E+01,0.527385E+01,
  0.533798E+01,0.540279E+01,0.546838E+01,0.553466E+01,0.560173E+01,
  0.566949E+01,0.573807E+01,0.580750E+01,0.587749E+01,0.594846E+01,
  0.602017E+01,0.609260E+01,0.616591E+01,0.623995E+01,0.631490E+01,
  0.639061E+01,0.646723E+01,0.654477E+01,0.662293E+01,0.670220E+01,
  0.678227E+01,0.686313E+01,0.694495E+01,0.702777E+01,0.711142E+01,
  0.719592E+01,0.728140E+01,0.736790E+01,0.745527E+01,0.754352E+01,
  0.763298E+01,0.772316E+01,0.781442E+01,0.790676E+01,0.800001E+01,
  0.809435E+01,0.818967E+01,0.828606E+01,0.838343E+01,0.848194E+01,
  0.858144E+01,0.868207E+01,0.878392E+01,0.888673E+01,0.899060E+01,
  0.909567E+01,0.920172E+01,0.930909E+01,0.941765E+01,0.952730E+01,
  0.963821E+01,0.975022E+01,0.986352E+01,0.997793E+01,0.100937E+02,
  0.102105E+02,0.103287E+02,0.104481E+02,0.105688E+02,0.106909E+02,
  0.108143E+02,0.109387E+02,0.110647E+02,0.111921E+02,0.113207E+02,
  0.114508E+02,0.115821E+02,0.117149E+02,0.118490E+02,0.119847E+02,
  0.121216E+02,0.122601E+02,0.124002E+02,0.125416E+02,0.126846E+02,
  0.128290E+02,0.129747E+02,0.131224E+02,0.132712E+02,0.134220E+02,
  0.135742E+02,0.137278E+02,0.138831E+02,0.140403E+02,0.141989E+02,
  0.143589E+02,0.145211E+02,0.146845E+02,0.148501E+02,0.150172E+02,
  0.151858E+02,0.153564E+02,0.155288E+02,0.157029E+02,0.158786E+02,
  0.160562E+02,0.162358E+02,0.164174E+02,0.166004E+02,0.167858E+02,
  0.169728E+02,0.171620E+02,0.173528E+02,0.175455E+02,0.177406E+02,
  0.179372E+02,0.181363E+02,0.183372E+02,0.185400E+02,0.187453E+02,
  0.189523E+02,0.191613E+02,0.193728E+02,0.195866E+02,0.198024E+02,
  0.200200E+02,0.202401E+02,0.204626E+02,0.206871E+02,0.209140E+02,
  0.211430E+02,0.213744E+02,0.216085E+02,0.218446E+02,0.220828E+02,
  0.223241E+02,0.225671E+02,0.228132E+02,0.230615E+02,0.233120E+02,
  0.235651E+02,0.238211E+02,0.240794E+02,0.243404E+02,0.246042E+02,
  0.248704E+02,0.251390E+02,0.254109E+02,0.256847E+02,0.259620E+02,
  0.262418E+02,0.265240E+02,0.268092E+02,0.270975E+02,0.273883E+02,
  0.276822E+02,0.279792E+02,0.282789E+02,0.285812E+02,0.288867E+02,
  0.291954E+02,0.295075E+02,0.298222E+02,0.301398E+02,0.304606E+02,
  0.307848E+02,0.311119E+02,0.314424E+02,0.317763E+02,0.321133E+02,
  0.324536E+02,0.327971E+02,0.331440E+02,0.334940E+02,0.338475E+02,
  0.342050E+02,0.345654E+02,0.349295E+02,0.352975E+02,0.356687E+02,
  0.360430E+02,0.364221E+02,0.368042E+02,0.371896E+02,0.375790E+02,
  0.379725E+02,0.383692E+02,0.387702E+02,0.391744E+02,0.395839E+02,
  0.399958E+02,0.404118E+02,0.408325E+02,0.412574E+02,0.416858E+02,
  0.421188E+02,0.425551E+02,0.429962E+02,0.434407E+02,0.438910E+02,
  0.443439E+02,0.448024E+02,0.452648E+02,0.457308E+02,0.462018E+02,
  0.466775E+02,0.471582E+02,0.476428E+02,0.481313E+02,0.486249E+02,
  0.491235E+02,0.496272E+02,0.501349E+02,0.506479E+02,0.511652E+02,
  0.516876E+02,0.522142E+02,0.527474E+02,0.532836E+02,0.538266E+02,
  0.543737E+02,0.549254E+02,0.554839E+02,0.560456E+02,0.566142E+02,
  0.571872E+02,0.577662E+02,0.583498E+02,0.589392E+02,0.595347E+02,
  0.601346E+02,0.607410E+02,0.613519E+02,0.619689E+02,0.625922E+02,
  0.632204E+02,0.638550E+02,0.644959E+02,0.651418E+02,0.657942E+02,
  0.664516E+02,0.671158E+02,0.677864E+02,0.684624E+02,0.691451E+02,
  0.698345E+02,0.705293E+02,0.712312E+02,0.719398E+02,0.726542E+02,
  0.733754E+02,0.741022E+02,0.748363E+02,0.755777E+02,0.763247E+02,
  0.770791E+02,0.778394E+02,0.786088E+02,0.793824E+02,0.801653E+02,
  0.809542E+02,0.817509E+02,0.825536E+02,0.833643E+02,0.841828E+02,
  0.850076E+02,0.858405E+02,0.866797E+02,0.875289E+02,0.883827E+02,
  0.892467E+02,0.901172E+02,0.909962E+02,0.918818E+02,0.927760E+02,
  0.936790E+02,0.945887E+02,0.955071E+02,0.964346E+02,0.973689E+02,
  0.983123E+02,0.992648E+02,0.100224E+03,0.101193E+03,0.102169E+03,
  0.103155E+03,0.104150E+03,0.105152E+03,0.106164E+03,0.107186E+03,
  0.108217E+03,0.109256E+03,0.110303E+03,0.111362E+03,0.112429E+03,
  0.113503E+03,0.114588E+03,0.115684E+03,0.116789E+03,0.117903E+03,
  0.119028E+03,0.120160E+03,0.121306E+03,0.122460E+03,0.123623E+03,
  0.124796E+03,0.125981E+03,0.127174E+03,0.128381E+03,0.129594E+03,
  0.130822E+03,0.132058E+03,0.133306E+03,0.134563E+03,0.135828E+03,
  0.137109E+03,0.138402E+03,0.139700E+03,0.141017E+03,0.142338E+03,
  0.143676E+03,0.145025E+03,0.146382E+03,0.147753E+03,0.149133E+03,
  0.150529E+03,0.151935E+03,0.153351E+03,0.154783E+03,0.156222E+03,
  0.157678E+03,0.159148E+03,0.160624E+03,0.162117E+03,0.163621E+03,
  0.165142E+03,0.166674E+03,0.168212E+03,0.169772E+03,0.171340E+03,
  0.172921E+03,0.174522E+03,0.176129E+03,0.177755E+03,0.179388E+03,
  0.181040E+03,0.182707E+03,0.184382E+03,0.186076E+03,0.187782E+03,
  0.189503E+03,0.191240E+03,0.192989E+03,0.194758E+03,0.196535E+03,
  0.198332E+03,0.200141E+03,0.201963E+03,0.203805E+03,0.205656E+03,
  0.207532E+03,0.209416E+03,0.211317E+03,0.213236E+03,0.215167E+03,
  0.217121E+03,0.219087E+03,0.221067E+03,0.223064E+03,0.225080E+03,
  0.227113E+03,0.229160E+03,0.231221E+03,0.233305E+03,0.235403E+03,
  0.237520E+03,0.239655E+03,0.241805E+03,0.243979E+03,0.246163E+03,
  0.248365E+03,0.250593E+03,0.252830E+03,0.255093E+03,0.257364E+03,
  0.259667E+03,0.261979E+03,0.264312E+03,0.266666E+03,0.269034E+03,
  0.271430E+03,0.273841E+03,0.276268E+03,0.278722E+03,0.281185E+03,
  0.283677E+03,0.286190E+03,0.288714E+03,0.291266E+03,0.293834E+03,
  0.296431E+03,0.299045E+03,0.301676E+03,0.304329E+03,0.307006E+03,
  0.309706E+03,0.312423E+03,0.315165E+03,0.317930E+03,0.320705E+03,
  0.323519E+03,0.326350E+03,0.329199E+03,0.332073E+03,0.334973E+03,
  0.337897E+03,0.340839E+03,0.343800E+03,0.346794E+03,0.349806E+03,
  0.352845E+03,0.355918E+03,0.358994E+03,0.362112E+03,0.365242E+03,
  0.368407E+03,0.371599E+03,0.374802E+03,0.378042E+03,0.381293E+03,
  0.384588E+03,0.387904E+03,0.391239E+03,0.394604E+03,0.397988E+03,
  0.401411E+03,0.404862E+03,0.408326E+03,0.411829E+03,0.415352E+03,
  0.418906E+03,0.422490E+03,0.426095E+03,0.429740E+03,0.433398E+03,
  0.437097E+03,0.440827E+03,0.444570E+03,0.448354E+03,0.452160E+03,
  0.455999E+03,0.459870E+03,0.463765E+03,0.467702E+03,0.471652E+03,
  0.475646E+03,0.479674E+03,0.483715E+03,0.487811E+03,0.491911E+03,
  0.496065E+03,0.500244E+03,0.504448E+03,0.508698E+03,0.512961E+03,
  0.517282E+03,0.521617E+03,0.525989E+03,0.530397E+03,0.534831E+03,
  0.539313E+03,0.543821E+03,0.548355E+03,0.552938E+03,0.557549E+03,
  0.562197E+03,0.566884E+03,0.571598E+03,0.576351E+03,0.581131E+03,
  0.585963E+03,0.590835E+03,0.595722E+03,0.600663E+03,0.605631E+03,
  0.610641E+03,0.615151E+03,0.619625E+03,0.624140E+03,0.628671E+03,
  0.633243E+03,0.637845E+03,0.642465E+03,0.647126E+03,0.651806E+03,
  0.656527E+03,0.661279E+03,0.666049E+03,0.670861E+03,0.675692E+03,
  0.680566E+03,0.685471E+03,0.690396E+03,0.695363E+03,0.700350E+03,
  0.705381E+03,0.710444E+03,0.715527E+03,0.720654E+03,0.725801E+03,
  0.730994E+03,0.736219E+03,0.741465E+03,0.746756E+03,0.752068E+03,
  0.757426E+03,0.762819E+03,0.768231E+03,0.773692E+03,0.779172E+03,
  0.784701E+03,0.790265E+03,0.795849E+03,0.801483E+03,0.807137E+03,
  0.812842E+03,0.818582E+03,0.824343E+03,0.830153E+03,0.835987E+03,
  0.841871E+03,0.847791E+03,0.853733E+03,0.859727E+03,0.865743E+03,
  0.871812E+03,0.877918E+03,0.884046E+03,0.890228E+03,0.896433E+03,
  0.902690E+03,0.908987E+03,0.915307E+03,0.921681E+03,0.928078E+03,
  0.934531E+03,0.941023E+03,0.947539E+03,0.954112E+03,0.960708E+03,
  0.967361E+03,0.974053E+03,0.980771E+03,0.987545E+03,0.994345E+03,
  0.100120E+04,0.100810E+04,0.101502E+04,0.102201E+04,0.102902E+04,
  0.103608E+04,0.104320E+04,0.105033E+04,0.105753E+04,0.106475E+04,
  0.107204E+04,0.107936E+04,0.108672E+04,0.109414E+04,0.110158E+04,
  0.110908E+04,0.111663E+04,0.112421E+04,0.113185E+04,0.113952E+04,
  0.114725E+04,0.115503E+04,0.116284E+04,0.117071E+04,0.117861E+04,
  0.118658E+04,0.119459E+04,0.120264E+04,0.121074E+04,0.121888E+04,
  0.122709E+04,0.123534E+04,0.124362E+04,0.125198E+04,0.126036E+04,
  0.126881E+04,0.127731E+04,0.128584E+04,0.129444E+04,0.130307E+04,
  0.131177E+04,0.132053E+04,0.132931E+04,0.133817E+04,0.134705E+04,
  0.135602E+04,0.136503E+04,0.137407E+04,0.138319E+04,0.139234E+04,
  0.140156E+04,0.141084E+04,0.142015E+04,0.142954E+04,0.143896E+04,
  0.144845E+04,0.145800E+04,0.146759E+04,0.147725E+04,0.148694E+04,
  0.149672E+04,0.150655E+04,0.151641E+04,0.152635E+04,0.153633E+04,
  0.154639E+04,0.155650E+04,0.156665E+04,0.157688E+04,0.158715E+04,
  0.159750E+04,0.160791E+04,0.161836E+04,0.162888E+04,0.163945E+04,
  0.165010E+04,0.166081E+04,0.167155E+04,0.168238E+04,0.169325E+04,
  0.170420E+04,0.171522E+04,0.172627E+04,0.173741E+04,0.174859E+04,
  0.175986E+04,0.177119E+04,0.178256E+04,0.179402E+04,0.180552E+04,
  0.181711E+04,0.182877E+04,0.184046E+04,0.185224E+04,0.186407E+04,
  0.187599E+04,0.188797E+04,0.190000E+04,0.191212E+04,0.192428E+04,
  0.193653E+04,0.194886E+04,0.196122E+04,0.197368E+04,0.198618E+04,
  0.199878E+04,0.201145E+04,0.202416E+04,0.203698E+04,0.204983E+04,
  0.206278E+04,0.207580E+04,0.208887E+04,0.210204E+04,0.211525E+04,
  0.212856E+04,0.214195E+04,0.215538E+04,0.216892E+04,0.218249E+04,
  0.219618E+04,0.220994E+04,0.222375E+04,0.223766E+04,0.225161E+04,
  0.226567E+04,0.227981E+04,0.229399E+04,0.230829E+04,0.232263E+04,
  0.233708E+04,0.235161E+04,0.236618E+04,0.238087E+04,0.239560E+04,
  0.241044E+04,0.242538E+04,0.244035E+04,0.245544E+04,0.247057E+04,
  0.248583E+04,0.250116E+04,0.251654E+04,0.253204E+04,0.254759E+04,
  0.256325E+04,0.257901E+04,0.259480E+04,0.261073E+04,0.262670E+04,
  0.264279E+04,0.265896E+04,0.267519E+04,0.269154E+04,0.270794E+04,
  0.272447E+04,0.274108E+04,0.275774E+04,0.277453E+04,0.279137E+04,
  0.280834E+04,0.282540E+04,0.284251E+04,0.285975E+04,0.287704E+04,
  0.289446E+04,0.291198E+04,0.292954E+04,0.294725E+04,0.296499E+04,
  0.298288E+04,0.300087E+04,0.301890E+04,0.303707E+04,0.305529E+04,
  0.307365E+04,0.309211E+04,0.311062E+04,0.312927E+04,0.314798E+04,
  0.316682E+04,0.318577E+04,0.320477E+04,0.322391E+04,0.324310E+04,
  0.326245E+04,0.328189E+04,0.330138E+04,0.332103E+04,0.334073E+04,
  0.336058E+04,0.338053E+04,0.340054E+04,0.342069E+04,0.344090E+04,
  0.346127E+04,0.348174E+04,0.350227E+04,0.352295E+04,0.354369E+04,
  0.356458E+04,0.358559E+04,0.360664E+04,0.362787E+04,0.364914E+04,
  0.367058E+04,0.369212E+04,0.371373E+04,0.373548E+04,0.375731E+04,
  0.377929E+04,0.380139E+04,0.382355E+04,0.384588E+04,0.386826E+04,
  0.389081E+04,0.391348E+04,0.393620E+04,0.395910E+04,0.398205E+04,
  0.400518E+04,0.402843E+04,0.405173E+04,0.407520E+04,0.409875E+04,
  0.412246E+04,0.414630E+04,0.417019E+04,0.419427E+04,0.421840E+04,
  0.424272E+04,0.426715E+04,0.429165E+04,0.431634E+04,0.434108E+04,
  0.436602E+04,0.439107E+04,0.441618E+04,0.444149E+04,0.446685E+04,
  0.449241E+04,0.451810E+04,0.454385E+04,0.456977E+04,0.459578E+04,
  0.462197E+04,0.464830E+04,0.467468E+04,0.470127E+04,0.472792E+04,
  0.475477E+04,0.478175E+04,0.480880E+04,0.483605E+04,0.486336E+04,
  0.489087E+04,0.491853E+04,0.494623E+04,0.497415E+04,0.500215E+04,
  0.503034E+04,0.505867E+04,0.508707E+04,0.511568E+04,0.514436E+04,
  0.517325E+04,0.520227E+04,0.523137E+04,0.526068E+04,0.529005E+04,
  0.531965E+04,0.534939E+04,0.537921E+04,0.540923E+04,0.543932E+04,
  0.546965E+04,0.550011E+04,0.553064E+04,0.556139E+04,0.559223E+04,
  0.562329E+04,0.565449E+04,0.568577E+04,0.571727E+04,0.574884E+04,
  0.578064E+04,0.581261E+04,0.584464E+04,0.587692E+04,0.590924E+04,
  0.594182E+04,0.597455E+04,0.600736E+04,0.604039E+04,0.607350E+04,
  0.610685E+04,0.614036E+04,0.617394E+04,0.620777E+04,0.624169E+04,
  0.627584E+04,0.631014E+04,0.634454E+04,0.637918E+04,0.641390E+04,
  0.644887E+04,0.648400E+04,0.651919E+04,0.655467E+04,0.659021E+04,
  0.662599E+04,0.666197E+04,0.669800E+04,0.673429E+04,0.677069E+04,
  0.680735E+04,0.684415E+04,0.688104E+04,0.691819E+04,0.695543E+04,
  0.699292E+04,0.703061E+04,0.706837E+04,0.710639E+04,0.714451E+04,
  0.718289E+04,0.722143E+04,0.726009E+04,0.729903E+04,0.733802E+04,
  0.737729E+04,0.741676E+04,0.745631E+04,0.749612E+04,0.753602E+04,
  0.757622E+04,0.761659E+04,0.765705E+04,0.769780E+04,0.773863E+04,
  0.777975E+04,0.782106E+04,0.786246E+04,0.790412E+04,0.794593E+04,
  0.798802E+04,0.803028E+04,0.807259E+04,0.811525E+04,0.815798E+04,
  0.820102E+04,0.824427E+04,0.828757E+04,0.833120E+04,0.837493E+04,
  0.841895E+04,0.846313E+04,0.850744E+04,0.855208E+04,0.859678E+04,
  0.864179E+04,0.868705E+04,0.873237E+04,0.877800E+04,0.882374E+04,
  0.886979E+04,0.891603E+04,0.896237E+04,0.900904E+04,0.905579E+04,
  0.910288E+04,0.915018E+04,0.919758E+04,0.924529E+04,0.929310E+04,
  0.934122E+04,0.938959E+04,0.943804E+04,0.948687E+04,0.953575E+04,
  0.958494E+04,0.963442E+04,0.968395E+04,0.973384E+04,0.978383E+04,
  0.983412E+04,0.988468E+04,0.993534E+04,0.998630E+04,0.100374E+05,
  0.100888E+05,0.101406E+05,0.101923E+05,0.102444E+05,0.102966E+05,
  0.103492E+05,0.104020E+05,0.104550E+05,0.105082E+05,0.105616E+05,
  0.106153E+05,0.106693E+05,0.107234E+05,0.107779E+05,0.108325E+05,
  0.108874E+05,0.109425E+05,0.109978E+05,0.110535E+05,0.111092E+05,
  0.111653E+05,0.112217E+05,0.112782E+05,0.113350E+05,0.113920E+05,
  0.114493E+05,0.115070E+05,0.115646E+05,0.116228E+05,0.116809E+05,
  0.117396E+05,0.117984E+05,0.118574E+05,0.119167E+05,0.119762E+05,
  0.120360E+05,0.120962E+05,0.121564E+05,0.122170E+05,0.122778E+05,
  0.123389E+05,0.124004E+05,0.124619E+05,0.125238E+05,0.125859E+05,
  0.126484E+05,0.127111E+05,0.127739E+05,0.128372E+05,0.129006E+05,
  0.129644E+05,0.130285E+05,0.130927E+05,0.131573E+05,0.132220E+05,
  0.132872E+05,0.133526E+05,0.134182E+05,0.134842E+05,0.135503E+05,
  0.136168E+05,0.136836E+05,0.137505E+05,0.138180E+05,0.138854E+05,
  0.139534E+05,0.140216E+05,0.140900E+05,0.141588E+05,0.142277E+05,
  0.142971E+05,0.143668E+05,0.144366E+05,0.145069E+05,0.145773E+05,
  0.146481E+05,0.147192E+05,0.147905E+05,0.148622E+05,0.149341E+05,
  0.150064E+05,0.150790E+05,0.151517E+05,0.152250E+05,0.152983E+05,
  0.153721E+05,0.154462E+05,0.155205E+05,0.155952E+05,0.156701E+05,
  0.157454E+05,0.158211E+05,0.158969E+05,0.159732E+05,0.160496E+05,
  0.161265E+05,0.162037E+05,0.162811E+05,0.163589E+05,0.164369E+05,
  0.165154E+05,0.165942E+05,0.166732E+05,0.167526E+05,0.168322E+05,
  0.169123E+05,0.169927E+05,0.170733E+05,0.171543E+05,0.172356E+05,
  0.173173E+05,0.173993E+05,0.174815E+05,0.175643E+05,0.176471E+05,
  0.177305E+05,0.178143E+05,0.178981E+05,0.179826E+05,0.180671E+05,
  0.181522E+05,0.182377E+05,0.183232E+05,0.184093E+05,0.184955E+05,
  0.185823E+05,0.186695E+05,0.187568E+05,0.188447E+05,0.189326E+05,
  0.190212E+05,0.191101E+05,0.191991E+05,0.192887E+05,0.193785E+05,
  0.194688E+05,0.195595E+05,0.196503E+05,0.197417E+05,0.198332E+05,
  0.199253E+05,0.200178E+05,0.201105E+05,0.202036E+05,0.202971E+05,
  0.203910E+05,0.204853E+05,0.205798E+05,0.206749E+05,0.207701E+05,
  0.208659E+05,0.209621E+05,0.210584E+05,0.211554E+05,0.212524E+05,
  0.213501E+05,0.214482E+05,0.215465E+05,0.216452E+05,0.217442E+05,
  0.218439E+05,0.219439E+05,0.220440E+05,0.221449E+05,0.222457E+05,
  0.223473E+05,0.224494E+05,0.225514E+05,0.226542E+05,0.227571E+05,
  0.228606E+05,0.229646E+05,0.230687E+05,0.231734E+05,0.232783E+05,
  0.233839E+05,0.234898E+05,0.235960E+05,0.237027E+05,0.238097E+05,
  0.239173E+05,0.240254E+05,0.241335E+05,0.242424E+05,0.243514E+05,
  0.244611E+05,0.245712E+05,0.246814E+05,0.247923E+05,0.249034E+05,
  0.250152E+05,0.250152E+05
};


/* Note: 0 element is a repeat of 1st element to cater for special case
 * of low temperatures (<= T_LOW) for which the array index is
 * rounded down due to machine precision.
 */
static REAL QSAT_WAT_LUT[1553] = {
  0.186905E-01,
  0.186905E-01,0.190449E-01,0.194059E-01,0.197727E-01,0.201462E-01,
  0.205261E-01,0.209122E-01,0.213052E-01,0.217050E-01,0.221116E-01,
  0.225252E-01,0.229463E-01,0.233740E-01,0.238090E-01,0.242518E-01,
  0.247017E-01,0.251595E-01,0.256252E-01,0.260981E-01,0.265795E-01,
  0.270691E-01,0.275667E-01,0.280733E-01,0.285876E-01,0.291105E-01,
  0.296429E-01,0.301835E-01,0.307336E-01,0.312927E-01,0.318611E-01,
  0.324390E-01,0.330262E-01,0.336232E-01,0.342306E-01,0.348472E-01,
  0.354748E-01,0.361117E-01,0.367599E-01,0.374185E-01,0.380879E-01,
  0.387689E-01,0.394602E-01,0.401626E-01,0.408771E-01,0.416033E-01,
  0.423411E-01,0.430908E-01,0.438524E-01,0.446263E-01,0.454124E-01,
  0.462122E-01,0.470247E-01,0.478491E-01,0.486874E-01,0.495393E-01,
  0.504057E-01,0.512847E-01,0.521784E-01,0.530853E-01,0.540076E-01,
  0.549444E-01,0.558959E-01,0.568633E-01,0.578448E-01,0.588428E-01,
  0.598566E-01,0.608858E-01,0.619313E-01,0.629926E-01,0.640706E-01,
  0.651665E-01,0.662795E-01,0.674095E-01,0.685570E-01,0.697219E-01,
  0.709063E-01,0.721076E-01,0.733284E-01,0.745679E-01,0.758265E-01,
  0.771039E-01,0.784026E-01,0.797212E-01,0.810577E-01,0.824164E-01,
  0.837971E-01,0.851970E-01,0.866198E-01,0.880620E-01,0.895281E-01,
  0.910178E-01,0.925278E-01,0.940622E-01,0.956177E-01,0.971984E-01,
  0.988051E-01,0.100433E+00,0.102085E+00,0.103764E+00,0.105467E+00,
  0.107196E+00,0.108953E+00,0.110732E+00,0.112541E+00,0.114376E+00,
  0.116238E+00,0.118130E+00,0.120046E+00,0.121993E+00,0.123969E+00,
  0.125973E+00,0.128009E+00,0.130075E+00,0.132167E+00,0.134296E+00,
  0.136452E+00,0.138642E+00,0.140861E+00,0.143115E+00,0.145404E+00,
  0.147723E+00,0.150078E+00,0.152466E+00,0.154889E+00,0.157346E+00,
  0.159841E+00,0.162372E+00,0.164939E+00,0.167545E+00,0.170185E+00,
  0.172866E+00,0.175584E+00,0.178340E+00,0.181139E+00,0.183977E+00,
  0.186855E+00,0.189773E+00,0.192737E+00,0.195736E+00,0.198783E+00,
  0.201875E+00,0.205007E+00,0.208186E+00,0.211409E+00,0.214676E+00,
  0.217993E+00,0.221355E+00,0.224764E+00,0.228220E+00,0.231728E+00,
  0.235284E+00,0.238888E+00,0.242542E+00,0.246251E+00,0.250010E+00,
  0.253821E+00,0.257688E+00,0.261602E+00,0.265575E+00,0.269607E+00,
  0.273689E+00,0.277830E+00,0.282027E+00,0.286287E+00,0.290598E+00,
  0.294972E+00,0.299405E+00,0.303904E+00,0.308462E+00,0.313082E+00,
  0.317763E+00,0.322512E+00,0.327324E+00,0.332201E+00,0.337141E+00,
  0.342154E+00,0.347234E+00,0.352387E+00,0.357601E+00,0.362889E+00,
  0.368257E+00,0.373685E+00,0.379194E+00,0.384773E+00,0.390433E+00,
  0.396159E+00,0.401968E+00,0.407861E+00,0.413820E+00,0.419866E+00,
  0.425999E+00,0.432203E+00,0.438494E+00,0.444867E+00,0.451332E+00,
  0.457879E+00,0.464510E+00,0.471226E+00,0.478037E+00,0.484935E+00,
  0.491920E+00,0.499005E+00,0.506181E+00,0.513447E+00,0.520816E+00,
  0.528279E+00,0.535835E+00,0.543497E+00,0.551256E+00,0.559113E+00,
  0.567081E+00,0.575147E+00,0.583315E+00,0.591585E+00,0.599970E+00,
  0.608472E+00,0.617069E+00,0.625785E+00,0.634609E+00,0.643556E+00,
  0.652611E+00,0.661782E+00,0.671077E+00,0.680487E+00,0.690015E+00,
  0.699656E+00,0.709433E+00,0.719344E+00,0.729363E+00,0.739518E+00,
  0.749795E+00,0.760217E+00,0.770763E+00,0.781454E+00,0.792258E+00,
  0.803208E+00,0.814309E+00,0.825528E+00,0.836914E+00,0.848422E+00,
  0.860086E+00,0.871891E+00,0.883837E+00,0.895944E+00,0.908214E+00,
  0.920611E+00,0.933175E+00,0.945890E+00,0.958776E+00,0.971812E+00,
  0.985027E+00,0.998379E+00,0.101193E+01,0.102561E+01,0.103949E+01,
  0.105352E+01,0.106774E+01,0.108213E+01,0.109669E+01,0.111144E+01,
  0.112636E+01,0.114148E+01,0.115676E+01,0.117226E+01,0.118791E+01,
  0.120377E+01,0.121984E+01,0.123608E+01,0.125252E+01,0.126919E+01,
  0.128604E+01,0.130309E+01,0.132036E+01,0.133782E+01,0.135549E+01,
  0.137339E+01,0.139150E+01,0.140984E+01,0.142839E+01,0.144715E+01,
  0.146616E+01,0.148538E+01,0.150482E+01,0.152450E+01,0.154445E+01,
  0.156459E+01,0.158502E+01,0.160564E+01,0.162654E+01,0.164766E+01,
  0.166906E+01,0.169070E+01,0.171257E+01,0.173473E+01,0.175718E+01,
  0.177984E+01,0.180282E+01,0.182602E+01,0.184951E+01,0.187327E+01,
  0.189733E+01,0.192165E+01,0.194629E+01,0.197118E+01,0.199636E+01,
  0.202185E+01,0.204762E+01,0.207372E+01,0.210010E+01,0.212678E+01,
  0.215379E+01,0.218109E+01,0.220873E+01,0.223668E+01,0.226497E+01,
  0.229357E+01,0.232249E+01,0.235176E+01,0.238134E+01,0.241129E+01,
  0.244157E+01,0.247217E+01,0.250316E+01,0.253447E+01,0.256617E+01,
  0.259821E+01,0.263064E+01,0.266341E+01,0.269661E+01,0.273009E+01,
  0.276403E+01,0.279834E+01,0.283302E+01,0.286811E+01,0.290358E+01,
  0.293943E+01,0.297571E+01,0.301236E+01,0.304946E+01,0.308702E+01,
  0.312491E+01,0.316326E+01,0.320208E+01,0.324130E+01,0.328092E+01,
  0.332102E+01,0.336162E+01,0.340264E+01,0.344407E+01,0.348601E+01,
  0.352838E+01,0.357118E+01,0.361449E+01,0.365834E+01,0.370264E+01,
  0.374737E+01,0.379265E+01,0.383839E+01,0.388469E+01,0.393144E+01,
  0.397876E+01,0.402656E+01,0.407492E+01,0.412378E+01,0.417313E+01,
  0.422306E+01,0.427359E+01,0.432454E+01,0.437617E+01,0.442834E+01,
  0.448102E+01,0.453433E+01,0.458816E+01,0.464253E+01,0.469764E+01,
  0.475321E+01,0.480942E+01,0.486629E+01,0.492372E+01,0.498173E+01,
  0.504041E+01,0.509967E+01,0.515962E+01,0.522029E+01,0.528142E+01,
  0.534337E+01,0.540595E+01,0.546912E+01,0.553292E+01,0.559757E+01,
  0.566273E+01,0.572864E+01,0.579532E+01,0.586266E+01,0.593075E+01,
  0.599952E+01,0.606895E+01,0.613918E+01,0.621021E+01,0.628191E+01,
  0.635433E+01,0.642755E+01,0.650162E+01,0.657639E+01,0.665188E+01,
  0.672823E+01,0.680532E+01,0.688329E+01,0.696198E+01,0.704157E+01,
  0.712206E+01,0.720319E+01,0.728534E+01,0.736829E+01,0.745204E+01,
  0.753671E+01,0.762218E+01,0.770860E+01,0.779588E+01,0.788408E+01,
  0.797314E+01,0.806318E+01,0.815408E+01,0.824599E+01,0.833874E+01,
  0.843254E+01,0.852721E+01,0.862293E+01,0.871954E+01,0.881724E+01,
  0.891579E+01,0.901547E+01,0.911624E+01,0.921778E+01,0.932061E+01,
  0.942438E+01,0.952910E+01,0.963497E+01,0.974181E+01,0.984982E+01,
  0.995887E+01,0.100690E+02,0.101804E+02,0.102926E+02,0.104063E+02,
  0.105210E+02,0.106367E+02,0.107536E+02,0.108719E+02,0.109912E+02,
  0.111116E+02,0.112333E+02,0.113563E+02,0.114804E+02,0.116056E+02,
  0.117325E+02,0.118602E+02,0.119892E+02,0.121197E+02,0.122513E+02,
  0.123844E+02,0.125186E+02,0.126543E+02,0.127912E+02,0.129295E+02,
  0.130691E+02,0.132101E+02,0.133527E+02,0.134965E+02,0.136415E+02,
  0.137882E+02,0.139361E+02,0.140855E+02,0.142366E+02,0.143889E+02,
  0.145429E+02,0.146982E+02,0.148552E+02,0.150135E+02,0.151735E+02,
  0.153349E+02,0.154979E+02,0.156624E+02,0.158286E+02,0.159965E+02,
  0.161659E+02,0.163367E+02,0.165094E+02,0.166838E+02,0.168597E+02,
  0.170375E+02,0.172168E+02,0.173979E+02,0.175806E+02,0.177651E+02,
  0.179513E+02,0.181394E+02,0.183293E+02,0.185210E+02,0.187146E+02,
  0.189098E+02,0.191066E+02,0.193059E+02,0.195065E+02,0.197095E+02,
  0.199142E+02,0.201206E+02,0.203291E+02,0.205397E+02,0.207522E+02,
  0.209664E+02,0.211831E+02,0.214013E+02,0.216221E+02,0.218448E+02,
  0.220692E+02,0.222959E+02,0.225250E+02,0.227559E+02,0.229887E+02,
  0.232239E+02,0.234614E+02,0.237014E+02,0.239428E+02,0.241872E+02,
  0.244335E+02,0.246824E+02,0.249332E+02,0.251860E+02,0.254419E+02,
  0.256993E+02,0.259600E+02,0.262225E+02,0.264873E+02,0.267552E+02,
  0.270248E+02,0.272970E+02,0.275719E+02,0.278497E+02,0.281295E+02,
  0.284117E+02,0.286965E+02,0.289843E+02,0.292743E+02,0.295671E+02,
  0.298624E+02,0.301605E+02,0.304616E+02,0.307650E+02,0.310708E+02,
  0.313803E+02,0.316915E+02,0.320064E+02,0.323238E+02,0.326437E+02,
  0.329666E+02,0.332928E+02,0.336215E+02,0.339534E+02,0.342885E+02,
  0.346263E+02,0.349666E+02,0.353109E+02,0.356572E+02,0.360076E+02,
  0.363606E+02,0.367164E+02,0.370757E+02,0.374383E+02,0.378038E+02,
  0.381727E+02,0.385453E+02,0.389206E+02,0.392989E+02,0.396807E+02,
  0.400663E+02,0.404555E+02,0.408478E+02,0.412428E+02,0.416417E+02,
  0.420445E+02,0.424502E+02,0.428600E+02,0.432733E+02,0.436900E+02,
  0.441106E+02,0.445343E+02,0.449620E+02,0.453930E+02,0.458280E+02,
  0.462672E+02,0.467096E+02,0.471561E+02,0.476070E+02,0.480610E+02,
  0.485186E+02,0.489813E+02,0.494474E+02,0.499170E+02,0.503909E+02,
  0.508693E+02,0.513511E+02,0.518376E+02,0.523277E+02,0.528232E+02,
  0.533213E+02,0.538240E+02,0.543315E+02,0.548437E+02,0.553596E+02,
  0.558802E+02,0.564046E+02,0.569340E+02,0.574672E+02,0.580061E+02,
  0.585481E+02,0.590963E+02,0.596482E+02,0.602041E+02,0.607649E+02,
  0.613311E+02,0.619025E+02,0.624779E+02,0.630574E+02,0.636422E+02,
  0.642324E+02,0.648280E+02,0.654278E+02,0.660332E+02,0.666426E+02,
  0.672577E+02,0.678771E+02,0.685034E+02,0.691328E+02,0.697694E+02,
  0.704103E+02,0.710556E+02,0.717081E+02,0.723639E+02,0.730269E+02,
  0.736945E+02,0.743681E+02,0.750463E+02,0.757309E+02,0.764214E+02,
  0.771167E+02,0.778182E+02,0.785246E+02,0.792373E+02,0.799564E+02,
  0.806804E+02,0.814109E+02,0.821479E+02,0.828898E+02,0.836384E+02,
  0.843922E+02,0.851525E+02,0.859198E+02,0.866920E+02,0.874712E+02,
  0.882574E+02,0.890486E+02,0.898470E+02,0.906525E+02,0.914634E+02,
  0.922814E+02,0.931048E+02,0.939356E+02,0.947736E+02,0.956171E+02,
  0.964681E+02,0.973246E+02,0.981907E+02,0.990605E+02,0.999399E+02,
  0.100825E+03,0.101718E+03,0.102617E+03,0.103523E+03,0.104438E+03,
  0.105358E+03,0.106287E+03,0.107221E+03,0.108166E+03,0.109115E+03,
  0.110074E+03,0.111039E+03,0.112012E+03,0.112992E+03,0.113981E+03,
  0.114978E+03,0.115981E+03,0.116993E+03,0.118013E+03,0.119041E+03,
  0.120077E+03,0.121122E+03,0.122173E+03,0.123234E+03,0.124301E+03,
  0.125377E+03,0.126463E+03,0.127556E+03,0.128657E+03,0.129769E+03,
  0.130889E+03,0.132017E+03,0.133152E+03,0.134299E+03,0.135453E+03,
  0.136614E+03,0.137786E+03,0.138967E+03,0.140158E+03,0.141356E+03,
  0.142565E+03,0.143781E+03,0.145010E+03,0.146247E+03,0.147491E+03,
  0.148746E+03,0.150011E+03,0.151284E+03,0.152571E+03,0.153862E+03,
  0.155168E+03,0.156481E+03,0.157805E+03,0.159137E+03,0.160478E+03,
  0.161832E+03,0.163198E+03,0.164569E+03,0.165958E+03,0.167348E+03,
  0.168757E+03,0.170174E+03,0.171599E+03,0.173037E+03,0.174483E+03,
  0.175944E+03,0.177414E+03,0.178892E+03,0.180387E+03,0.181886E+03,
  0.183402E+03,0.184930E+03,0.186463E+03,0.188012E+03,0.189571E+03,
  0.191146E+03,0.192730E+03,0.194320E+03,0.195930E+03,0.197546E+03,
  0.199175E+03,0.200821E+03,0.202473E+03,0.204142E+03,0.205817E+03,
  0.207510E+03,0.209216E+03,0.210928E+03,0.212658E+03,0.214398E+03,
  0.216152E+03,0.217920E+03,0.219698E+03,0.221495E+03,0.223297E+03,
  0.225119E+03,0.226951E+03,0.228793E+03,0.230654E+03,0.232522E+03,
  0.234413E+03,0.236311E+03,0.238223E+03,0.240151E+03,0.242090E+03,
  0.244049E+03,0.246019E+03,0.248000E+03,0.249996E+03,0.252009E+03,
  0.254037E+03,0.256077E+03,0.258128E+03,0.260200E+03,0.262284E+03,
  0.264384E+03,0.266500E+03,0.268629E+03,0.270779E+03,0.272936E+03,
  0.275110E+03,0.277306E+03,0.279509E+03,0.281734E+03,0.283966E+03,
  0.286227E+03,0.288494E+03,0.290780E+03,0.293083E+03,0.295398E+03,
  0.297737E+03,0.300089E+03,0.302453E+03,0.304841E+03,0.307237E+03,
  0.309656E+03,0.312095E+03,0.314541E+03,0.317012E+03,0.319496E+03,
  0.322005E+03,0.324527E+03,0.327063E+03,0.329618E+03,0.332193E+03,
  0.334788E+03,0.337396E+03,0.340025E+03,0.342673E+03,0.345329E+03,
  0.348019E+03,0.350722E+03,0.353440E+03,0.356178E+03,0.358938E+03,
  0.361718E+03,0.364513E+03,0.367322E+03,0.370160E+03,0.373012E+03,
  0.375885E+03,0.378788E+03,0.381691E+03,0.384631E+03,0.387579E+03,
  0.390556E+03,0.393556E+03,0.396563E+03,0.399601E+03,0.402646E+03,
  0.405730E+03,0.408829E+03,0.411944E+03,0.415083E+03,0.418236E+03,
  0.421422E+03,0.424632E+03,0.427849E+03,0.431099E+03,0.434365E+03,
  0.437655E+03,0.440970E+03,0.444301E+03,0.447666E+03,0.451038E+03,
  0.454445E+03,0.457876E+03,0.461316E+03,0.464790E+03,0.468281E+03,
  0.471798E+03,0.475342E+03,0.478902E+03,0.482497E+03,0.486101E+03,
  0.489741E+03,0.493408E+03,0.497083E+03,0.500804E+03,0.504524E+03,
  0.508290E+03,0.512074E+03,0.515877E+03,0.519717E+03,0.523566E+03,
  0.527462E+03,0.531367E+03,0.535301E+03,0.539264E+03,0.543245E+03,
  0.547265E+03,0.551305E+03,0.555363E+03,0.559462E+03,0.563579E+03,
  0.567727E+03,0.571905E+03,0.576102E+03,0.580329E+03,0.584576E+03,
  0.588865E+03,0.593185E+03,0.597514E+03,0.601885E+03,0.606276E+03,
  0.610699E+03,0.615151E+03,0.619625E+03,0.624140E+03,0.628671E+03,
  0.633243E+03,0.637845E+03,0.642465E+03,0.647126E+03,0.651806E+03,
  0.656527E+03,0.661279E+03,0.666049E+03,0.670861E+03,0.675692E+03,
  0.680566E+03,0.685471E+03,0.690396E+03,0.695363E+03,0.700350E+03,
  0.705381E+03,0.710444E+03,0.715527E+03,0.720654E+03,0.725801E+03,
  0.730994E+03,0.736219E+03,0.741465E+03,0.746756E+03,0.752068E+03,
  0.757426E+03,0.762819E+03,0.768231E+03,0.773692E+03,0.779172E+03,
  0.784701E+03,0.790265E+03,0.795849E+03,0.801483E+03,0.807137E+03,
  0.812842E+03,0.818582E+03,0.824343E+03,0.830153E+03,0.835987E+03,
  0.841871E+03,0.847791E+03,0.853733E+03,0.859727E+03,0.865743E+03,
  0.871812E+03,0.877918E+03,0.884046E+03,0.890228E+03,0.896433E+03,
  0.902690E+03,0.908987E+03,0.915307E+03,0.921681E+03,0.928078E+03,
  0.934531E+03,0.941023E+03,0.947539E+03,0.954112E+03,0.960708E+03,
  0.967361E+03,0.974053E+03,0.980771E+03,0.987545E+03,0.994345E+03,
  0.100120E+04,0.100810E+04,0.101502E+04,0.102201E+04,0.102902E+04,
  0.103608E+04,0.104320E+04,0.105033E+04,0.105753E+04,0.106475E+04,
  0.107204E+04,0.107936E+04,0.108672E+04,0.109414E+04,0.110158E+04,
  0.110908E+04,0.111663E+04,0.112421E+04,0.113185E+04,0.113952E+04,
  0.114725E+04,0.115503E+04,0.116284E+04,0.117071E+04,0.117861E+04,
  0.118658E+04,0.119459E+04,0.120264E+04,0.121074E+04,0.121888E+04,
  0.122709E+04,0.123534E+04,0.124362E+04,0.125198E+04,0.126036E+04,
  0.126881E+04,0.127731E+04,0.128584E+04,0.129444E+04,0.130307E+04,
  0.131177E+04,0.132053E+04,0.132931E+04,0.133817E+04,0.134705E+04,
  0.135602E+04,0.136503E+04,0.137407E+04,0.138319E+04,0.139234E+04,
  0.140156E+04,0.141084E+04,0.142015E+04,0.142954E+04,0.143896E+04,
  0.144845E+04,0.145800E+04,0.146759E+04,0.147725E+04,0.148694E+04,
  0.149672E+04,0.150655E+04,0.151641E+04,0.152635E+04,0.153633E+04,
  0.154639E+04,0.155650E+04,0.156665E+04,0.157688E+04,0.158715E+04,
  0.159750E+04,0.160791E+04,0.161836E+04,0.162888E+04,0.163945E+04,
  0.165010E+04,0.166081E+04,0.167155E+04,0.168238E+04,0.169325E+04,
  0.170420E+04,0.171522E+04,0.172627E+04,0.173741E+04,0.174859E+04,
  0.175986E+04,0.177119E+04,0.178256E+04,0.179402E+04,0.180552E+04,
  0.181711E+04,0.182877E+04,0.184046E+04,0.185224E+04,0.186407E+04,
  0.187599E+04,0.188797E+04,0.190000E+04,0.191212E+04,0.192428E+04,
  0.193653E+04,0.194886E+04,0.196122E+04,0.197368E+04,0.198618E+04,
  0.199878E+04,0.201145E+04,0.202416E+04,0.203698E+04,0.204983E+04,
  0.206278E+04,0.207580E+04,0.208887E+04,0.210204E+04,0.211525E+04,
  0.212856E+04,0.214195E+04,0.215538E+04,0.216892E+04,0.218249E+04,
  0.219618E+04,0.220994E+04,0.222375E+04,0.223766E+04,0.225161E+04,
  0.226567E+04,0.227981E+04,0.229399E+04,0.230829E+04,0.232263E+04,
  0.233708E+04,0.235161E+04,0.236618E+04,0.238087E+04,0.239560E+04,
  0.241044E+04,0.242538E+04,0.244035E+04,0.245544E+04,0.247057E+04,
  0.248583E+04,0.250116E+04,0.251654E+04,0.253204E+04,0.254759E+04,
  0.256325E+04,0.257901E+04,0.259480E+04,0.261073E+04,0.262670E+04,
  0.264279E+04,0.265896E+04,0.267519E+04,0.269154E+04,0.270794E+04,
  0.272447E+04,0.274108E+04,0.275774E+04,0.277453E+04,0.279137E+04,
  0.280834E+04,0.282540E+04,0.284251E+04,0.285975E+04,0.287704E+04,
  0.289446E+04,0.291198E+04,0.292954E+04,0.294725E+04,0.296499E+04,
  0.298288E+04,0.300087E+04,0.301890E+04,0.303707E+04,0.305529E+04,
  0.307365E+04,0.309211E+04,0.311062E+04,0.312927E+04,0.314798E+04,
  0.316682E+04,0.318577E+04,0.320477E+04,0.322391E+04,0.324310E+04,
  0.326245E+04,0.328189E+04,0.330138E+04,0.332103E+04,0.334073E+04,
  0.336058E+04,0.338053E+04,0.340054E+04,0.342069E+04,0.344090E+04,
  0.346127E+04,0.348174E+04,0.350227E+04,0.352295E+04,0.354369E+04,
  0.356458E+04,0.358559E+04,0.360664E+04,0.362787E+04,0.364914E+04,
  0.367058E+04,0.369212E+04,0.371373E+04,0.373548E+04,0.375731E+04,
  0.377929E+04,0.380139E+04,0.382355E+04,0.384588E+04,0.386826E+04,
  0.389081E+04,0.391348E+04,0.393620E+04,0.395910E+04,0.398205E+04,
  0.400518E+04,0.402843E+04,0.405173E+04,0.407520E+04,0.409875E+04,
  0.412246E+04,0.414630E+04,0.417019E+04,0.419427E+04,0.421840E+04,
  0.424272E+04,0.426715E+04,0.429165E+04,0.431634E+04,0.434108E+04,
  0.436602E+04,0.439107E+04,0.441618E+04,0.444149E+04,0.446685E+04,
  0.449241E+04,0.451810E+04,0.454385E+04,0.456977E+04,0.459578E+04,
  0.462197E+04,0.464830E+04,0.467468E+04,0.470127E+04,0.472792E+04,
  0.475477E+04,0.478175E+04,0.480880E+04,0.483605E+04,0.486336E+04,
  0.489087E+04,0.491853E+04,0.494623E+04,0.497415E+04,0.500215E+04,
  0.503034E+04,0.505867E+04,0.508707E+04,0.511568E+04,0.514436E+04,
  0.517325E+04,0.520227E+04,0.523137E+04,0.526068E+04,0.529005E+04,
  0.531965E+04,0.534939E+04,0.537921E+04,0.540923E+04,0.543932E+04,
  0.546965E+04,0.550011E+04,0.553064E+04,0.556139E+04,0.559223E+04,
  0.562329E+04,0.565449E+04,0.568577E+04,0.571727E+04,0.574884E+04,
  0.578064E+04,0.581261E+04,0.584464E+04,0.587692E+04,0.590924E+04,
  0.594182E+04,0.597455E+04,0.600736E+04,0.604039E+04,0.607350E+04,
  0.610685E+04,0.614036E+04,0.617394E+04,0.620777E+04,0.624169E+04,
  0.627584E+04,0.631014E+04,0.634454E+04,0.637918E+04,0.641390E+04,
  0.644887E+04,0.648400E+04,0.651919E+04,0.655467E+04,0.659021E+04,
  0.662599E+04,0.666197E+04,0.669800E+04,0.673429E+04,0.677069E+04,
  0.680735E+04,0.684415E+04,0.688104E+04,0.691819E+04,0.695543E+04,
  0.699292E+04,0.703061E+04,0.706837E+04,0.710639E+04,0.714451E+04,
  0.718289E+04,0.722143E+04,0.726009E+04,0.729903E+04,0.733802E+04,
  0.737729E+04,0.741676E+04,0.745631E+04,0.749612E+04,0.753602E+04,
  0.757622E+04,0.761659E+04,0.765705E+04,0.769780E+04,0.773863E+04,
  0.777975E+04,0.782106E+04,0.786246E+04,0.790412E+04,0.794593E+04,
  0.798802E+04,0.803028E+04,0.807259E+04,0.811525E+04,0.815798E+04,
  0.820102E+04,0.824427E+04,0.828757E+04,0.833120E+04,0.837493E+04,
  0.841895E+04,0.846313E+04,0.850744E+04,0.855208E+04,0.859678E+04,
  0.864179E+04,0.868705E+04,0.873237E+04,0.877800E+04,0.882374E+04,
  0.886979E+04,0.891603E+04,0.896237E+04,0.900904E+04,0.905579E+04,
  0.910288E+04,0.915018E+04,0.919758E+04,0.924529E+04,0.929310E+04,
  0.934122E+04,0.938959E+04,0.943804E+04,0.948687E+04,0.953575E+04,
  0.958494E+04,0.963442E+04,0.968395E+04,0.973384E+04,0.978383E+04,
  0.983412E+04,0.988468E+04,0.993534E+04,0.998630E+04,0.100374E+05,
  0.100888E+05,0.101406E+05,0.101923E+05,0.102444E+05,0.102966E+05,
  0.103492E+05,0.104020E+05,0.104550E+05,0.105082E+05,0.105616E+05,
  0.106153E+05,0.106693E+05,0.107234E+05,0.107779E+05,0.108325E+05,
  0.108874E+05,0.109425E+05,0.109978E+05,0.110535E+05,0.111092E+05,
  0.111653E+05,0.112217E+05,0.112782E+05,0.113350E+05,0.113920E+05,
  0.114493E+05,0.115070E+05,0.115646E+05,0.116228E+05,0.116809E+05,
  0.117396E+05,0.117984E+05,0.118574E+05,0.119167E+05,0.119762E+05,
  0.120360E+05,0.120962E+05,0.121564E+05,0.122170E+05,0.122778E+05,
  0.123389E+05,0.124004E+05,0.124619E+05,0.125238E+05,0.125859E+05,
  0.126484E+05,0.127111E+05,0.127739E+05,0.128372E+05,0.129006E+05,
  0.129644E+05,0.130285E+05,0.130927E+05,0.131573E+05,0.132220E+05,
  0.132872E+05,0.133526E+05,0.134182E+05,0.134842E+05,0.135503E+05,
  0.136168E+05,0.136836E+05,0.137505E+05,0.138180E+05,0.138854E+05,
  0.139534E+05,0.140216E+05,0.140900E+05,0.141588E+05,0.142277E+05,
  0.142971E+05,0.143668E+05,0.144366E+05,0.145069E+05,0.145773E+05,
  0.146481E+05,0.147192E+05,0.147905E+05,0.148622E+05,0.149341E+05,
  0.150064E+05,0.150790E+05,0.151517E+05,0.152250E+05,0.152983E+05,
  0.153721E+05,0.154462E+05,0.155205E+05,0.155952E+05,0.156701E+05,
  0.157454E+05,0.158211E+05,0.158969E+05,0.159732E+05,0.160496E+05,
  0.161265E+05,0.162037E+05,0.162811E+05,0.163589E+05,0.164369E+05,
  0.165154E+05,0.165942E+05,0.166732E+05,0.167526E+05,0.168322E+05,
  0.169123E+05,0.169927E+05,0.170733E+05,0.171543E+05,0.172356E+05,
  0.173173E+05,0.173993E+05,0.174815E+05,0.175643E+05,0.176471E+05,
  0.177305E+05,0.178143E+05,0.178981E+05,0.179826E+05,0.180671E+05,
  0.181522E+05,0.182377E+05,0.183232E+05,0.184093E+05,0.184955E+05,
  0.185823E+05,0.186695E+05,0.187568E+05,0.188447E+05,0.189326E+05,
  0.190212E+05,0.191101E+05,0.191991E+05,0.192887E+05,0.193785E+05,
  0.194688E+05,0.195595E+05,0.196503E+05,0.197417E+05,0.198332E+05,
  0.199253E+05,0.200178E+05,0.201105E+05,0.202036E+05,0.202971E+05,
  0.203910E+05,0.204853E+05,0.205798E+05,0.206749E+05,0.207701E+05,
  0.208659E+05,0.209621E+05,0.210584E+05,0.211554E+05,0.212524E+05,
  0.213501E+05,0.214482E+05,0.215465E+05,0.216452E+05,0.217442E+05,
  0.218439E+05,0.219439E+05,0.220440E+05,0.221449E+05,0.222457E+05,
  0.223473E+05,0.224494E+05,0.225514E+05,0.226542E+05,0.227571E+05,
  0.228606E+05,0.229646E+05,0.230687E+05,0.231734E+05,0.232783E+05,
  0.233839E+05,0.234898E+05,0.235960E+05,0.237027E+05,0.238097E+05,
  0.239173E+05,0.240254E+05,0.241335E+05,0.242424E+05,0.243514E+05,
  0.244611E+05,0.245712E+05,0.246814E+05,0.247923E+05,0.249034E+05,
  0.250152E+05,0.250152E+05
};


static void QSAT_CALC(REAL QS[] /*OUT*/, REAL T[], REAL P[],
  INTEGER POINTS, REAL ES[])
{
  REAL    ONE_MINUS_EPSILON = 1.0 - EPSILON;

  /* Lowest temperature for which look-up table of
   * saturation water vapour pressure is valid (K).
   */
  REAL    T_LOW             = 183.15;

  /* Highest temperature for which look-up table of
   * saturation water vapour pressure is valid (K).
   */
  REAL    T_HIGH            = 338.15;

  /* Temperature increment of the look-up of saturation vapour
   * pressures.
   */
  REAL    DELTA_T           = 0.1;

  INTEGER ITABLE;
  REAL    ATABLE;

  /* Factor that converts from sat vapour pressure
   * in a pure water system to sat vapour pressure in air.
   */
  REAL    FSUBW;
  REAL    TT;
  INTEGER I;

  for ( I = 0; I < POINTS; I++ )
  {
    REAL T_2 = (T[I] - ZERODEGC);

    T_2      = SQR(T_2);

    /* Compute the factor that converts from sat vapour pressure in a
     * pure water system to sat vapour pressure in air, FSUBW.
     * This formula is taken from equation A4.7 of adrian gill's book:
     * Atmosphere-Ocean Dynamics. Note that his formula works in terms
     * of pressure in mb and temperature in Celsius, so conversion of
     * units leads to the slightly different equation used here.
     */
    FSUBW    = 1.0 + 1.0E-8 * P[I] * (4.5 + 6.0E-4 * T_2);

    /* Use the lookup table to find saturated vapour pressure, and store
     * it in QS.
     */
    TT       = MAX(T_LOW, T[I]);
    TT       = MIN(T_HIGH, TT);

    ATABLE   = (TT - T_LOW + DELTA_T) / DELTA_T;
    ITABLE   = (INTEGER) ATABLE;
    ATABLE   = ATABLE - (REAL) ITABLE;

    QS[I]    = (1.0 - ATABLE) * ES[ITABLE] + ATABLE * ES[ITABLE + 1];

    /* Multiply by FSUBW to convert to saturated vapour pressure in air
     * (Equation A4.6 of Adrian Gill's book).
     */
    QS[I]    = QS[I] * FSUBW;

    /* Now form the accurate expression for QS, which is a rearranged
     * version of equation A4.3 of Gill's book.
     * Note that at very low pressures we apply a fix, to prevent a
     * singularity (Qsat tends to 1. kg/kg).
     */
    QS[I]    = (EPSILON * QS[I])
      / (MAX(P[I], QS[I]) - ONE_MINUS_EPSILON * QS[I]);
  }
}


/* Saturation Specific Humidity Scheme (Qsat): Vapour to Liquid/Ice.
 * Returns a saturation mixing ratio given a temperature and pressure
 * using saturation vapour pressures calculated using the Goff-Gratch
 * formulae, adopted by the WMO as taken from Landolt-Bornstein, 1987
 * Numerical Data and Functional Relationships in Science and
 * Technology. Group V/Vol 4b Meteorology. Physical and Chemical
 * Properties of Air, p35
 *
 * QS[POINTS] saturation mixing ratio Kg/Kg.
 * T[POINTS] temperature K.
 * P[POINTS] pressure Pa.
 */
void QSAT(REAL QS[], REAL T[], REAL P[], INTEGER POINTS)
{
  QSAT_CALC(QS, T, P, POINTS, QSAT_LUT);
}


/* Saturation Specific Humidity Scheme (Qsat_Wat): Vapour to Liquid.
 * Returns a saturation mixing ratio given a temperature and pressure
 * using saturation vapour pressures calculated using the Goff-Gratch
 * formulae, adopted by the WMO as taken from Landolt-Bornstein, 1987
 * Numerical Data and Functional Relationships in Science and
 * Technology. Group V/Vol 4B Meteorology. Physical and Chemical
 * Properties of Air, P35
 *
 * Values in the lookup table are over water above and below 0 deg C.
 *
 * Note : for vapour pressure over water this formula is valid for
 * temperatures between 373K and 223K.  The values for saturated vapour
 * over water in the lookup table below are out of the lower end of
 * this range.  However it is standard WMO practice to use the formula
 * below its accepted range for use with the calculation of dew points
 * in the upper atmosphere.
 */
void QSAT_WAT(REAL QSL[], REAL T[], REAL P[], INTEGER POINTS)
{
  QSAT_CALC(QSL, T, P, POINTS, QSAT_WAT_LUT);
}


/************************************************************************/
/************************************************************************/










#ifdef jalkjdflkajsdf


/* This must be called before the first call to LSP_ICE.
 * It initialises many of the constants and allocates memory.
 * Returns 0 if all allocation was successull, -1 if malloc failed.
 */
int LSP_INIT(INTEGER POINTS);


/* This should be called to free all memory allocated by LSP_INIT(). */
void LSP_DONE(void);


void LSP_ICE(
  /* [POINTS] Air pressure at this level (Pa). */
  REAL    P[],

  /* Thickness of the layer in metres. */
  REAL    DZ,

  /* Timestep of physics in model (s). */
  REAL    TIMESTEPFIXED,

  /* [POINTS] Critical humidity for cloud formation. */
  REAL    RHCRIT[],

  /* [POINTS] Cloud ice (kg water per kg air). */
  REAL    QCF[],

  /* [POINTS] Cloud liquid water (kg water per kg air). */
  REAL    QCL[],

  /* [POINTS] Specific humidity at this level (kg wat per kg air). */
  REAL    Q[],

  /* [POINTS] On input: Rate of rainfall entering this layer from above.
   * On output: Rate of rainfall leaving this layer.
   * (kg per sq m per s).
   */
  REAL    RAIN[],

  /* [POINTS] On input: Rate of snowfall entering this layer from above.
   * On Output: Rate of snowfall leaving this layer.
   * (kg per sq m per s).
   */
  REAL    SNOW[],

  /* [POINTS] On input: Fall speed of ice into layer from above.
   * On output: Fall speed of ice into layer below.
   * (m per s).
   */
  REAL    VF[],

  /* [POINTS] Temperature at this level (K). */
  REAL    T[],

  /* [POINTS] Liquid cloud fraction in this layer. */
  REAL    CFLIQ[],

  /* [POINTS] Frozen cloud fraction in this layer. */
  REAL    CFICE[],

  /* [POINTS] Land/sea mask */
  LOGICAL BLAND[],

  /* Number of iterations in microphysics.
   * Advise 1 iteration for every 10 minutes or less of timestep.
   */
  INTEGER LS_ITER
);


struct
{
  /* Saturated sp humidity for (T,p) in layer */
  REAL *  QS;

  /* Saturated sp humidity for (T,p) in layer wrt water at all temps */
  REAL *  QSL;

  /* Cumulative fall out of snow within iterations. */
  REAL *  SNOWT;

  /* Density of air in the layer. */
  REAL *  RHO;

  /* 1.0/RHO to speed up calculations. */
  REAL *  RHOR;

  /* CFL limit */
  REAL *  DHI;

  /* 1.0/DHI */
  REAL *  DHIR;

  /* Critical humidity for ice deposition */
  REAL *  RATEQS;

  INTEGER POINTS;
} LSP_ICE_VARS;


void LSP_ICE(
  REAL    P[],
  REAL    DZ,
  REAL    TIMESTEPFIXED,
  REAL    RHCRIT[],
  REAL    QCF[],
  REAL    QCL[],
  REAL    Q[],
  REAL    RAIN[],
  REAL    SNOW[],
  REAL    VF[],
  REAL    T[],
  REAL    CFLIQ[],
  REAL    CFICE[],
  LOGICAL BLAND[],
  INTEGER LS_ITER
)
{
  /* Latent heat of condensation / Cp (K). */
  REAL    LCRCP;

  /* Latent heat of fusion / Cp (K). */
  REAL    LFRCP;

  /* Sum of the above (S for Sublimation). */
  REAL    LSRCP;

  /* Constant in wet bulb temperature calculation. */
  REAL    CONW;

  /* Loop counter (horizontal field index). */
  INTEGER I;

  /* Counter for the iterations */
  INTEGER J;

  /* Virtual temperature as at start of loop. */
  REAL    VTEMP;

  /* temperature degree C as at start of loop. */
  REAL    TEMPC;

  /* saturation vapour pressure (wrt ice below zero) */
  REAL    ESI;

  /* saturation vapour pressure (wrt water at all T) */
  REAL    ESW;

  /* increment to/from ice/snow */
  REAL    DQI;

  /* increment to/from cloud water */
  REAL    DQIL;

  /* increment to/from rain */
  REAL    DPR;

  /* fraction of ice inferred for fall speed calculations. */
  REAL    CFICETEMP;

  /* fallspeed for ice */
  REAL    FQI;

  /* 1.0/(DHI*LS_ITER) */
  REAL    DHILSITERR;

  /* saved flux of ice out of layer */
  REAL    FQIRQI;

  /* updated ice for long timestep */
  REAL    QUP;

  /* updated liquid cloud in implicit calculations */
  REAL    QCLNEW;

  /* term in melting */
  REAL    TEMP7;

  /* term in evaporation of rain */
  REAL    PR02;

  /* square of pr02 */
  REAL    PR04;

  /* term in autoconversion of cloud to rain */
  REAL    QC;

  /* denominator in deposition or evaporation of ice */
  REAL    APLUSB;

  /* density correction for fall speed */
  REAL    CORR;

  /* density correction for fall speed */
  REAL    ROCOR;

  /* surface air density */
  REAL    RHO1;

  /* Mean fall speed of rain */
  REAL    VR1;

  /* Mean fall speed of snow */
  REAL    VS1;

  /* Inverse lambda in rain exponential distribution */
  REAL    LAMR1;

  /* Expression containing calculations with lambda */
  REAL    LAMFAC1;

  /* lambda in snow exponential distribution */
  REAL    LAMS1;

  /* Mean velocity difference between rain and snow */
  REAL    FV1;

  /* Timestep of each iteration */
  REAL    TIMESTEP;

  /* Temperature correction of viscosity etc. */
  REAL    CORR2;

  /* Relative humidity required for nucleation */
  REAL    RHNUC;

  /* Temperature Factor for X1I in Cox Golding calculation */
  REAL    TCG;

  /* Inverse of TCG */
  REAL    TCGI;

  /* Constant effecting rate of deposition/sublimation of ice */
  REAL    RATEQ;

  /* Constant representing effect of sub grid distribution of ice */
  REAL    RATEQCF;

  /* Critical humidity for rain evaporation */
  REAL    RATEQSL;

  /**********************************************************************/

  LCRCP = LC / CP;
  LFRCP = LF / CP;
  LSRCP = LCRCP + LFRCP;
  CONW  = R / (EPSILON * LC);

  /**********************************************************************/

  /* ------------------------------------------------------------------ */
  /* 2.1 Start the microphysics calculations                            */
  /* ------------------------------------------------------------------ */
  /* Set up the iterations */

  TIMESTEP = TIMESTEPFIXED / (float) LS_ITER;

  /* Set up sub grid scale constants. For no sub grid scale variability
   * use the set up RATEQ=1.0, RATEQCF=0.0, RATEQS=1.0.
   * Ideally represent these in a comdeck but require RHCRIT
   */

  RATEQ   = 1.0;
  RATEQCF = 0.0;

  /* Set up SNOWT(I) to be zero for all I. */
  for ( I = 0; I < LSP_ICE_VARS.POINTS; I++ ) {
    LSP_ICE_VARS.SNOWT[I]  = 0.0;
    LSP_ICE_VARS.RATEQS[I] = RHCRIT[I];
  }

  /* ------------------------------------------------------------------ */
  /* 2.2 Start iterating.                                               */
  /* ------------------------------------------------------------------ */

  for ( J = 0; J < LS_ITER; J++ ) {

    /* ---------------------------------------------------------------- */
    /* 2.3 Calculate sat humidity mixing ratio(respect to ice below     */
    /*     zero)                                                        */
    /* ---------------------------------------------------------------- */
    QSAT(LSP_ICE_VARS.QS, T, P, LSP_ICE_VARS.POINTS);

    /* ---------------------------------------------------------------- */
    /* 2.4 Calculate sat humidity mixing ratio(respect to water at all  */
    /*     temps)                                                       */
    /* ---------------------------------------------------------------- */
    QSAT_WAT(LSP_ICE_VARS.QSL, T, P, LSP_ICE_VARS.POINTS);

    /* ---------------------------------------------------------------- */
    /* 2.5 Start loop over points.                                      */
    /* ---------------------------------------------------------------- */
    for ( I = 0; I < LSP_ICE_VARS.POINTS; I++ ) {
      REAL RHODZ;

      /* -------------------------------------------------------------- */
      /* 3.1 Calculate density of air, RHO, via virtual temperature.    */
      /* -------------------------------------------------------------- */
      VTEMP                = T[I] * (1.0 + C_VIRTUAL * Q[I] - QCL[I]
        - QCF[I]);
      LSP_ICE_VARS.RHO[I]  = P[I] / (R * VTEMP);
      RHODZ                = LSP_ICE_VARS.RHO[I] * DZ;
      LSP_ICE_VARS.RHOR[I] = 1.0 / LSP_ICE_VARS.RHO[I];
      RHO1                 = 1.0;

      /* Correction factor of fall speeds etc. due to density. */
      CORR      = pow(RHO1 * LSP_ICE_VARS.RHOR[I], 0.4);

      /* Correction factor in viscosity etc. due to temperature. */
      CORR2     = pow(T[I] / 273.0, 1.5) * (393.0 / (T[I] + 120.0));

      /* Combined correction factor */
      ROCOR     = (REAL) sqrt((double) (LSP_ICE_VARS.RHO[I] * CORR
        * CORR2));

      /* Calculate RATEQS as a function of RHCRIT and CFICE[I] */
      LSP_ICE_VARS.RATEQS[I] = RHCRIT[I] + CFICE[I] * (1.0 - RHCRIT[I]);
      RATEQSL                = RHCRIT[I] + CFLIQ[I] * (1.0 - RHCRIT[I]);

      /* -------------------------------------------------------------- */
      /* 3.2 set T deg C sat vapour pressures in N/m2                   */
      /* -------------------------------------------------------------- */
      TEMPC = T[I] - ZERODEGC;
      ESI   = LSP_ICE_VARS.QS[I] * P[I] / EPSILON;
      ESW   = LSP_ICE_VARS.QSL[I] * P[I] / EPSILON;

      /* Calculate a temperature factor for N0snow. CX[13]=1.0 if there
       * is a temperature dependence, and 0.0 if there is not.
       */
      TCG   = exp(- CX[13] * TEMPC / 8.18);
      TCGI  = 1.0 / TCG;

      /* -------------------------------------------------------------- */
      /* 4 Check that ice cloud fraction is sensible.                   */
      /* -------------------------------------------------------------- */
      CFICETEMP = CFICE[I];

      /* The possibility exists in multiple iterations that ice cloud
       * fraction is equal to zero but nucleation and deposition from the
       * last iteration has produced a finite ice content. Hence this
       * section produces a fix which will stop the scheme crashing.
       * Only need to use for more than 1 iteration
       */
      if ( LS_ITER > 1 ) {
        if ( QCF[I] > 0.0 && CFICE[I] < 0.1) {
          CFICETEMP = MAX(CFLIQ[I], 0.1);
        }                                                          
      }                                                            

      /* -------------------------------------------------------------- */
      /* 5.1 Calculate fallspeeds for ice and water and limit           */
      /*     to CFL criteria for time to fall through layer.            */
      /* -------------------------------------------------------------- */

      /* Estimate fall speed out of this layer. We want to avoid advecting
       * very small amounts of snow between layers, as this can cause
       * numerical problems in other routines, so if QCF is smaller than
       * a single nucleation mass per metre cubed don't advect it.
       */
      if ( QCF[I] > M0 ) {
        /* Estimate the mean fall speed across the entire gridbox.
         * Use a top hat distrubution within the gridbox
         */
        FQI = CONSTP[3] * CORR
          * pow(LSP_ICE_VARS.RHO[I] * QCF[I] * CONSTP[1] * TCGI
            / CFICETEMP, CX[3]);
      }
      else {                                                              
        /* QCF is smaller than zero so set fall speed to zero */
        FQI = 0.0;
      }                                                            

      /* Calculate CFL quantity of timestep over level separation. */
      LSP_ICE_VARS.DHI[I]     = TIMESTEP / DZ;

      /* Define DHIR and DHILSITERR to speed up calculations. */
      LSP_ICE_VARS.DHIR[I]    = 1.0 / LSP_ICE_VARS.DHI[I];
      DHILSITERR              = 1.0 / (LSP_ICE_VARS.DHI[I] * LS_ITER);

      /* See if fall speed is small enough that not all the ice falls out
       * of the layer.
       */
      if ( VF[I] < LSP_ICE_VARS.DHIR[I] ) {
        /* short timestep solution */
        VF[I] = FQI;

        if ( VF[I] < LSP_ICE_VARS.DHIR[I] ) {
          /* flux out is just represented by the fall speed estimated
           * above
           */
          FQIRQI = FQI * LSP_ICE_VARS.RHO[I] * QCF[I];
        }
        else {
          /* cannot allow more ice to leave than was already there */
          FQIRQI = LSP_ICE_VARS.RHO[I] * QCF[I] * LSP_ICE_VARS.DHIR[I];
        }

        /* calculate new ice content in this layer by flux divergence */
        QCF[I] = QCF[I] + (SNOW[I] - FQIRQI) * LSP_ICE_VARS.DHI[I]
          * LSP_ICE_VARS.RHOR[I];
      }
      else {                                                              
        /* long timestep case */
        QUP    = SNOW[I] * LSP_ICE_VARS.RHOR[I] / VF[I];
        FQIRQI = SNOW[I] - (RHODZ * (QUP - QCF[I]) / TIMESTEP);
        QCF[I] = QUP;

        /* VF must be as least as great as the fall velocity of the
         * current layer 
         */
        if ( VF[I] < FQI ) VF[I] = FQI;
      }                                                            

      /* -------------------------------------------------------------- */
      /* 5.2 Snow is used to save fall out of layer                     */
      /*     for calculation of fall into next layer                    */
      /* -------------------------------------------------------------- */
      LSP_ICE_VARS.SNOWT[I] = LSP_ICE_VARS.SNOWT[I] + FQIRQI / LS_ITER;

      /* -------------------------------------------------------------- */
      /* Transfer processes only active at T less than 0 deg C          */
      /* -------------------------------------------------------------- */
      if ( T[I] < ZERODEGC ) {
        /* ------------------------------------------------------------ */
        /* 6   Nucleation of snow - explicit                            */
        /* ------------------------------------------------------------ */

        /* ------------------------------------------------------------ */
        /* 6.1 Homogenous nucleation takes place at temperatures less   */
        /*     than THOMO                                               */
        /* ------------------------------------------------------------ */
        if ( T[I] < (ZERODEGC + THOMO) ) {
          /* Turn all liquid to ice */
          QCF[I] = QCF[I] + QCL[I];
          T[I]   = T[I] + LFRCP * QCL[I];
          QCL[I] = 0.0;
        }                                                         

        /* ------------------------------------------------------------ */
        /* 6.2 Heteorgenous nucleation occurs for temps less than TNUC  */
        /*     deg C                                                    */
        /* ------------------------------------------------------------ */
        if ( T[I] < (ZERODEGC + TNUC) ) {
          REAL TC = T[I] - ZERODEGC;

          /* Calculate number of active ice nucleii */
          DQI     = MIN(0.01 * exp(-0.6 * TEMPC), 1.0E5);

          /* Each nucleus can grow to arbitary mass of M0 kg */
          DQI     = M0 * DQI * LSP_ICE_VARS.RHOR[I];

          /* RHNUC represents how much moisture is available for ice
           * formation.
           */
          RHNUC   = (188.92 +2.81 * TC + 0.013336 * SQR(TC) - 10.0)
            * 0.01;
          RHNUC   = MIN(RHNUC, 1.0);

          /* Predict transfer of mass to ice. */
          DQI     = MAX(MIN(DQI, Q[I] + QCL[I] - LSP_ICE_VARS.RATEQS[I]
            * MAX(LSP_ICE_VARS.QSL[I] * RHNUC, LSP_ICE_VARS.QS[I])), 0.0);
          QCF[I]  = QCF[I] + DQI;

          /* This comes initially from liquid water */
          DQIL    = MIN(DQI, QCL[I]);
          QCL[I]  = QCL[I] - DQIL;
          T[I]    = T[I] + LFRCP * DQIL;

          /* If more moisture is required then nucleation removes from
           * vapour.
           */
          DQI     = DQI - DQIL;
          T[I]    = T[I] + LSRCP * DQI;
          Q[I]    = Q[I] - DQI;
        }

        /* ------------------------------------------------------------ */
        /* 7 Deposition/Sublimation of snow - explicit                  */
        /* ------------------------------------------------------------ */
        if ( QCF[I] > M0 ) {
          REAL T_2 = SQR(T[I]);
          REAL T_3 = T[I] * T_2;

          /* Calculate transfer rate as a function of QCF and T */
          PR02   = LSP_ICE_VARS.RHO[I] * QCF[I] * CONSTP[1] * TCGI;
          APLUSB = (APB1 - APB2 * T[I]) * ESI;
          APLUSB = APLUSB + T_3 * P[I] * APB3;
          DQI    = TCG * CONSTP[5] * T_2 * ESI * RATEQ
            * (MIN(Q[I] + QCL[I], LSP_ICE_VARS.QSL[I]) - RATEQCF * QCF[I]
              - LSP_ICE_VARS.RATEQS[I] * LSP_ICE_VARS.QS[I])
            * (0.65 * CONSTP[13] * CORR2 * pow(PR02, CX[1])
              + CONSTP[6] * ROCOR * pow(PR02, CX[2]))
              / (LSP_ICE_VARS.QS[I] * APLUSB * LSP_ICE_VARS.RHO[I]);

          /* Limits depend on whether deposition or sublimation occurs */
          if ( DQI > 0.0 ) {
            /* Deposition is occuring. Limit is available moisture. */
            DQI = MIN(DQI * TIMESTEP, (Q[I] + QCL[I] - RATEQCF * QCF[I]
              - LSP_ICE_VARS.RATEQS[I] * LSP_ICE_VARS.QS[I]) / (1.0
                + RATEQCF));
          }
          else {                                                          
            /* Sublimation is occuring. Limits are spare moisture capacity
             * and QCF
             */
            DQI = MAX(MAX(DQI * TIMESTEP,
              (Q[I] + QCL[I] - RATEQCF * QCF[I]
                - LSP_ICE_VARS.RATEQS[I] * LSP_ICE_VARS.QS[I]) / (1.0
                  + RATEQCF)),
              - QCF[I]);
          }                                                        

          /* Adjust ice content */
          QCF[I] = QCF[I] + DQI;
          DQIL   = MAX(MIN(DQI, QCL[I]), 0.0);

          /* Adjust liquid content (deposits before vapour by Bergeron
           * Findeison process).
           */
          QCL[I] = QCL[I] - DQIL;
          T[I]   = T[I] + LFRCP * DQIL;
          DQI    = DQI - DQIL;

          /* Adjust vapour content */
          Q[I]   = Q[I] - DQI;
          T[I]   = T[I] + LSRCP * DQI;
        }                                                          

        /* ------------------------------------------------------------ */
        /* 8 Riming of snow by cloud water -implicit in QCL             */
        /* ------------------------------------------------------------ */
        if ( QCF[I] > M0 && QCL[I] > 0.0 ) {                        
          QCLNEW = QCL[I] / (1.0 + CONSTP[4] * TCG * CORR * TIMESTEP
            * pow(LSP_ICE_VARS.RHO[I] * QCF[I]  * CONSTP[1] * TCGI,
              CX[4]));

          /* Recalculate water contents */
          QCF[I] = QCF[I] + (QCL[I] - QCLNEW);
          T[I]   = T[I] + LFRCP * (QCL[I] - QCLNEW);
          QCL[I] = QCLNEW;
        }                                                          

        /* ------------------------------------------------------------ */
        /* 9 Capture of rain by snow - implicit in rain                 */
        /*   This is a candidate to be removed to speed things up.      */
        /* ------------------------------------------------------------ */
        if ( RAIN[I] > 0.0 && QCF[I] > M0 ) {
          REAL L_2;
          REAL L_3;
          REAL L_4;
          REAL L_5;
          REAL L_6;

          /* Calculate velocities */
          VR1 = CORR * CONSTP[11] / 6.0
            * pow(RAIN[I] / (CONSTP[8] * CORR), CX[5]);
          VS1 = CONSTP[3] * CORR
            * pow(LSP_ICE_VARS.RHO[I] * QCF[I] * CONSTP[1] * TCGI, CX[3]);

          /* Estimate the mean absolute differences in velocities. */
          FV1 = MAX(ABS(VR1 - VS1), (VR1 + VS1) / 8.0);

          /* Calculate functions of slope parameter lambda */
          /* MOD fdm2f404 changes next 3 active lines */
          /* LAMR1 and LAMS1 are INVERSE lambda values */
          LAMR1   = pow(RAIN[I] / (CONSTP[8] * CORR), CX[10]);
          LAMS1   = pow(LSP_ICE_VARS.RHO[I] * QCF[I] * CONSTP[1] * TCGI,
            -CX[6]);

          L_2     = SQR(LAMR1);
          L_3     = L_2 * LAMR1;
          L_4     = SQR(L_2);
          L_5     = L_2 * L_3;
          L_6     = SQR(L_3);

          LAMFAC1 = CONSTP[16] * (L_6 * pow(LAMS1, CX[16]))
                  + CONSTP[15] * (L_5 * pow(LAMS1, CX[15]))
                  + CONSTP[14] * (L_4 * pow(LAMS1, CX[14]));

          /* Calculate transfer */
          DPR = TCG * CONSTP[9] * pow(LAMS1, -CX[8]) * pow(LAMR1, -CX[9])
            * FV1 * LAMFAC1 * TIMESTEP * LSP_ICE_VARS.RHOR[I];
          DPR = MIN(DPR, RAIN[I] * (LSP_ICE_VARS.DHI[I] * LS_ITER)
            * LSP_ICE_VARS.RHOR[I]);

          /* Adjust ice and rain contents */
          QCF[I]  = QCF[I] + DPR;
          RAIN[I] = RAIN[I] - DPR * LSP_ICE_VARS.RHO[I] * DHILSITERR;
          T[I]    = T[I] + LFRCP * DPR;
        }                                                          

        /* ------------------------------------------------------------ */
        /* End of transfer processes only active at T less than 0 deg C */
        /* ------------------------------------------------------------ */
      }                                                            

      /* -------------------------------------------------------------- */
      /* 10 Evaporate melting snow - implicit in subsaturation          */
      /* -------------------------------------------------------------- */
      if ( QCF[I] > M0 && T[I] > ZERODEGC ) {
        /* Calculate transfer as a function of QCF, T and specific
         * humidity
         */
        PR02 = LSP_ICE_VARS.RHO[I] * QCF[I] * CONSTP[1] * TCGI;
        PR04 = (APB4 - APB5 * T[I]) * ESW + APB6 * P[I] * CUBE(T[I]);
        DPR  = TCG * RATEQ * CONSTP[5] * SQR(T[I]) * ESW * TIMESTEP
          * (0.65 * CONSTP[13] * CORR2 * pow(PR02, CX[1])
          + CONSTP[6] * ROCOR * pow(PR02, CX[2]))
            / (LSP_ICE_VARS.QSL[I] * LSP_ICE_VARS.RHO[I] * PR04);
        DPR  = DPR
          * MAX((LSP_ICE_VARS.RATEQS[I] * LSP_ICE_VARS.QSL[I]+ RATEQCF
            * QCF[I] - Q[I] - QCL[I]), 0.0)
          / (1.0 + DPR * (1.0 + RATEQCF));

        /* Extra check to see we don't get a negative QCF */
        DPR  = MIN(DPR, QCF[I]);

        /* Update values of ice and vapour */
        QCF[I] = QCF[I] - DPR;
        Q[I]   = Q[I] + DPR;
        T[I]   = T[I]- DPR * LSRCP;
      }                                                            

      /* -------------------------------------------------------------- */
      /* 11 Melting of snow - explicit                                  */
      /*    USE WET BULB TEMP (DEG.C) IN SNOW MELT CALC.                */
      /*    Use a numerical approximation.                              */
      /* -------------------------------------------------------------- */
      if ( QCF[I] > M0 && T[I] > ZERODEGC ) {
        TEMPC = T[I] - ZERODEGC;

        /* An approximate calculation of wet bulb temperature */
        TEMP7 = TEMPC - RATEQ
          * MAX((LSP_ICE_VARS.RATEQS[I] * LSP_ICE_VARS.QSL[I]
            + RATEQCF * QCF[I] - Q[I]), 0.0)
          * (TW1 + TW2 * (P[I] - TW3) - TW4 * (T[I] - TW5) );
        TEMP7 = MAX(TEMP7, 0.0);
        /* End of wet bulb temp formulations. */

        PR02 = LSP_ICE_VARS.RHO[I] * QCF[I] * CONSTP[1] * TCGI;
        DPR  = TCG * CONSTP[7] * TIMESTEP
          * (0.65 * CONSTP[13] * CORR2 * pow(PR02, CX[1])
            + CONSTP[6] * ROCOR * pow(PR02, CX[2]))
          * LSP_ICE_VARS.RHOR[I];

        /* Solve implicitly in terms of temperature */
        DPR  = TEMP7 * (1.0 - 1.0 / (1.0 + DPR * LFRCP)) / LFRCP;
        DPR  = MIN(DPR, QCF[I]);

        /* Update values of ice and Rain */
        QCF[I]  = QCF[I] - DPR;
        RAIN[I] = RAIN[I] + DPR * LSP_ICE_VARS.RHO[I] * DHILSITERR;
        T[I]    = T[I] - LFRCP * DPR;
      }                                                            

      /* -------------------------------------------------------------- */
      /* 12 Evaporation of rain - implicit in subsaturation             */
      /* -------------------------------------------------------------- */
      if ( RAIN[I] > 0.0 ) {
        PR04 = (APB4 - APB5 * T[I]) * ESW + APB6 * P[I] * CUBE(T[I]);

        /* New, consistent evaporation method, with rain fall speed
         * relationship.
         * LAMR1 is now an INVERSE lambda value
         */
        LAMR1 = pow(RAIN[I] / (CONSTP[8] * CORR), CX[10]);
        DPR   = CONSTP[2] * SQR(T[I]) * ESW * TIMESTEP;
        DPR   = DPR * ( (0.78 * CORR2 * pow(LAMR1, CX[12]))
          + (CONSTP[12] * ROCOR * (pow(LAMR1, CX[11]))) );
        DPR   = DPR * RATEQ / (LSP_ICE_VARS.QSL[I] * LSP_ICE_VARS.RHO[I]
          * PR04);

        /* Calculate transfers. */
        DPR   = DPR * MAX((RATEQSL * LSP_ICE_VARS.QSL[I] - Q[I] - QCL[I]),
          0.0) / (1.0 + DPR);
        DPR   = DPR * LSP_ICE_VARS.RHO[I] * DHILSITERR;
        DPR   = MIN(DPR, RAIN[I]);

        /* Update values of rain and vapour */
        RAIN[I] = RAIN[I] - DPR;
        Q[I]    = Q[I] + DPR * LSP_ICE_VARS.DHI[I] * LS_ITER
          * LSP_ICE_VARS.RHOR[I];
        T[I]    = T[I] - DPR * LCRCP * LSP_ICE_VARS.DHI[I] * LS_ITER
          * LSP_ICE_VARS.RHOR[I];
      }                                                            

      /* -------------------------------------------------------------- */
      /* 13 Accretion of cloud on rain - implicit in liquid water       */
      /*    content                                                     */
      /* -------------------------------------------------------------- */
      if ( RAIN[I] > 0.0 && QCL[I] > 0.0 ) {
        /* New accretion formulation. */
        PR02   = RAIN[I] / (CONSTP[8] * CORR);
        QCLNEW = QCL[I] / ((1.0 + CONSTP[10] * CORR * TIMESTEP
          * pow(PR02, CX[7])));

        /* Now calculate increments to rain. */
        RAIN[I] = RAIN[I] + (QCL[I] - QCLNEW) * LSP_ICE_VARS.RHO[I]
          * DHILSITERR;
        QCL[I]  = QCL[I] - (QCL[I] - QCLNEW);
      }                                                          

      /* -------------------------------------------------------------- */
      /* 14 Autoconversion of cloud to rain - explicit                  */
      /* -------------------------------------------------------------- */
      if ( QCL[I] > 0.0 && CFLIQ[I] > 0.0 ) {
        /* Use a liquid cloud fraction here as this term is very
         * non-linear
         * The section below is a simple way of proceeding.
         */
        if ( BLAND[I] ) {
          /* Land point */
          QC  = MIN(AUTOLIM_LAND * CFLIQ[I] * LSP_ICE_VARS.RHOR[I],
            QCL[I]);
          DPR = MIN(AUTORATE_LAND * pow(LSP_ICE_VARS.RHO[I] * QCL[I]
            / CFLIQ[I], 1.333) * TIMESTEP * QCL[I] / CORR2, QCL[I] - QC);
        }
        else {
          /* Sea point */
          QC  = MIN(AUTOLIM_SEA * CFLIQ[I] * LSP_ICE_VARS.RHOR[I],
            QCL[I]);
          DPR = MIN(AUTORATE_SEA * pow(LSP_ICE_VARS.RHO[I] * QCL[I]
            / CFLIQ[I], 1.333) * TIMESTEP * QCL[I] / CORR2, QCL[I] - QC);
        }                                                        
        /* End of calculation of autoconversion amount DPR */

        QCL[I]  = QCL[I] - DPR;
        RAIN[I] = RAIN[I] + DPR * LSP_ICE_VARS.RHO[I] * DHILSITERR;
      }

    /* ---------------------------------------------------------------- */
    /* 15 Now continue the loops over points and iterations.            */
    /* ---------------------------------------------------------------- */
    }
  }


  /* Copy contents of SNOWT to SNOW, to fall into next layer down */
  for ( I = 0; I < LSP_ICE_VARS.POINTS; I++ ) {
    SNOW[I] = LSP_ICE_VARS.SNOWT[I];

    /* ---------------------------------------------------------------- */
    /* 16 Melt any SNOW which has reached here, as long as T is large   */
    /*    enough                                                        */
    /* ---------------------------------------------------------------- */

    /* Only use if long timestep case. In which case melt the excess
     * snow which falls straight through a layer. Use DQI variable to
     * save space.
     * DQI APPROXIMATELY represents the excess SNOW.
     */

    DQI = MIN(VF[I] * LSP_ICE_VARS.DHI[I] - 1.0, 1.0);

    if ( DQI > 0.0 ) {
      /* Long timestep case */
      TEMPC = T[I] - ZERODEGC;

      if ( SNOW[I] > 0.0 && T[I] > ZERODEGC ) {
        /* Numerical approximation of wet bulb temperature. */
        TEMP7 = TEMPC - RATEQ
          * MAX((LSP_ICE_VARS.RATEQS[I] * LSP_ICE_VARS.QSL[I] + RATEQCF
            * QCF[I] - Q[I]), 0.0)
          * (TW1 + TW2 * (P[I] -TW3 ) - TW4 * (T[I] - TW5) );
        TEMP7 = MAX(TEMP7, 0.0);
        /* End of wet bulb calculation */

        DPR = TEMP7 / (LFRCP * LS_ITER);
        DPR = MIN(DPR, SNOW[I] * LSP_ICE_VARS.DHI[I]
          * LSP_ICE_VARS.RHOR[I] * DQI);

        /* Update values of snow and rain */
        SNOW[I] = SNOW[I] - DPR * LSP_ICE_VARS.RHO[I]
          * LSP_ICE_VARS.DHIR[I];
        RAIN[I] = RAIN[I] + DPR * LSP_ICE_VARS.RHO[I]
          * LSP_ICE_VARS.DHIR[I];
        T[I]    = T[I] - LFRCP * DPR * LS_ITER;
      }
    }

    /* ---------------------------------------------------------------- */
    /* 17 Remove any small amount of QCF which is left over to be tidy. */
    /*    If QCF is less than 1e-10 kg/kg (hardwired) and isn't growing */
    /*    by deposition (assumed to be given by RHCRIT) then remove it. */
    /* ---------------------------------------------------------------- */
    if ( QCF[I] < 1.0E-10
      && (T[I] > ZERODEGC || (Q[I] + QCL[I] < RHCRIT[I]
          * LSP_ICE_VARS.QS[I])
        || QCF[I] < 0.0) ) {
      Q[I]   = Q[I] + QCF[I];
      T[I]   = T[I] - LSRCP * QCF[I];
      QCF[I] = 0.0;
    }
  }
}
void LSP_DONE(void)
{
#define FREE(p) free(p); p = NULL

  FREE(LSP_ICE_VARS.QS);
  FREE(LSP_ICE_VARS.QSL);
  FREE(LSP_ICE_VARS.SNOWT);
  FREE(LSP_ICE_VARS.RHO);
  FREE(LSP_ICE_VARS.RHOR);
  FREE(LSP_ICE_VARS.DHI);
  FREE(LSP_ICE_VARS.DHIR);
  FREE(LSP_ICE_VARS.RATEQS);

#undef FREE
}


int LSP_INIT(POINTS)
{
#define alloc_real(p, size) p = malloc(sizeof(REAL) * size);\
  if ( p == NULL ) { LSP_DONE(); return -1; }


  REAL R_THRESH_3 = CUBE(R_THRESH);

  /* The numbers 5907.24 and 4188.79 represent combinations of
   * physical constants. Do NOT change them.
   * PLEASE REFER TO UMDP 26 EQUATIONS P26.129 TO P26.136 FOR
   * AN EXPLANATION OF THE AUTOCONVERSION PARAMETERIZATION.
   * REMEMBER THAT THE PARAMETERIZATION IS VERY ROUGH.
   */
  C_VIRTUAL     = 1.0 / EPSILON - 1.0;
  AUTORATE_LAND = 5907.24 * EC_AUTO * INHOMOG_RATE * N_DROP_LAND_CR;
  AUTORATE_SEA  = 5907.24 * EC_AUTO * INHOMOG_RATE * N_DROP_SEA_CR;
  AUTOLIM_LAND  = 4188.79 * R_THRESH_3 * N_DROP_LAND * INHOMOG_LIM;
  AUTOLIM_SEA   = 4188.79 * R_THRESH_3 * N_DROP_SEA  * INHOMOG_LIM;

  LSPCON(CX, CONSTP);

  LSP_ICE_VARS.QSL    = NULL;
  LSP_ICE_VARS.SNOWT  = NULL;
  LSP_ICE_VARS.RHO    = NULL;
  LSP_ICE_VARS.RHOR   = NULL;
  LSP_ICE_VARS.DHI    = NULL;
  LSP_ICE_VARS.DHIR   = NULL;
  LSP_ICE_VARS.RATEQS = NULL;

  alloc_real(LSP_ICE_VARS.QS, POINTS);
  alloc_real(LSP_ICE_VARS.QSL, POINTS);
  alloc_real(LSP_ICE_VARS.SNOWT, POINTS);
  alloc_real(LSP_ICE_VARS.RHO, POINTS);
  alloc_real(LSP_ICE_VARS.RHOR, POINTS);
  alloc_real(LSP_ICE_VARS.DHI, POINTS);
  alloc_real(LSP_ICE_VARS.DHIR, POINTS);
  alloc_real(LSP_ICE_VARS.RATEQS, POINTS);

  LSP_ICE_VARS.POINTS = POINTS;

  return 0;


#undef alloc_real
}



#endif


