/*
 * This is the LSP_ICE FORTRAN subroutine converted to C.
 *
 * HISTORY
 * 2004-05-25  ao  started conversion.
 *
 * AUTHORS
 * Andrew Openshaw
 */

#if !defined(LSP_ICE_H)
#define LSP_ICE_H

/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "FORTRAN_TYPES.h"

/************************************************************************/


/* This must be called before the first call to LSP_ICE.
 * It initialises many of the constants.
 */
void LSP_INIT(void);


/* Calculate rain and snow fluxes for one level at one grid point. */
void LSP_ICE(
  /* Air pressure at this level (Pa). */
  REAL    P,

  /* Thickness of the layer in metres. */
  REAL    DZ,

  /* Timestep of physics in model (s). */
  REAL    TIMESTEPFIXED,

  /* Critical humidity for cloud formation. */
  REAL    RHCRIT,

  /* Cloud ice (kg water per kg air). */
  REAL    QCF,

  /* Cloud liquid water (kg water per kg air). */
  REAL    QCL,

  /* Specific humidity at this level (kg wat per kg air). */
  REAL    Q,

  /* On input: Rate of rainfall entering this layer from above.
   * On output: Rate of rainfall leaving this layer.
   * (kg per sq m per s).
   */
  REAL *  RAIN,

  /* On input: Rate of snowfall entering this layer from above.
   * On Output: Rate of snowfall leaving this layer.
   * (kg per sq m per s).
   */
  REAL *  SNOW,

  /* On input: Fall speed of ice into layer from above.
   * On output: Fall speed of ice into layer below.
   * (m per s).
   */
  REAL *  VF,

  /* Temperature at this level (K). */
  REAL    T,

  /* Liquid cloud fraction in this layer. */
  REAL    CFLIQ,

  /* Frozen cloud fraction in this layer. */
  REAL    CFICE,

  /* Land/sea mask */
  LOGICAL BLAND,

  /* Number of iterations in microphysics.
   * Advise 1 iteration for every 10 minutes or less of timestep.
   */
  INTEGER LS_ITER
);


/* Large scale cloud scheme.
 * Purpose:
 *   This subroutine calculates liquid and ice cloud fractional cover
 *   for use with the enhanced precipitation microphysics scheme.
 *
 * Method:
 *   Statistical cloud scheme separates input moisture into specific
 *   humidity and cloud liquid water. Temperature calculated from liquid
 *   water temperature. Cloud fractions calculated from statistical
 *   relation between cloud fraction and cloud liquid/ice water content.
 *
 * Documentation: UMDP No. 29
 *
 * Returns -1 if malloc failed, 0 if OK.
 */
int LS_CLD(
  /* [LEVELS] pressure at all points (Pa). */
  REAL       P[],

  /* [LEVELS] Critical relative humidity.
   * See the the paragraph incorporating eqs P292.11 to P292.14;
   * the values need to be tuned for the give set of levels.
   */
  REAL       RHCRIT[],

  /* No. of levels being processed. */
  INTEGER    LEVELS,

  /* [LEVELS]
   * On input: Liquid/frozen water temperature (TL) (K).
   * On output: Temperature at processed levels (K).
   */
  REAL       T[],

  /*OUT [LEVELS]
   * Cloud fraction at processed levels (decimal fraction).
   */
  REAL       CF[],

  /* [LEVELS]
   * On input: Total water content (QW) (kg per kg air).
   * On output: Specific humidity at processed levels
   *   (kg water per kg air).
   */
  REAL       Q[],

  /* [LEVELS]
   * Cloud ice content at processed levels (kg water per kg air).
   */
  REAL       QCF[],

  /*OUT [LEVELS]
   * Cloud liquid water content at processed levels (kg per kg air).
   */
  REAL       QCL[],

  /*OUT [LEVELS]
   * Liquid cloud fraction at processed levels (decimal fraction).
   */
  REAL       CFL[],

  /*OUT [LEVELS]
   * Frozen cloud fraction at processed levels (decimal fraction).
   */
  REAL       CFF[]
);


/************************************************************************/
/************************************************************************/

#endif
